@@ -27,7 +27,9 @@
 #include <string.h>
 #include <gst/gst-i18n-plugin.h>
 
-static GstElementDetails gst_ogg_demux_details =
+#include <gst/base/gsttypefindhelper.h>
+
+static const GstElementDetails gst_ogg_demux_details =
 GST_ELEMENT_DETAILS ("Ogg demuxer",
     "Codec/Demuxer",
     "demux ogg streams (info about ogg: http://xiph.org)",
@@ -133,10 +135,13 @@ struct _GstOggPad
   GstClockTime start_time;      /* the timestamp of the first sample */
 
   gint64 first_granule;         /* the granulepos of first page == first sample in next page */
-  GstClockTime first_time;      /* the timestamp of the second page */
+  GstClockTime first_time;      /* the timestamp of the second page or granuletime of first page */
 
   ogg_stream_state stream;
 
+  gboolean discont;
+  GstFlowReturn last_ret;       /* last return of _pad_push() */
+
   gboolean dynamic;             /* True if the internal element had dynamic pads */
   guint padaddedid;             /* The signal id for element::pad-added */
 };
@@ -177,6 +182,7 @@ struct _GstOggDemux
   gboolean segment_running;
 
   GstEvent *event;
+  GstEvent *newsegment;         /* pending newsegment to be sent from _loop */
 
   gint64 current_granule;
 
@@ -222,7 +228,6 @@ static const GstQueryType *gst_ogg_pad_query_types (GstPad * pad);
 static gboolean gst_ogg_pad_src_query (GstPad * pad, GstQuery * query);
 static gboolean gst_ogg_pad_event (GstPad * pad, GstEvent * event);
 static GstCaps *gst_ogg_pad_getcaps (GstPad * pad);
-static GstCaps *gst_ogg_type_find (ogg_packet * packet);
 static GstOggPad *gst_ogg_chain_get_stream (GstOggChain * chain,
     glong serialno);
 
@@ -232,6 +237,8 @@ static gboolean gst_ogg_pad_query_convert (GstOggPad * pad,
 static GstClockTime gst_annodex_granule_to_time (gint64 granulepos,
     gint64 granulerate_n, gint64 granulerate_d, guint8 granuleshift);
 
+static GstFlowReturn gst_ogg_demux_combine_flows (GstOggDemux * ogg,
+    GstOggPad * pad, GstFlowReturn ret);
 
 static GstPadClass *ogg_pad_parent_class = NULL;
 
@@ -410,6 +417,20 @@ gst_ogg_pad_src_query (GstPad * pad, GstQuery * query)
       gst_query_set_duration (query, GST_FORMAT_TIME, ogg->total_time);
       break;
     }
+    case GST_QUERY_SEEKING:
+    {
+      GstFormat format;
+
+      gst_query_parse_seeking (query, &format, NULL, NULL, NULL);
+      if (format == GST_FORMAT_TIME) {
+        gst_query_set_seeking (query, GST_FORMAT_TIME, ogg->seekable,
+            0, ogg->total_time);
+      } else {
+        res = FALSE;
+      }
+      break;
+    }
+
     default:
       res = gst_pad_query_default (pad, query);
       break;
@@ -750,6 +771,8 @@ gst_ogg_pad_internal_chain (GstPad * pad, GstBuffer * buffer)
 
   if (oggpad->start_time == GST_CLOCK_TIME_NONE) {
     oggpad->start_time = timestamp;
+    GST_DEBUG_OBJECT (oggpad, "new start time: %" GST_TIME_FORMAT,
+        GST_TIME_ARGS (timestamp));
   }
 
   gst_buffer_unref (buffer);
@@ -781,6 +804,7 @@ internal_element_pad_added_cb (GstElement * element, GstPad * pad,
 static gboolean
 gst_ogg_pad_typefind (GstOggPad * pad, ogg_packet * packet)
 {
+  GstBuffer *buf;
   GstCaps *caps;
   GstElement *element = NULL;
 
@@ -791,11 +815,20 @@ gst_ogg_pad_typefind (GstOggPad * pad, ogg_packet * packet)
   if (GST_PAD_CAPS (pad) != NULL)
     return TRUE;
 
-  caps = gst_ogg_type_find (packet);
+  /* The ogg spec defines that the first packet of an ogg stream must identify
+   * the stream. Therefore ogg can use a simplified approach to typefinding
+   * and only needs to check the first packet */
+  buf = gst_buffer_new ();
+  GST_BUFFER_DATA (buf) = packet->packet;
+  GST_BUFFER_SIZE (buf) = packet->bytes;
+  GST_BUFFER_OFFSET (buf) = 0;
+
+  caps = gst_type_find_helper_for_buffer (GST_OBJECT (pad), buf, NULL);
+  gst_buffer_unref (buf);
 
   if (caps == NULL) {
     GST_WARNING_OBJECT (ogg,
-        "couldn't find caps for stream with serial %08lx", pad->serialno);
+        "couldn't find caps for stream with serial %08x", pad->serialno);
     caps = gst_caps_new_simple ("application/octet-stream", NULL);
   } else {
     /* ogg requires you to use a decoder element to define the
@@ -803,6 +836,8 @@ gst_ogg_pad_typefind (GstOggPad * pad, ogg_packet * packet)
      * we are in the streaming mode to calculate the first timestamp. */
     GList *factories;
 
+    GST_LOG_OBJECT (ogg, "found caps: %" GST_PTR_FORMAT, caps);
+
     /* first filter out the interesting element factories */
     factories = gst_default_registry_feature_filter (
         (GstPluginFeatureFilter) gst_ogg_demux_factory_filter, FALSE, caps);
@@ -874,7 +909,7 @@ gst_ogg_demux_chain_elem_pad (GstOggPad * pad, ogg_packet * packet)
   if (!pad->elem_pad)
     goto no_decoder;
 
-  GST_DEBUG_OBJECT (ogg, "%p init decoder serial %08lx", pad, pad->serialno);
+  GST_DEBUG_OBJECT (ogg, "%p init decoder serial %08x", pad, pad->serialno);
 
   buf = gst_buffer_new_and_alloc (packet->bytes);
   memcpy (GST_BUFFER_DATA (buf), packet->packet, packet->bytes);
@@ -891,7 +926,7 @@ gst_ogg_demux_chain_elem_pad (GstOggPad * pad, ogg_packet * packet)
 no_decoder:
   {
     GST_WARNING_OBJECT (ogg,
-        "pad %08lx does not have elem_pad, no decoder ?", pad);
+        "pad %08x does not have elem_pad, no decoder ?", pad);
     return GST_FLOW_ERROR;
   }
 decoder_error:
@@ -911,7 +946,7 @@ gst_ogg_demux_queue_data (GstOggPad * pad, ogg_packet * packet)
   GstOggDemux *ogg = pad->ogg;
 #endif
 
-  GST_DEBUG_OBJECT (ogg, "%p queueing data serial %08lx", pad, pad->serialno);
+  GST_DEBUG_OBJECT (ogg, "%p queueing data serial %08x", pad, pad->serialno);
 
   buf = gst_buffer_new_and_alloc (packet->bytes);
   memcpy (buf->data, packet->packet, packet->bytes);
@@ -928,18 +963,21 @@ static GstFlowReturn
 gst_ogg_demux_chain_peer (GstOggPad * pad, ogg_packet * packet)
 {
   GstBuffer *buf;
-  GstFlowReturn ret;
+  GstFlowReturn ret, cret;
   GstOggDemux *ogg = pad->ogg;
   GstFormat format;
   gint64 current_time;
   GstOggChain *chain;
 
   GST_DEBUG_OBJECT (ogg,
-      "%p streaming to peer serial %08lx", pad, pad->serialno);
+      "%p streaming to peer serial %08x", pad, pad->serialno);
 
   ret =
-      gst_pad_alloc_buffer_and_set_caps (GST_PAD (pad), GST_BUFFER_OFFSET_NONE,
-      packet->bytes, GST_PAD_CAPS (pad), &buf);
+      gst_pad_alloc_buffer_and_set_caps (GST_PAD_CAST (pad),
+      GST_BUFFER_OFFSET_NONE, packet->bytes, GST_PAD_CAPS (pad), &buf);
+
+  /* combine flows */
+  cret = gst_ogg_demux_combine_flows (ogg, pad, ret);
   if (ret != GST_FLOW_OK)
     goto no_buffer;
 
@@ -949,10 +987,16 @@ gst_ogg_demux_chain_peer (GstOggPad * pad, ogg_packet * packet)
   GST_BUFFER_OFFSET (buf) = -1;
   GST_BUFFER_OFFSET_END (buf) = packet->granulepos;
 
-  ret = gst_pad_push (GST_PAD (pad), buf);
-  /* ignore not linked */
-  if (ret == GST_FLOW_NOT_LINKED)
-    ret = GST_FLOW_OK;
+  /* Mark discont on the buffer */
+  if (pad->discont) {
+    GST_BUFFER_FLAG_SET (buf, GST_BUFFER_FLAG_DISCONT);
+    pad->discont = FALSE;
+  }
+
+  ret = gst_pad_push (GST_PAD_CAST (pad), buf);
+
+  /* combine flows */
+  cret = gst_ogg_demux_combine_flows (ogg, pad, ret);
 
   /* we're done with skeleton stuff */
   if (pad->is_skeleton)
@@ -984,22 +1028,22 @@ gst_ogg_demux_chain_peer (GstOggPad * pad, ogg_packet * packet)
       GST_TIME_ARGS (current_time));
 
 done:
-  return ret;
+  /* return combined flow result */
+  return cret;
 
   /* special cases */
 no_buffer:
   {
     GST_DEBUG_OBJECT (ogg,
-        "%p could not get buffer from peer %08lx, %d (%s)", pad,
-        pad->serialno, ret, gst_flow_get_name (ret));
-    if (ret == GST_FLOW_NOT_LINKED)
-      ret = GST_FLOW_OK;
-    return ret;
+        "%p could not get buffer from peer %08x, %d (%s), combined %d (%s)",
+        pad, pad->serialno, ret, gst_flow_get_name (ret),
+        cret, gst_flow_get_name (cret));
+    goto done;
   }
 convert_failed:
   {
     GST_WARNING_OBJECT (ogg, "could not convert granulepos to time");
-    return ret;
+    goto done;
   }
 }
 
@@ -1015,7 +1059,7 @@ gst_ogg_pad_submit_packet (GstOggPad * pad, ogg_packet * packet)
 
   GstOggDemux *ogg = pad->ogg;
 
-  GST_DEBUG_OBJECT (ogg, "%p submit packet serial %08lx", pad, pad->serialno);
+  GST_DEBUG_OBJECT (ogg, "%p submit packet serial %08x", pad, pad->serialno);
 
   if (!pad->have_type) {
     if (!ogg->have_fishead && packet->bytes == SKELETON_FISHEAD_SIZE &&
@@ -1033,9 +1077,17 @@ gst_ogg_pad_submit_packet (GstOggPad * pad, ogg_packet * packet)
 
   granule = packet->granulepos;
   if (granule != -1) {
+    GST_DEBUG_OBJECT (ogg, "%p has granulepos %" G_GINT64_FORMAT, pad, granule);
     ogg->current_granule = granule;
     pad->current_granule = granule;
+    /* granulepos 0 and -1 are considered header packets.
+     * Note that since theora is busted, it can create non-header
+     * packets with 0 granulepos. We will correct for this when our
+     * internal decoder produced a frame and we don't have a
+     * granulepos because in that case the granulpos must have been 0 */
     if (pad->first_granule == -1 && granule != 0) {
+      GST_DEBUG_OBJECT (ogg, "%p found first granulepos %" G_GINT64_FORMAT, pad,
+          granule);
       pad->first_granule = granule;
     }
   }
@@ -1051,6 +1103,15 @@ gst_ogg_pad_submit_packet (GstOggPad * pad, ogg_packet * packet)
   if (pad->start_time != GST_CLOCK_TIME_NONE) {
     GstOggChain *chain = pad->chain;
 
+    /* correction for busted ogg, if the internal decoder outputed
+     * a timestamp but we did not get a granulepos, it must have
+     * been 0 and the time is therefore also 0 */
+    if (pad->first_granule == -1) {
+      GST_DEBUG_OBJECT (ogg, "%p found start time without granulepos", pad);
+      pad->first_granule = 0;
+      pad->first_time = 0;
+    }
+
     /* check if complete chain has start time */
     if (chain == ogg->building_chain) {
 
@@ -1098,7 +1159,7 @@ gst_ogg_pad_submit_packet (GstOggPad * pad, ogg_packet * packet)
     }
   }
 
-  /* if we are building a chain, store buffer for when we activate 
+  /* if we are building a chain, store buffer for when we activate
    * it. This path is taken if we operate in streaming mode. */
   if (ogg->building_chain) {
     ret = gst_ogg_demux_queue_data (pad, packet);
@@ -1156,7 +1217,7 @@ gst_ogg_pad_submit_page (GstOggPad * pad, ogg_page * page)
 choked:
   {
     GST_WARNING_OBJECT (ogg,
-        "ogg stream choked on page (serial %08lx), resetting stream",
+        "ogg stream choked on page (serial %08x), resetting stream",
         pad->serialno);
     gst_ogg_pad_reset (pad);
     /* we continue to recover */
@@ -1204,6 +1265,18 @@ gst_ogg_chain_free (GstOggChain * chain)
   g_free (chain);
 }
 
+static void
+gst_ogg_chain_mark_discont (GstOggChain * chain)
+{
+  gint i;
+
+  for (i = 0; i < chain->streams->len; i++) {
+    GstOggPad *pad = g_array_index (chain->streams, GstOggPad *, i);
+
+    pad->discont = TRUE;
+  }
+}
+
 static GstOggPad *
 gst_ogg_chain_new_stream (GstOggChain * chain, glong serialno)
 {
@@ -1211,7 +1284,7 @@ gst_ogg_chain_new_stream (GstOggChain * chain, glong serialno)
   GstTagList *list;
   gchar *name;
 
-  GST_DEBUG_OBJECT (chain->ogg, "creating new stream %08lx in chain %p",
+  GST_DEBUG_OBJECT (chain->ogg, "creating new stream %08x in chain %p",
       serialno, chain);
 
   ret = g_object_new (GST_TYPE_OGG_PAD, NULL);
@@ -1225,12 +1298,13 @@ gst_ogg_chain_new_stream (GstOggChain * chain, glong serialno)
   GST_PAD_DIRECTION (ret) = GST_PAD_SRC;
   ret->chain = chain;
   ret->ogg = chain->ogg;
+  ret->discont = TRUE;
   gst_object_set_name (GST_OBJECT (ret), name);
   g_free (name);
 
   ret->serialno = serialno;
   if (ogg_stream_init (&ret->stream, serialno) != 0) {
-    GST_ERROR ("Could not initialize ogg_stream struct for serial %08lx.",
+    GST_ERROR ("Could not initialize ogg_stream struct for serial %08x.",
         serialno);
     gst_object_unref (ret);
     return NULL;
@@ -1242,7 +1316,7 @@ gst_ogg_chain_new_stream (GstOggChain * chain, glong serialno)
   gst_tag_list_free (list);
 
   GST_DEBUG_OBJECT (chain->ogg,
-      "created new ogg src %p for stream with serial %08lx", ret, serialno);
+      "created new ogg src %p for stream with serial %08x", ret, serialno);
 
   g_array_append_val (chain->streams, ret);
 
@@ -1365,6 +1439,7 @@ gst_ogg_demux_init (GstOggDemux * ogg, GstOggDemuxClass * g_class)
   ogg->chain_lock = g_mutex_new ();
   ogg->chains = g_array_new (FALSE, TRUE, sizeof (GstOggChain *));
 
+  ogg->newsegment = NULL;
 }
 
 static void
@@ -1374,9 +1449,13 @@ gst_ogg_demux_finalize (GObject * object)
 
   ogg = GST_OGG_DEMUX (object);
 
+  g_array_free (ogg->chains, TRUE);
   g_mutex_free (ogg->chain_lock);
   ogg_sync_clear (&ogg->sync);
 
+  if (ogg->newsegment)
+    gst_event_unref (ogg->newsegment);
+
   G_OBJECT_CLASS (parent_class)->finalize (object);
 }
 
@@ -1420,6 +1499,8 @@ gst_ogg_demux_submit_buffer (GstOggDemux * ogg, GstBuffer * buffer)
   size = GST_BUFFER_SIZE (buffer);
   data = GST_BUFFER_DATA (buffer);
 
+  GST_DEBUG_OBJECT (ogg, "submitting %u bytes", size);
+
   oggbuffer = ogg_sync_buffer (&ogg->sync, size);
   memcpy (oggbuffer, data, size);
   ogg_sync_wrote (&ogg->sync, size);
@@ -1435,7 +1516,7 @@ gst_ogg_demux_submit_buffer (GstOggDemux * ogg, GstBuffer * buffer)
 static void
 gst_ogg_demux_seek (GstOggDemux * ogg, gint64 offset)
 {
-  GST_LOG_OBJECT (ogg, "seeking to %lld", offset);
+  GST_LOG_OBJECT (ogg, "seeking to %" G_GINT64_FORMAT, offset);
 
   ogg->offset = offset;
   ogg_sync_reset (&ogg->sync);
@@ -1453,7 +1534,8 @@ gst_ogg_demux_get_data (GstOggDemux * ogg)
   GstBuffer *buffer;
   gint size;
 
-  GST_LOG_OBJECT (ogg, "get data %lld %lld", ogg->offset, ogg->length);
+  GST_LOG_OBJECT (ogg, "get data %" G_GINT64_FORMAT " %" G_GINT64_FORMAT,
+      ogg->offset, ogg->length);
   if (ogg->offset == ogg->length)
     goto eos;
 
@@ -1481,13 +1563,19 @@ error:
 }
 
 /* Read the next page from the current offset.
+ * boundary: number of bytes ahead we allow looking for;
+ * -1 if no boundary
+ * returns the offset the next page starts at, or OV_FALSE if we couldn't
+ * find a new page within the boundary
  */
 static gint64
 gst_ogg_demux_get_next_page (GstOggDemux * ogg, ogg_page * og, gint64 boundary)
 {
   gint64 end_offset = 0;
 
-  GST_LOG_OBJECT (ogg, "get next page %lld", boundary);
+  GST_LOG_OBJECT (ogg,
+      "get next page, current offset %" G_GINT64_FORMAT ", bytes boundary %"
+      G_GINT64_FORMAT, ogg->offset, boundary);
 
   if (boundary > 0)
     end_offset = ogg->offset + boundary;
@@ -1496,13 +1584,16 @@ gst_ogg_demux_get_next_page (GstOggDemux * ogg, ogg_page * og, gint64 boundary)
     glong more;
 
     if (boundary > 0 && ogg->offset >= end_offset) {
-      GST_LOG_OBJECT (ogg, "offset %lld >= end_offset %lld", ogg->offset,
-          end_offset);
+      GST_LOG_OBJECT (ogg,
+          "offset %" G_GINT64_FORMAT " >= end_offset %" G_GINT64_FORMAT,
+          ogg->offset, end_offset);
       return OV_FALSE;
     }
 
     more = ogg_sync_pageseek (&ogg->sync, og);
 
+    GST_LOG_OBJECT (ogg, "pageseek gave %ld", more);
+
     if (more < 0) {
       GST_LOG_OBJECT (ogg, "skipped %ld bytes", more);
       /* skipped n bytes */
@@ -1530,7 +1621,8 @@ gst_ogg_demux_get_next_page (GstOggDemux * ogg, ogg_page * og, gint64 boundary)
       ogg_sync_reset (&ogg->sync);
 
       GST_LOG_OBJECT (ogg,
-          "got page at %lld, serial %08lx, end at %lld, granule %lld", ret,
+          "got page at %" G_GINT64_FORMAT ", serial %08x, end at %"
+          G_GINT64_FORMAT ", granule %" G_GINT64_FORMAT, ret,
           ogg_page_serialno (og), ogg->offset, ogg_page_granulepos (og));
 
       return ret;
@@ -1643,6 +1735,10 @@ gst_ogg_demux_activate_chain (GstOggDemux * ogg, GstOggChain * chain,
     pad = g_array_index (chain->streams, GstOggPad *, i);
     GST_DEBUG_OBJECT (ogg, "adding pad %" GST_PTR_FORMAT, pad);
 
+    /* mark discont */
+    pad->discont = TRUE;
+    pad->last_ret = GST_FLOW_OK;
+
     /* activate first */
     gst_pad_set_active (GST_PAD_CAST (pad), TRUE);
 
@@ -1667,6 +1763,11 @@ gst_ogg_demux_activate_chain (GstOggDemux * ogg, GstOggChain * chain,
     for (headers = pad->headers; headers; headers = g_list_next (headers)) {
       GstBuffer *buffer = GST_BUFFER (headers->data);
 
+      if (pad->discont) {
+        GST_BUFFER_FLAG_SET (buffer, GST_BUFFER_FLAG_DISCONT);
+        pad->discont = FALSE;
+      }
+
       /* we don't care about the return value here */
       gst_pad_push (GST_PAD_CAST (pad), buffer);
     }
@@ -1743,6 +1844,7 @@ gst_ogg_demux_do_seek (GstOggDemux * ogg, gint64 position, gboolean accurate,
 
       if (bisect <= begin)
         bisect = begin;
+      GST_DEBUG_OBJECT (ogg, "Initial guess: %" G_GINT64_FORMAT, bisect);
     }
     gst_ogg_demux_seek (ogg, bisect);
 
@@ -1754,6 +1856,8 @@ gst_ogg_demux_do_seek (GstOggDemux * ogg, gint64 position, gboolean accurate,
           ", end %" G_GINT64_FORMAT, bisect, begin, end);
 
       result = gst_ogg_demux_get_next_page (ogg, &og, end - ogg->offset);
+      GST_LOG_OBJECT (ogg, "looking for next page returned %" G_GINT64_FORMAT,
+          result);
       if (result == OV_EREAD) {
         goto seek_error;
       }
@@ -1772,14 +1876,19 @@ gst_ogg_demux_do_seek (GstOggDemux * ogg, gint64 position, gboolean accurate,
           gst_ogg_demux_seek (ogg, bisect);
         }
       } else {
+        /* found offset of next ogg page */
         gint64 granulepos;
         GstClockTime granuletime;
         GstFormat format;
         GstOggPad *pad;
 
+        GST_LOG_OBJECT (ogg, "found next ogg page at %" G_GINT64_FORMAT,
+            result);
         granulepos = ogg_page_granulepos (&og);
-        if (granulepos == -1)
+        if (granulepos == -1) {
+          GST_LOG_OBJECT (ogg, "granulepos of next page is -1");
           continue;
+        }
 
         pad = gst_ogg_chain_get_stream (chain, ogg_page_serialno (&og));
         if (pad == NULL || pad->is_skeleton)
@@ -1792,10 +1901,12 @@ gst_ogg_demux_do_seek (GstOggDemux * ogg, gint64 position, gboolean accurate,
           GST_WARNING_OBJECT (ogg, "could not convert granulepos to time");
           granuletime = target;
         } else {
-          if (granuletime < pad->first_time)
+          if (granuletime < pad->start_time)
             continue;
+          GST_LOG_OBJECT (ogg, "granulepos %" G_GINT64_FORMAT " maps to time %"
+              GST_TIME_FORMAT, granulepos, GST_TIME_ARGS (granuletime));
 
-          granuletime -= pad->first_time;
+          granuletime -= pad->start_time;
         }
 
         GST_DEBUG_OBJECT (ogg,
@@ -1923,17 +2034,16 @@ gst_ogg_demux_perform_seek (GstOggDemux * ogg, GstEvent * event)
 
   if (ogg->segment_running && !flush) {
     /* create the segment event to close the current segment */
-    if ((chain = ogg->current_chain)) { 
-    GstEvent *newseg;
+    if ((chain = ogg->current_chain)) {
+      GstEvent *newseg;
 
-    newseg = gst_event_new_new_segment (TRUE, ogg->segment.rate,
-        GST_FORMAT_TIME, ogg->segment.start + chain->segment_start, 
-	ogg->segment.last_stop + chain->segment_start,
-        ogg->segment.time);
+      newseg = gst_event_new_new_segment (TRUE, ogg->segment.rate,
+          GST_FORMAT_TIME, ogg->segment.start + chain->segment_start,
+          ogg->segment.last_stop + chain->segment_start, ogg->segment.time);
 
       /* send segment on old chain */
-    gst_ogg_demux_send_event (ogg, newseg);
-  }
+      gst_ogg_demux_send_event (ogg, newseg);
+    }
   }
 
   if (event) {
@@ -1968,10 +2078,16 @@ gst_ogg_demux_perform_seek (GstOggDemux * ogg, GstEvent * event)
 
   res = gst_ogg_demux_do_seek (ogg, ogg->segment.last_stop, accurate, &chain);
 
+  /* seek failed, make sure we continue the current chain */
+  if (!res) {
+    chain = ogg->current_chain;
+  }
+
   /* now we have a new position, prepare for streaming again */
   {
     GstEvent *event;
     gint64 stop;
+    gint64 start;
 
     /* we have to send the flush to the old chain, not the new one */
     if (flush)
@@ -1983,19 +2099,24 @@ gst_ogg_demux_perform_seek (GstOggDemux * ogg, GstEvent * event)
     if (stop != -1)
       stop += chain->segment_start;
 
+    start = ogg->segment.last_stop;
+    if (chain->segment_start != GST_CLOCK_TIME_NONE)
+      start += chain->segment_start;
+
     /* create the segment event we are going to send out */
     event = gst_event_new_new_segment (FALSE, ogg->segment.rate,
-        ogg->segment.format, 
-	ogg->segment.last_stop + chain->segment_start, 
-	stop, 
-	ogg->segment.time);
+        ogg->segment.format, start, stop, ogg->segment.time);
 
     if (chain != ogg->current_chain) {
       /* switch to different chain, send segment on new chain */
       gst_ogg_demux_activate_chain (ogg, chain, event);
     } else {
-      /* send segment on current chain */
-      gst_ogg_demux_send_event (ogg, event);
+      /* mark discont and send segment on current chain */
+      gst_ogg_chain_mark_discont (chain);
+      /* This event should be sent from the streaming thread (sink pad task) */
+      if (ogg->newsegment)
+        gst_event_unref (ogg->newsegment);
+      ogg->newsegment = event;
     }
 
     /* notify start of new segment */
@@ -2015,7 +2136,7 @@ gst_ogg_demux_perform_seek (GstOggDemux * ogg, GstEvent * event)
   /* streaming can continue now */
   GST_PAD_STREAM_UNLOCK (ogg->sinkpad);
 
-  return TRUE;
+  return res;
 
 error:
   {
@@ -2040,8 +2161,8 @@ gst_ogg_demux_bisect_forward_serialno (GstOggDemux * ogg,
   GstOggChain *nextchain;
 
   GST_LOG_OBJECT (ogg,
-      "bisect begin: %lld, searched: %lld, end %lld, chain: %p", begin,
-      searched, end, chain);
+      "bisect begin: %" G_GINT64_FORMAT ", searched: %" G_GINT64_FORMAT
+      ", end %" G_GINT64_FORMAT ", chain: %p", begin, searched, end, chain);
 
   /* the below guards against garbage seperating the last and
    * first pages of two links. */
@@ -2075,12 +2196,12 @@ gst_ogg_demux_bisect_forward_serialno (GstOggDemux * ogg,
     }
   }
 
-  GST_LOG_OBJECT (ogg, "current chain ends at %lld", searched);
+  GST_LOG_OBJECT (ogg, "current chain ends at %" G_GINT64_FORMAT, searched);
 
   chain->end_offset = searched;
   gst_ogg_demux_read_end_chain (ogg, chain);
 
-  GST_LOG_OBJECT (ogg, "found begin at %lld", next);
+  GST_LOG_OBJECT (ogg, "found begin at %" G_GINT64_FORMAT, next);
 
   gst_ogg_demux_seek (ogg, next);
   nextchain = gst_ogg_demux_read_chain (ogg);
@@ -2117,7 +2238,7 @@ gst_ogg_demux_read_chain (GstOggDemux * ogg)
   gboolean done;
   gint i;
 
-  GST_LOG_OBJECT (ogg, "reading chain at %lld", offset);
+  GST_LOG_OBJECT (ogg, "reading chain at %" G_GINT64_FORMAT, offset);
 
   /* first read the BOS pages, do typefind on them, create
    * the decoders, send data to the decoders. */
@@ -2136,6 +2257,12 @@ gst_ogg_demux_read_chain (GstOggDemux * ogg)
     }
 
     serial = ogg_page_serialno (&op);
+    if (gst_ogg_chain_get_stream (chain, serial) != NULL) {
+      GST_WARNING_OBJECT (ogg, "found serial %08x BOS page twice, ignoring",
+          serial);
+      continue;
+    }
+
     pad = gst_ogg_chain_new_stream (chain, serial);
     gst_ogg_pad_submit_page (pad, &op);
   }
@@ -2182,7 +2309,7 @@ gst_ogg_demux_read_chain (GstOggDemux * ogg)
       }
       /* the timestamp will be filled in when we submit the pages */
       done &= (pad->start_time != GST_CLOCK_TIME_NONE);
-      GST_LOG_OBJECT (ogg, "done %08lx now %d", serial, done);
+      GST_LOG_OBJECT (ogg, "done %08x now %d", serial, done);
     }
 
     /* we read a page not belonging to the current chain: seek back to the
@@ -2208,11 +2335,17 @@ gst_ogg_demux_read_chain (GstOggDemux * ogg)
     if (pad->is_skeleton)
       continue;
 
+    GST_LOG_OBJECT (ogg, "convert first granule %" G_GUINT64_FORMAT " to time ",
+        pad->first_granule);
+
     target = GST_FORMAT_TIME;
     if (!gst_ogg_pad_query_convert (pad,
             GST_FORMAT_DEFAULT, pad->first_granule, &target,
             (gint64 *) & pad->first_time)) {
       GST_WARNING_OBJECT (ogg, "could not convert granulepos to time");
+    } else {
+      GST_LOG_OBJECT (ogg, "converted to first time %" GST_TIME_FORMAT,
+          GST_TIME_ARGS (pad->first_time));
     }
 
     pad->mode = GST_OGG_PAD_MODE_STREAMING;
@@ -2220,7 +2353,7 @@ gst_ogg_demux_read_chain (GstOggDemux * ogg)
   return chain;
 }
 
-/* read the last pages from the ogg stream to get the final 
+/* read the last pages from the ogg stream to get the final
  * page end_offsets.
  */
 static gint
@@ -2414,7 +2547,7 @@ gst_ogg_demux_find_chains (GstOggDemux * ogg)
   if (!res || ogg->length <= 0)
     goto no_length;
 
-  GST_DEBUG_OBJECT (ogg, "file length %lld", ogg->length);
+  GST_DEBUG_OBJECT (ogg, "file length %" G_GINT64_FORMAT, ogg->length);
 
   /* read chain from offset 0, this is the first chain of the
    * ogg file. */
@@ -2498,7 +2631,8 @@ gst_ogg_demux_chain (GstPad * pad, GstBuffer * buffer)
       granule = ogg_page_granulepos (&page);
 
       GST_LOG_OBJECT (ogg,
-          "processing ogg page (serial %08lx, pageno %ld, granule pos %llu, bos %d)",
+          "processing ogg page (serial %08x, pageno %ld, granulepos %"
+          G_GINT64_FORMAT ", bos %d)",
           serialno, ogg_page_pageno (&page), granule, ogg_page_bos (&page));
 
       if (ogg_page_bos (&page)) {
@@ -2577,7 +2711,7 @@ unknown_chain:
 unknown_pad:
   {
     GST_ELEMENT_ERROR (ogg, STREAM, DECODE,
-        (NULL), ("unknown ogg pad for serial %08d detected", serialno));
+        (NULL), ("unknown ogg pad for serial %08x detected", serialno));
     gst_ogg_demux_send_event (ogg, gst_event_new_eos ());
     return GST_FLOW_ERROR;
   }
@@ -2602,14 +2736,49 @@ gst_ogg_demux_send_event (GstOggDemux * ogg, GstEvent * event)
   gst_event_unref (event);
 }
 
-/* random access code 
+static GstFlowReturn
+gst_ogg_demux_combine_flows (GstOggDemux * ogg, GstOggPad * pad,
+    GstFlowReturn ret)
+{
+  GstOggChain *chain;
+
+  /* store the value */
+  pad->last_ret = ret;
+  /* if it's success we can return the value right away */
+  if (GST_FLOW_IS_SUCCESS (ret))
+    goto done;
+
+  /* any other error that is not-linked can be returned right
+   * away */
+  if (ret != GST_FLOW_NOT_LINKED)
+    goto done;
+
+  /* only return NOT_LINKED if all other pads returned NOT_LINKED */
+  chain = ogg->current_chain;
+  if (chain) {
+    gint i;
+
+    for (i = 0; i < chain->streams->len; i++) {
+      GstOggPad *opad = g_array_index (chain->streams, GstOggPad *, i);
+
+      ret = opad->last_ret;
+      /* some other return value (must be SUCCESS but we can return
+       * other values as well) */
+      if (ret != GST_FLOW_NOT_LINKED)
+        goto done;
+    }
+    /* if we get here, all other pads were unlinked and we return
+     * NOT_LINKED then */
+  }
+done:
+  return ret;
+}
+
+/* random access code
  *
- * - first find all the chains and streams by scanning the 
- *   file.
- * - then get and chain buffers, just like the streaming
- *   case.
- * - when seeking, we can use the chain info to perform the
- *   seek.
+ * - first find all the chains and streams by scanning the file.
+ * - then get and chain buffers, just like the streaming case.
+ * - when seeking, we can use the chain info to perform the seek.
  */
 static void
 gst_ogg_demux_loop (GstOggPad * pad)
@@ -2623,8 +2792,9 @@ gst_ogg_demux_loop (GstOggPad * pad)
 
   if (ogg->need_chains) {
     gboolean got_chains;
+    gboolean res;
 
-    /* this is the only place where we write chains */
+    /* this is the only place where we write chains and thus need to lock. */
     GST_CHAIN_LOCK (ogg);
     ogg->chain_error = GST_FLOW_OK;
     got_chains = gst_ogg_demux_find_chains (ogg);
@@ -2641,15 +2811,21 @@ gst_ogg_demux_loop (GstOggPad * pad)
     GST_OBJECT_UNLOCK (ogg);
 
     /* and seek to configured positions without FLUSH */
-    gst_ogg_demux_perform_seek (ogg, event);
+    res = gst_ogg_demux_perform_seek (ogg, event);
     if (event)
       gst_event_unref (event);
+
+    if (!res)
+      goto seek_failed;
   }
 
-  GST_LOG_OBJECT (ogg, "pull data %lld", ogg->offset);
-  if (ogg->offset == ogg->length)
+  if (ogg->offset == ogg->length) {
+    GST_LOG_OBJECT (ogg, "no more data to pull %" G_GINT64_FORMAT
+        " == %" G_GINT64_FORMAT, ogg->offset, ogg->length);
     goto eos;
+  }
 
+  GST_LOG_OBJECT (ogg, "pull data %" G_GINT64_FORMAT, ogg->offset);
   ret = gst_pad_pull_range (ogg->sinkpad, ogg->offset, CHUNKSIZE, &buffer);
   if (ret != GST_FLOW_OK) {
     GST_LOG_OBJECT (ogg, "Failed pull_range");
@@ -2658,6 +2834,11 @@ gst_ogg_demux_loop (GstOggPad * pad)
 
   ogg->offset += GST_BUFFER_SIZE (buffer);
 
+  if (G_UNLIKELY (ogg->newsegment)) {
+    gst_ogg_demux_send_event (ogg, ogg->newsegment);
+    ogg->newsegment = NULL;
+  }
+
   ret = gst_ogg_demux_chain (ogg->sinkpad, buffer);
   if (ret != GST_FLOW_OK) {
     GST_LOG_OBJECT (ogg, "Failed demux_chain");
@@ -2678,37 +2859,52 @@ chain_read_failed:
     ret = ogg->chain_error;
     goto pause;
   }
+seek_failed:
+  {
+    GST_ELEMENT_ERROR (ogg, STREAM, DEMUX, (NULL),
+        ("failed to start demuxing ogg"));
+    ret = GST_FLOW_ERROR;
+    goto pause;
+  }
 eos:
   {
-    ret = GST_FLOW_OK;
-    /* segment playback just posts a segment end message instead of
-     * pushing out EOS. */
-    ogg->segment_running = FALSE;
-    if (ogg->segment.flags & GST_SEEK_FLAG_SEGMENT) {
-      gint64 stop;
-
-      if ((stop = ogg->segment.stop) == -1)
-        stop = ogg->segment.duration;
-
-      GST_LOG_OBJECT (ogg, "Sending segment done, at end of segment");
-      gst_element_post_message (GST_ELEMENT (ogg),
-          gst_message_new_segment_done (GST_OBJECT (ogg), GST_FORMAT_TIME,
-              stop));
-    } else {
-      GST_LOG_OBJECT (ogg, "Sending EOS, at end of stream");
-      gst_ogg_demux_send_event (ogg, gst_event_new_eos ());
-    }
+    ret = GST_FLOW_UNEXPECTED;
     goto pause;
   }
 pause:
   {
-    GST_LOG_OBJECT (ogg, "pausing task, reason %s", gst_flow_get_name (ret));
+    const gchar *reason = gst_flow_get_name (ret);
+
+    GST_LOG_OBJECT (ogg, "pausing task, reason %s", reason);
+    ogg->segment_running = FALSE;
     gst_pad_pause_task (ogg->sinkpad);
-    if (GST_FLOW_IS_FATAL (ret)) {
-      gst_ogg_demux_send_event (ogg, gst_event_new_eos ());
-      GST_ELEMENT_ERROR (ogg, STREAM, FAILED,
-          (_("Internal data stream error.")),
-          ("stream stopped, reason %s", gst_flow_get_name (ret)));
+
+    if (GST_FLOW_IS_FATAL (ret) || ret == GST_FLOW_NOT_LINKED) {
+      if (ret == GST_FLOW_UNEXPECTED) {
+        /* perform EOS logic */
+        if (ogg->segment.flags & GST_SEEK_FLAG_SEGMENT) {
+          gint64 stop;
+
+          /* for segment playback we need to post when (in stream time)
+           * we stopped, this is either stop (when set) or the duration. */
+          if ((stop = ogg->segment.stop) == -1)
+            stop = ogg->segment.duration;
+
+          GST_LOG_OBJECT (ogg, "Sending segment done, at end of segment");
+          gst_element_post_message (GST_ELEMENT (ogg),
+              gst_message_new_segment_done (GST_OBJECT (ogg), GST_FORMAT_TIME,
+                  stop));
+        } else {
+          /* normal playback, send EOS to all linked pads */
+          GST_LOG_OBJECT (ogg, "Sending EOS, at end of stream");
+          gst_ogg_demux_send_event (ogg, gst_event_new_eos ());
+        }
+      } else {
+        GST_ELEMENT_ERROR (ogg, STREAM, FAILED,
+            (_("Internal data stream error.")),
+            ("stream stopped, reason %s", reason));
+        gst_ogg_demux_send_event (ogg, gst_event_new_eos ());
+      }
     }
     return;
   }
@@ -2855,80 +3051,6 @@ gst_annodex_granule_to_time (gint64 granulepos, gint64 granulerate_n,
   return res;
 }
 
-/*** typefinding **************************************************************/
-/* ogg supports its own typefinding because the ogg spec defines that the first
- * packet of an ogg stream must identify the stream. Therefore ogg can use a
- * simplified approach at typefinding.
- */
-typedef struct
-{
-  ogg_packet *packet;
-  guint best_probability;
-  GstCaps *caps;
-}
-OggTypeFind;
-
-static guint8 *
-ogg_find_peek (gpointer data, gint64 offset, guint size)
-{
-  OggTypeFind *find = (OggTypeFind *) data;
-
-  /* We don't support negative offset (from stream end); nothing embedded in ogg
-   * ever needs them */
-  if (offset >= 0 && offset + size <= find->packet->bytes) {
-    return ((guint8 *) find->packet->packet) + offset;
-  } else {
-    return NULL;
-  }
-}
-static void
-ogg_find_suggest (gpointer data, guint probability, const GstCaps * caps)
-{
-  OggTypeFind *find = (OggTypeFind *) data;
-
-  if (probability > find->best_probability) {
-    GstCaps *copy = gst_caps_copy (caps);
-
-    gst_caps_replace (&find->caps, copy);
-    gst_caps_unref (copy);
-
-    find->best_probability = probability;
-  }
-}
-static GstCaps *
-gst_ogg_type_find (ogg_packet * packet)
-{
-  GstTypeFind gst_find;
-  OggTypeFind find;
-  GList *walk, *type_list = NULL;
-
-  walk = type_list = gst_type_find_factory_get_list ();
-
-  find.packet = packet;
-  find.best_probability = 0;
-  find.caps = NULL;
-  gst_find.data = &find;
-  gst_find.peek = ogg_find_peek;
-  gst_find.suggest = ogg_find_suggest;
-  gst_find.get_length = NULL;
-
-  while (walk) {
-    GstTypeFindFactory *factory = GST_TYPE_FIND_FACTORY (walk->data);
-
-    gst_type_find_factory_call_function (factory, &gst_find);
-    if (find.best_probability >= GST_TYPE_FIND_MAXIMUM)
-      break;
-    walk = g_list_next (walk);
-  }
-
-  gst_plugin_feature_list_free (type_list);
-
-  if (find.best_probability > 0)
-    return find.caps;
-
-  return NULL;
-}
-
 gboolean
 gst_ogg_demux_plugin_init (GstPlugin * plugin)
 {
@@ -2987,7 +3109,7 @@ gst_ogg_print (GstOggDemux * ogg)
     for (j = 0; j < chain->streams->len; j++) {
       GstOggPad *stream = g_array_index (chain->streams, GstOggPad *, j);
 
-      GST_INFO_OBJECT (ogg, "  stream %08lx:", stream->serialno);
+      GST_INFO_OBJECT (ogg, "  stream %08x:", stream->serialno);
       GST_INFO_OBJECT (ogg, "   start time:       %" GST_TIME_FORMAT,
           GST_TIME_ARGS (stream->start_time));
       GST_INFO_OBJECT (ogg, "   first granulepos: %" G_GINT64_FORMAT,
