@@ -58,6 +58,7 @@
 
 #include <gst/gsttagsetter.h>
 #include <gst/tag/tag.h>
+#include <gst/audio/multichannel.h>
 #include "vorbisenc.h"
 
 GST_DEBUG_CATEGORY_EXTERN (vorbisenc_debug);
@@ -67,7 +68,7 @@ static GstPadTemplate *gst_vorbis_enc_src_template,
     *gst_vorbis_enc_sink_template;
 
 /* elementfactory information */
-GstElementDetails vorbisenc_details =
+static const GstElementDetails vorbisenc_details =
 GST_ELEMENT_DETAILS ("Vorbis audio encoder",
     "Codec/Encoder/Audio",
     "Encodes audio in Vorbis format",
@@ -120,6 +121,7 @@ static gboolean gst_vorbis_enc_sink_event (GstPad * pad, GstEvent * event);
 static GstFlowReturn gst_vorbis_enc_chain (GstPad * pad, GstBuffer * buffer);
 static gboolean gst_vorbis_enc_setup (GstVorbisEnc * vorbisenc);
 
+static void gst_vorbis_enc_dispose (GObject * object);
 static void gst_vorbis_enc_get_property (GObject * object, guint prop_id,
     GValue * value, GParamSpec * pspec);
 static void gst_vorbis_enc_set_property (GObject * object, guint prop_id,
@@ -149,12 +151,12 @@ vorbis_caps_factory (void)
 static GstCaps *
 raw_caps_factory (void)
 {
-  /* lowest sample rate is in vorbis/lib/modes/setup_8.h, 8000 Hz
-   * highest sample rate is in vorbis/lib/modes/setup_44.h, 50000 Hz */
+  /* lowest, highest sample rates come from vorbis/lib/modes/setup_X.h: 
+   * 1-200000 Hz */
   return
       gst_caps_new_simple ("audio/x-raw-float",
-      "rate", GST_TYPE_INT_RANGE, 8000, 50000,
-      "channels", GST_TYPE_INT_RANGE, 1, 2,
+      "rate", GST_TYPE_INT_RANGE, 1, 200000,
+      "channels", GST_TYPE_INT_RANGE, 1, 256,
       "endianness", G_TYPE_INT, G_BYTE_ORDER, "width", G_TYPE_INT, 32, NULL);
 }
 
@@ -189,6 +191,7 @@ gst_vorbis_enc_class_init (GstVorbisEncClass * klass)
 
   gobject_class->set_property = gst_vorbis_enc_set_property;
   gobject_class->get_property = gst_vorbis_enc_get_property;
+  gobject_class->dispose = gst_vorbis_enc_dispose;
 
   g_object_class_install_property (G_OBJECT_CLASS (klass), ARG_MAX_BITRATE,
       g_param_spec_int ("max-bitrate", "Maximum Bitrate",
@@ -221,6 +224,113 @@ gst_vorbis_enc_class_init (GstVorbisEncClass * klass)
       GST_DEBUG_FUNCPTR (gst_vorbis_enc_change_state);
 }
 
+static void
+gst_vorbis_enc_dispose (GObject * object)
+{
+  GstVorbisEnc *vorbisenc = GST_VORBISENC (object);
+
+  if (vorbisenc->sinkcaps) {
+    gst_caps_unref (vorbisenc->sinkcaps);
+    vorbisenc->sinkcaps = NULL;
+  }
+
+  G_OBJECT_CLASS (parent_class)->dispose (object);
+}
+
+static GstAudioChannelPosition vorbischannelpositions[][6] = {
+  {                             /* Mono */
+      GST_AUDIO_CHANNEL_POSITION_FRONT_MONO},
+  {                             /* Stereo */
+        GST_AUDIO_CHANNEL_POSITION_FRONT_LEFT,
+      GST_AUDIO_CHANNEL_POSITION_FRONT_RIGHT},
+  {                             /* Stereo + Centre */
+        GST_AUDIO_CHANNEL_POSITION_FRONT_LEFT,
+        GST_AUDIO_CHANNEL_POSITION_FRONT_CENTER,
+      GST_AUDIO_CHANNEL_POSITION_FRONT_RIGHT},
+  {                             /* Quadraphonic */
+        GST_AUDIO_CHANNEL_POSITION_FRONT_LEFT,
+        GST_AUDIO_CHANNEL_POSITION_FRONT_RIGHT,
+        GST_AUDIO_CHANNEL_POSITION_REAR_LEFT,
+        GST_AUDIO_CHANNEL_POSITION_REAR_RIGHT,
+      },
+  {                             /* Stereo + Centre + rear stereo */
+        GST_AUDIO_CHANNEL_POSITION_FRONT_LEFT,
+        GST_AUDIO_CHANNEL_POSITION_FRONT_CENTER,
+        GST_AUDIO_CHANNEL_POSITION_FRONT_RIGHT,
+        GST_AUDIO_CHANNEL_POSITION_REAR_LEFT,
+        GST_AUDIO_CHANNEL_POSITION_REAR_RIGHT,
+      },
+  {                             /* Full 5.1 Surround */
+        GST_AUDIO_CHANNEL_POSITION_FRONT_LEFT,
+        GST_AUDIO_CHANNEL_POSITION_FRONT_CENTER,
+        GST_AUDIO_CHANNEL_POSITION_FRONT_RIGHT,
+        GST_AUDIO_CHANNEL_POSITION_REAR_LEFT,
+        GST_AUDIO_CHANNEL_POSITION_REAR_RIGHT,
+        GST_AUDIO_CHANNEL_POSITION_LFE,
+      },
+};
+static GstCaps *
+gst_vorbis_enc_generate_sink_caps (void)
+{
+  GstCaps *caps = gst_caps_new_empty ();
+  int i, c;
+
+  gst_caps_append_structure (caps, gst_structure_new ("audio/x-raw-float",
+          "rate", GST_TYPE_INT_RANGE, 1, 200000,
+          "channels", G_TYPE_INT, 1,
+          "endianness", G_TYPE_INT, G_BYTE_ORDER, "width", G_TYPE_INT, 32,
+          NULL));
+
+  gst_caps_append_structure (caps, gst_structure_new ("audio/x-raw-float",
+          "rate", GST_TYPE_INT_RANGE, 1, 200000,
+          "channels", G_TYPE_INT, 2,
+          "endianness", G_TYPE_INT, G_BYTE_ORDER, "width", G_TYPE_INT, 32,
+          NULL));
+
+  for (i = 3; i <= 6; i++) {
+    GValue chanpos = { 0 };
+    GValue pos = { 0 };
+    GstStructure *structure;
+
+    g_value_init (&chanpos, GST_TYPE_ARRAY);
+    g_value_init (&pos, GST_TYPE_AUDIO_CHANNEL_POSITION);
+
+    for (c = 0; c < i; c++) {
+      g_value_set_enum (&pos, vorbischannelpositions[i - 1][c]);
+      gst_value_array_append_value (&chanpos, &pos);
+    }
+    g_value_unset (&pos);
+
+    structure = gst_structure_new ("audio/x-raw-float",
+        "rate", GST_TYPE_INT_RANGE, 1, 200000,
+        "channels", G_TYPE_INT, i,
+        "endianness", G_TYPE_INT, G_BYTE_ORDER, "width", G_TYPE_INT, 32, NULL);
+    gst_structure_set_value (structure, "channel-positions", &chanpos);
+    g_value_unset (&chanpos);
+
+    gst_caps_append_structure (caps, structure);
+  }
+
+  gst_caps_append_structure (caps, gst_structure_new ("audio/x-raw-float",
+          "rate", GST_TYPE_INT_RANGE, 1, 200000,
+          "channels", GST_TYPE_INT_RANGE, 7, 256,
+          "endianness", G_TYPE_INT, G_BYTE_ORDER, "width", G_TYPE_INT, 32,
+          NULL));
+
+  return caps;
+}
+
+static GstCaps *
+gst_vorbis_enc_sink_getcaps (GstPad * pad)
+{
+  GstVorbisEnc *vorbisenc = GST_VORBISENC (GST_PAD_PARENT (pad));
+
+  if (vorbisenc->sinkcaps == NULL)
+    vorbisenc->sinkcaps = gst_vorbis_enc_generate_sink_caps ();
+
+  return gst_caps_ref (vorbisenc->sinkcaps);
+}
+
 static gboolean
 gst_vorbis_enc_sink_setcaps (GstPad * pad, GstCaps * caps)
 {
@@ -488,6 +598,8 @@ gst_vorbis_enc_init (GstVorbisEnc * vorbisenc, GstVorbisEncClass * klass)
       GST_DEBUG_FUNCPTR (gst_vorbis_enc_chain));
   gst_pad_set_setcaps_function (vorbisenc->sinkpad,
       GST_DEBUG_FUNCPTR (gst_vorbis_enc_sink_setcaps));
+  gst_pad_set_getcaps_function (vorbisenc->sinkpad,
+      GST_DEBUG_FUNCPTR (gst_vorbis_enc_sink_getcaps));
   gst_pad_set_query_function (vorbisenc->sinkpad,
       GST_DEBUG_FUNCPTR (gst_vorbis_enc_sink_query));
   gst_element_add_pad (GST_ELEMENT (vorbisenc), vorbisenc->sinkpad);
@@ -512,92 +624,52 @@ gst_vorbis_enc_init (GstVorbisEnc * vorbisenc, GstVorbisEncClass * klass)
   vorbisenc->last_message = NULL;
 }
 
-
-static gchar *
-gst_vorbis_enc_get_tag_value (const GstTagList * list, const gchar * tag,
-    int index)
-{
-  GType tag_type;
-  gchar *vorbisvalue = NULL;
-
-  if (tag == NULL)
-    return NULL;
-
-  tag_type = gst_tag_get_type (tag);
-
-  /* get tag name right */
-  if ((strcmp (tag, GST_TAG_TRACK_NUMBER) == 0)
-      || (strcmp (tag, GST_TAG_ALBUM_VOLUME_NUMBER) == 0)
-      || (strcmp (tag, GST_TAG_TRACK_COUNT) == 0)
-      || (strcmp (tag, GST_TAG_ALBUM_VOLUME_COUNT) == 0)) {
-    guint track_no;
-
-    if (!gst_tag_list_get_uint_index (list, tag, index, &track_no))
-      g_return_val_if_reached (NULL);
-
-    vorbisvalue = g_strdup_printf ("%u", track_no);
-  } else if (tag_type == GST_TYPE_DATE) {
-    GDate *date;
-
-    if (!gst_tag_list_get_date_index (list, tag, index, &date))
-      g_return_val_if_reached (NULL);
-
-    vorbisvalue =
-        g_strdup_printf ("%04d-%02d-%02d", (gint) g_date_get_year (date),
-        (gint) g_date_get_month (date), (gint) g_date_get_day (date));
-    g_date_free (date);
-  } else if (tag_type == G_TYPE_STRING) {
-    if (!gst_tag_list_get_string_index (list, tag, index, &vorbisvalue))
-      g_return_val_if_reached (NULL);
-  }
-
-  return vorbisvalue;
-}
-
 static void
 gst_vorbis_enc_metadata_set1 (const GstTagList * list, const gchar * tag,
     gpointer vorbisenc)
 {
-  const gchar *vorbistag = NULL;
-  gchar *vorbisvalue = NULL;
-  guint i, count;
   GstVorbisEnc *enc = GST_VORBISENC (vorbisenc);
+  GList *vc_list, *l;
 
-  vorbistag = gst_tag_to_vorbis_tag (tag);
-  if (vorbistag == NULL) {
-    return;
-  }
+  vc_list = gst_tag_to_vorbis_comments (list, tag);
 
-  count = gst_tag_list_get_tag_size (list, tag);
-  for (i = 0; i < count; i++) {
-    vorbisvalue = gst_vorbis_enc_get_tag_value (list, tag, i);
+  for (l = vc_list; l != NULL; l = l->next) {
+    const gchar *vc_string = (const gchar *) l->data;
+    gchar *key = NULL, *val = NULL;
 
-    if (vorbisvalue != NULL) {
-      gchar *tmptag = g_strdup (vorbistag);
-
-      vorbis_comment_add_tag (&enc->vc, tmptag, vorbisvalue);
-      g_free (tmptag);
-      g_free (vorbisvalue);
+    if (gst_tag_parse_extended_comment (vc_string, &key, NULL, &val, TRUE)) {
+      vorbis_comment_add_tag (&enc->vc, key, val);
+      g_free (key);
+      g_free (val);
     }
   }
+
+  g_list_foreach (vc_list, (GFunc) g_free, NULL);
+  g_list_free (vc_list);
 }
 
 static void
-gst_vorbis_enc_set_metadata (GstVorbisEnc * vorbisenc)
+gst_vorbis_enc_set_metadata (GstVorbisEnc * enc)
 {
-  GstTagList *copy;
+  GstTagList *merged_tags;
   const GstTagList *user_tags;
 
-  user_tags = gst_tag_setter_get_tag_list (GST_TAG_SETTER (vorbisenc));
-  if (!(vorbisenc->tags || user_tags))
-    return;
+  vorbis_comment_init (&enc->vc);
 
-  copy =
-      gst_tag_list_merge (user_tags, vorbisenc->tags,
-      gst_tag_setter_get_tag_merge_mode (GST_TAG_SETTER (vorbisenc)));
-  vorbis_comment_init (&vorbisenc->vc);
-  gst_tag_list_foreach (copy, gst_vorbis_enc_metadata_set1, vorbisenc);
-  gst_tag_list_free (copy);
+  user_tags = gst_tag_setter_get_tag_list (GST_TAG_SETTER (enc));
+
+  GST_DEBUG_OBJECT (enc, "upstream tags = %" GST_PTR_FORMAT, enc->tags);
+  GST_DEBUG_OBJECT (enc, "user-set tags = %" GST_PTR_FORMAT, user_tags);
+
+  /* gst_tag_list_merge() will handle NULL for either or both lists fine */
+  merged_tags = gst_tag_list_merge (user_tags, enc->tags,
+      gst_tag_setter_get_tag_merge_mode (GST_TAG_SETTER (enc)));
+
+  if (merged_tags) {
+    GST_DEBUG_OBJECT (enc, "merged   tags = %" GST_PTR_FORMAT, merged_tags);
+    gst_tag_list_foreach (merged_tags, gst_vorbis_enc_metadata_set1, enc);
+    gst_tag_list_free (merged_tags);
+  }
 }
 
 static gchar *
@@ -787,6 +859,11 @@ gst_vorbis_enc_buffer_from_packet (GstVorbisEnc * vorbisenc,
   GST_BUFFER_DURATION (outbuf) =
       vorbisenc->next_ts - GST_BUFFER_TIMESTAMP (outbuf);
 
+  if (vorbisenc->next_discont) {
+    GST_BUFFER_FLAG_SET (outbuf, GST_BUFFER_FLAG_DISCONT);
+    vorbisenc->next_discont = FALSE;
+  }
+
   GST_LOG_OBJECT (vorbisenc, "encoded buffer of %d bytes",
       GST_BUFFER_SIZE (outbuf));
   return outbuf;
@@ -807,6 +884,8 @@ gst_vorbis_enc_buffer_from_header_packet (GstVorbisEnc * vorbisenc,
   GST_BUFFER_TIMESTAMP (outbuf) = GST_CLOCK_TIME_NONE;
   GST_BUFFER_DURATION (outbuf) = GST_CLOCK_TIME_NONE;
 
+  gst_buffer_set_caps (outbuf, vorbisenc->srccaps);
+
   GST_DEBUG ("created header packet buffer, %d bytes",
       GST_BUFFER_SIZE (outbuf));
   return outbuf;
@@ -818,6 +897,8 @@ gst_vorbis_enc_push_buffer (GstVorbisEnc * vorbisenc, GstBuffer * buffer)
 {
   vorbisenc->bytes_out += GST_BUFFER_SIZE (buffer);
 
+  GST_DEBUG_OBJECT (vorbisenc, "Pushing buffer with GP %lld, ts %lld",
+      GST_BUFFER_OFFSET_END (buffer), GST_BUFFER_TIMESTAMP (buffer));
   return gst_pad_push (vorbisenc->srcpad, buffer);
 }
 
@@ -913,6 +994,41 @@ gst_vorbis_enc_sink_event (GstPad * pad, GstEvent * event)
   return res;
 }
 
+static gboolean
+gst_vorbis_enc_buffer_check_discontinuous (GstVorbisEnc * vorbisenc,
+    GstBuffer * buffer)
+{
+  gboolean ret = FALSE;
+
+  if (GST_BUFFER_TIMESTAMP (buffer) != GST_CLOCK_TIME_NONE &&
+      vorbisenc->expected_ts != GST_CLOCK_TIME_NONE &&
+      GST_BUFFER_TIMESTAMP (buffer) != vorbisenc->expected_ts) {
+    /* It turns out that a lot of elements don't generate perfect streams due
+     * to rounding errors. So, we permit small errors (< 1/2 a sample) without
+     * causing a discont.
+     */
+    int halfsample = GST_SECOND / vorbisenc->frequency / 2;
+
+    if ((GstClockTimeDiff) (GST_BUFFER_TIMESTAMP (buffer) -
+            vorbisenc->expected_ts) > halfsample) {
+      GST_DEBUG_OBJECT (vorbisenc, "Expected TS % " GST_TIME_FORMAT
+          ", buffer TS % " GST_TIME_FORMAT,
+          GST_TIME_ARGS (vorbisenc->expected_ts),
+          GST_TIME_ARGS (GST_BUFFER_TIMESTAMP (buffer)));
+      ret = TRUE;
+    }
+  }
+
+  if (GST_BUFFER_TIMESTAMP (buffer) != GST_CLOCK_TIME_NONE &&
+      GST_BUFFER_DURATION (buffer) != GST_CLOCK_TIME_NONE) {
+    vorbisenc->expected_ts = GST_BUFFER_TIMESTAMP (buffer) +
+        GST_BUFFER_DURATION (buffer);
+  } else
+    vorbisenc->expected_ts = GST_CLOCK_TIME_NONE;
+
+  return ret;
+}
+
 static GstFlowReturn
 gst_vorbis_enc_chain (GstPad * pad, GstBuffer * buffer)
 {
@@ -923,6 +1039,7 @@ gst_vorbis_enc_chain (GstPad * pad, GstBuffer * buffer)
   gulong i, j;
   float **vorbis_buffer;
   GstBuffer *buf1, *buf2, *buf3;
+  gboolean first = FALSE;
 
   vorbisenc = GST_VORBISENC (GST_PAD_PARENT (pad));
 
@@ -958,7 +1075,8 @@ gst_vorbis_enc_chain (GstPad * pad, GstBuffer * buffer)
     buf3 = gst_vorbis_enc_buffer_from_header_packet (vorbisenc, &header_code);
 
     /* mark and put on caps */
-    caps = gst_pad_get_caps (vorbisenc->srcpad);
+    vorbisenc->srccaps = gst_caps_new_simple ("audio/x-vorbis", NULL);
+    caps = vorbisenc->srccaps;
     caps = gst_vorbis_enc_set_header_on_caps (caps, buf1, buf2, buf3);
 
     /* negotiate with these caps */
@@ -969,23 +1087,23 @@ gst_vorbis_enc_chain (GstPad * pad, GstBuffer * buffer)
     gst_buffer_set_caps (buf2, caps);
     gst_buffer_set_caps (buf3, caps);
 
-    gst_caps_unref (caps);
-
     /* push out buffers */
+    /* push_buffer takes the reference even for failure */
     if ((ret = gst_vorbis_enc_push_buffer (vorbisenc, buf1)) != GST_FLOW_OK)
       goto failed_header_push;
-    buf1 = NULL;
-    if ((ret = gst_vorbis_enc_push_buffer (vorbisenc, buf2)) != GST_FLOW_OK)
+    if ((ret = gst_vorbis_enc_push_buffer (vorbisenc, buf2)) != GST_FLOW_OK) {
+      buf2 = NULL;
       goto failed_header_push;
-    buf2 = NULL;
-    if ((ret = gst_vorbis_enc_push_buffer (vorbisenc, buf3)) != GST_FLOW_OK)
+    }
+    if ((ret = gst_vorbis_enc_push_buffer (vorbisenc, buf3)) != GST_FLOW_OK) {
+      buf3 = NULL;
       goto failed_header_push;
-    buf3 = NULL;
-
+    }
 
     /* now adjust starting granulepos accordingly if the buffer's timestamp is
        nonzero */
     vorbisenc->next_ts = GST_BUFFER_TIMESTAMP (buffer);
+    vorbisenc->expected_ts = GST_BUFFER_TIMESTAMP (buffer);
     vorbisenc->granulepos_offset = gst_util_uint64_scale
         (GST_BUFFER_TIMESTAMP (buffer), vorbisenc->frequency, GST_SECOND);
     vorbisenc->subgranule_offset = 0;
@@ -993,6 +1111,41 @@ gst_vorbis_enc_chain (GstPad * pad, GstBuffer * buffer)
         vorbisenc->next_ts - granulepos_to_timestamp_offset (vorbisenc, 0);
 
     vorbisenc->header_sent = TRUE;
+    first = TRUE;
+  }
+
+  if (vorbisenc->expected_ts != GST_CLOCK_TIME_NONE &&
+      GST_BUFFER_TIMESTAMP (buffer) < vorbisenc->expected_ts) {
+    GST_WARNING_OBJECT (vorbisenc, "Buffer is older than previous "
+        "timestamp + duration, cannot handle. Dropping buffer.");
+    gst_buffer_unref (buffer);
+    return GST_FLOW_OK;
+  }
+
+  if (gst_vorbis_enc_buffer_check_discontinuous (vorbisenc, buffer) && !first) {
+    GST_WARNING_OBJECT (vorbisenc, "Buffer is discontinuous, flushing encoder "
+        "and restarting (Discont from %" GST_TIME_FORMAT
+        " to %" GST_TIME_FORMAT ")", GST_TIME_ARGS (vorbisenc->next_ts),
+        GST_TIME_ARGS (GST_BUFFER_TIMESTAMP (buffer)));
+    /* Re-initialise encoder (there's unfortunately no API to flush it) */
+    if ((ret = gst_vorbis_enc_clear (vorbisenc)) != GST_FLOW_OK)
+      return ret;
+    if (!gst_vorbis_enc_setup (vorbisenc))
+      return GST_FLOW_ERROR;    /* Should be impossible, we can only get here if
+                                   we successfully initialised earlier */
+
+    /* Now, set our granulepos offset appropriately. */
+    vorbisenc->next_ts = GST_BUFFER_TIMESTAMP (buffer);
+    /* We need to round to the nearest whole number of samples, not just do
+     * a truncating division here */
+    vorbisenc->granulepos_offset = gst_util_uint64_scale
+        (GST_BUFFER_TIMESTAMP (buffer) + GST_SECOND / vorbisenc->frequency / 2
+        - vorbisenc->subgranule_offset, vorbisenc->frequency, GST_SECOND);
+
+    vorbisenc->header_sent = TRUE;
+
+    /* And our next output buffer must have DISCONT set on it */
+    vorbisenc->next_discont = TRUE;
   }
 
   /* data to encode */
@@ -1031,8 +1184,7 @@ not_setup:
 failed_header_push:
   {
     GST_WARNING_OBJECT (vorbisenc, "Failed to push headers");
-    if (buf1)
-      gst_buffer_unref (buf1);
+    /* buf1 is always already unreffed */
     if (buf2)
       gst_buffer_unref (buf2);
     if (buf3)
@@ -1191,6 +1343,7 @@ gst_vorbis_enc_change_state (GstElement * element, GstStateChange transition)
       break;
     case GST_STATE_CHANGE_READY_TO_PAUSED:
       vorbisenc->setup = FALSE;
+      vorbisenc->next_discont = FALSE;
       vorbisenc->header_sent = FALSE;
       break;
     case GST_STATE_CHANGE_PAUSED_TO_PLAYING:
@@ -1210,6 +1363,10 @@ gst_vorbis_enc_change_state (GstElement * element, GstStateChange transition)
       vorbis_info_clear (&vorbisenc->vi);
       g_free (vorbisenc->last_message);
       vorbisenc->last_message = NULL;
+      if (vorbisenc->srccaps) {
+        gst_caps_unref (vorbisenc->srccaps);
+        vorbisenc->srccaps = NULL;
+      }
       break;
     case GST_STATE_CHANGE_READY_TO_NULL:
       gst_tag_list_free (vorbisenc->tags);
