@@ -34,13 +34,18 @@ enum
   LAST_SIGNAL
 };
 
-/* we tollerate a 10th of a second diff before we start resyncing. This
+/* we tollerate half a second diff before we start resyncing. This
  * should be enough to compensate for various rounding errors in the timestamp
- * and sample offset position. */
-#define DIFF_TOLERANCE  10
+ * and sample offset position. 
+ * This is an emergency resync fallback since buffers marked as DISCONT will
+ * always lock to the correct timestamp immediatly and buffers not marked as
+ * DISCONT are contiguous by definition.
+ */
+#define DIFF_TOLERANCE  2
 
-#define DEFAULT_BUFFER_TIME     200 * GST_USECOND
-#define DEFAULT_LATENCY_TIME    10 * GST_USECOND
+/* FIXME: 0.11, store the buffer_time and latency_time in nanoseconds */
+#define DEFAULT_BUFFER_TIME     ((200 * GST_MSECOND) / GST_USECOND)
+#define DEFAULT_LATENCY_TIME    ((10 * GST_MSECOND) / GST_USECOND)
 #define DEFAULT_PROVIDE_CLOCK   TRUE
 
 enum
@@ -96,8 +101,6 @@ gst_base_audio_sink_base_init (gpointer g_class)
 static void
 gst_base_audio_sink_class_init (GstBaseAudioSinkClass * klass)
 {
-  gchar *longdesc;
-
   GObjectClass *gobject_class;
   GstElementClass *gstelement_class;
   GstBaseSinkClass *gstbasesink_class;
@@ -112,22 +115,17 @@ gst_base_audio_sink_class_init (GstBaseAudioSinkClass * klass)
       GST_DEBUG_FUNCPTR (gst_base_audio_sink_get_property);
   gobject_class->dispose = GST_DEBUG_FUNCPTR (gst_base_audio_sink_dispose);
 
-  longdesc =
-      g_strdup_printf
-      ("Size of audio buffer in microseconds (use -1 for default of %"
-      G_GUINT64_FORMAT " us)", DEFAULT_BUFFER_TIME / GST_USECOND);
-  g_object_class_install_property (G_OBJECT_CLASS (klass), PROP_BUFFER_TIME,
-      g_param_spec_int64 ("buffer-time", "Buffer Time", longdesc, -1,
+  g_object_class_install_property (gobject_class, PROP_BUFFER_TIME,
+      g_param_spec_int64 ("buffer-time", "Buffer Time",
+          "Size of audio buffer in microseconds", 1,
           G_MAXINT64, DEFAULT_BUFFER_TIME, G_PARAM_READWRITE));
-  g_free (longdesc);
-  longdesc =
-      g_strdup_printf ("Audio latency in microseconds (use -1 for default of %"
-      G_GUINT64_FORMAT " us)", DEFAULT_LATENCY_TIME / GST_USECOND);
-  g_object_class_install_property (G_OBJECT_CLASS (klass), PROP_LATENCY_TIME,
-      g_param_spec_int64 ("latency-time", "Latency Time", longdesc, -1,
+
+  g_object_class_install_property (gobject_class, PROP_LATENCY_TIME,
+      g_param_spec_int64 ("latency-time", "Latency Time",
+          "Audio latency in microseconds", 1,
           G_MAXINT64, DEFAULT_LATENCY_TIME, G_PARAM_READWRITE));
-  g_free (longdesc);
-  g_object_class_install_property (G_OBJECT_CLASS (klass), PROP_PROVIDE_CLOCK,
+
+  g_object_class_install_property (gobject_class, PROP_PROVIDE_CLOCK,
       g_param_spec_boolean ("provide-clock", "Provide Clock",
           "Provide a clock to be used as the global pipeline clock",
           DEFAULT_PROVIDE_CLOCK, G_PARAM_READWRITE));
@@ -186,7 +184,7 @@ gst_base_audio_sink_provide_clock (GstElement * elem)
 
   sink = GST_BASE_AUDIO_SINK (elem);
 
-  /* we have no ringbuffer (must be NULL state */
+  /* we have no ringbuffer (must be NULL state) */
   if (sink->ringbuffer == NULL)
     goto wrong_state;
 
@@ -202,6 +200,7 @@ gst_base_audio_sink_provide_clock (GstElement * elem)
 
   return clock;
 
+  /* ERRORS */
 wrong_state:
   {
     GST_DEBUG_OBJECT (sink, "ringbuffer not acquired");
@@ -314,12 +313,12 @@ gst_base_audio_sink_setcaps (GstBaseSink * bsink, GstCaps * caps)
   if (!gst_ring_buffer_acquire (sink->ringbuffer, spec))
     goto acquire_error;
 
-  /* calculate actual latency and buffer times */
-  spec->latency_time =
-      spec->segsize * GST_MSECOND / (spec->rate * spec->bytes_per_sample);
-  spec->buffer_time =
-      spec->segtotal * spec->segsize * GST_MSECOND / (spec->rate *
-      spec->bytes_per_sample);
+  /* calculate actual latency and buffer times. 
+   * FIXME: In 0.11, store the latency_time internally in ns */
+  spec->latency_time = gst_util_uint64_scale (spec->segsize,
+      (GST_SECOND / GST_USECOND), spec->rate * spec->bytes_per_sample);
+
+  spec->buffer_time = spec->segtotal * spec->latency_time;
 
   gst_ring_buffer_debug_spec_buff (spec);
 
@@ -361,6 +360,12 @@ gst_base_audio_sink_drain (GstBaseAudioSink * sink)
   if (!sink->ringbuffer->spec.rate)
     return TRUE;
 
+  /* need to start playback before we can drain, but only when
+   * we have successfully negotiated a format and thus aqcuired the
+   * ringbuffer. */
+  if (gst_ring_buffer_is_acquired (sink->ringbuffer))
+    gst_ring_buffer_start (sink->ringbuffer);
+
   if (sink->next_sample != -1) {
     GstClockTime time;
     GstClock *clock;
@@ -394,18 +399,16 @@ gst_base_audio_sink_event (GstBaseSink * bsink, GstEvent * event)
 
   switch (GST_EVENT_TYPE (event)) {
     case GST_EVENT_FLUSH_START:
-      gst_ring_buffer_set_flushing (sink->ringbuffer, TRUE);
+      if (sink->ringbuffer)
+        gst_ring_buffer_set_flushing (sink->ringbuffer, TRUE);
       break;
     case GST_EVENT_FLUSH_STOP:
       /* always resync on sample after a flush */
       sink->next_sample = -1;
-      gst_ring_buffer_set_flushing (sink->ringbuffer, FALSE);
+      if (sink->ringbuffer)
+        gst_ring_buffer_set_flushing (sink->ringbuffer, FALSE);
       break;
     case GST_EVENT_EOS:
-      /* need to start playback when we reach EOS, but only when
-       * we have successfully negotiated a format. */
-      if (gst_ring_buffer_is_acquired (sink->ringbuffer))
-        gst_ring_buffer_start (sink->ringbuffer);
       /* now wait till we played everything */
       gst_base_audio_sink_drain (sink);
       break;
@@ -488,11 +491,6 @@ gst_base_audio_sink_render (GstBaseSink * bsink, GstBuffer * buf)
 
   sink = GST_BASE_AUDIO_SINK (bsink);
 
-  if (G_UNLIKELY (GST_BUFFER_FLAG_IS_SET (buf, GST_BUFFER_FLAG_DISCONT))) {
-    /* always resync after a discont */
-    sink->next_sample = -1;
-  }
-
   ringbuf = sink->ringbuffer;
 
   /* can't do anything when we don't have the device */
@@ -579,27 +577,49 @@ gst_base_audio_sink_render (GstBaseSink * bsink, GstBuffer * buf)
       ", render offset %llu, samples %lu",
       GST_TIME_ARGS (render_time), render_offset, samples);
 
-  /* roundoff errors in timestamp conversion */
-  if (G_LIKELY (sink->next_sample != -1)) {
-    diff = ABS ((gint64) render_offset - (gint64) sink->next_sample);
 
-    /* we tollerate a 10th of a second diff before we start resyncing. This
-     * should be enough to compensate for various rounding errors in the timestamp
-     * and sample offset position. We always resync if we got a discont anyway. */
-    if (diff < ringbuf->spec.rate / DIFF_TOLERANCE) {
-      GST_DEBUG_OBJECT (sink,
-          "align with prev sample, %" G_GINT64_FORMAT " < %lu", diff,
-          ringbuf->spec.rate / DIFF_TOLERANCE);
-      /* just align with previous sample then */
-      render_offset = sink->next_sample;
-    } else {
-      GST_DEBUG_OBJECT (sink,
-          "resync after discont with previous sample of diff: %lu", diff);
-    }
-  } else {
-    GST_DEBUG_OBJECT (sink, "resync after discont");
+  /* never try to align samples when we are slaved to another clock, just
+   * trust the rate control algorithm to align the two clocks. We don't take
+   * the LOCK to read the clock because it does not really matter here and the
+   * clock is not changed while playing normally. */
+  if (GST_ELEMENT_CLOCK (sink) != sink->provided_clock) {
+    GST_DEBUG_OBJECT (sink, "no align needed: we are slaved");
+    goto no_align;
   }
 
+  /* always resync after a discont */
+  if (G_UNLIKELY (GST_BUFFER_FLAG_IS_SET (buf, GST_BUFFER_FLAG_DISCONT))) {
+    GST_DEBUG_OBJECT (sink, "resync after discont");
+    goto no_align;
+  }
+
+  if (G_UNLIKELY (sink->next_sample == -1)) {
+    GST_DEBUG_OBJECT (sink,
+        "no align possible: no previous sample position known");
+    goto no_align;
+  }
+
+  /* now try to align the sample to the previous one */
+  diff = ABS ((gint64) render_offset - (gint64) sink->next_sample);
+
+  /* we tollerate half a second diff before we start resyncing. This
+   * should be enough to compensate for various rounding errors in the timestamp
+   * and sample offset position. We always resync if we got a discont anyway and
+   * non-discont should be aligned by definition. */
+  if (G_LIKELY (diff < ringbuf->spec.rate / DIFF_TOLERANCE)) {
+    GST_DEBUG_OBJECT (sink,
+        "align with prev sample, %" G_GINT64_FORMAT " < %lu", diff,
+        ringbuf->spec.rate / DIFF_TOLERANCE);
+    /* just align with previous sample then */
+    render_offset = sink->next_sample;
+  } else {
+    /* timestamps drifted apart from previous samples too much, we need to
+     * resync. We log this as an element warning. */
+    GST_ELEMENT_WARNING (sink, CORE, CLOCK, (NULL),
+        ("Unexpected discontinuity in audio timestamps of more than half a second"));
+  }
+
+no_align:
   crate =
       gst_guint64_to_gdouble (crate_num) / gst_guint64_to_gdouble (crate_denom);
   GST_DEBUG_OBJECT (sink,
@@ -787,6 +807,8 @@ gst_base_audio_sink_change_state (GstElement * element,
       GST_OBJECT_UNLOCK (sink);
       break;
     case GST_STATE_CHANGE_PAUSED_TO_READY:
+      /* make sure we unblock before calling the parent state change
+       * so it can grab the STREAM_LOCK */
       gst_ring_buffer_set_flushing (sink->ringbuffer, TRUE);
       break;
     default:
@@ -813,8 +835,10 @@ gst_base_audio_sink_change_state (GstElement * element,
 
   return ret;
 
+  /* ERRORS */
 open_failed:
   {
+    /* subclass must post a meaningfull error message */
     GST_DEBUG_OBJECT (sink, "open failed");
     return GST_STATE_CHANGE_FAILURE;
   }
