@@ -1,8 +1,6 @@
 /* GStreamer
  * Copyright (C) 2005 Wim Taymans <wim@fluendo.com>
  *
- * gstringbuffer.c: 
- *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Library General Public
  * License as published by the Free Software Foundation; either
@@ -18,6 +16,7 @@
  * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
  * Boston, MA 02111-1307, USA.
  */
+
 /**
  * SECTION:gstringbuffer
  * @short_description: Base class for audio ringbuffer implementations
@@ -30,9 +29,9 @@
  * </para>
  * <para>
  * The ringbuffer abstracts a circular buffer of data. One reader and
- * one writer can operate on the data from different threads in a lockfree 
+ * one writer can operate on the data from different threads in a lockfree
  * manner. The base class is sufficiently flexible to be used as an
- * abstraction for DMA based ringbuffers as well as a pure software 
+ * abstraction for DMA based ringbuffers as well as a pure software
  * implementations.
  * </para>
  * </refsect2>
@@ -117,6 +116,8 @@ gst_ring_buffer_dispose (GObject * object)
 {
   GstRingBuffer *ringbuffer = GST_RING_BUFFER (object);
 
+  gst_caps_replace (&ringbuffer->spec.caps, NULL);
+
   G_OBJECT_CLASS (parent_class)->dispose (G_OBJECT (ringbuffer));
 }
 
@@ -131,55 +132,61 @@ gst_ring_buffer_finalize (GObject * object)
   G_OBJECT_CLASS (parent_class)->finalize (G_OBJECT (ringbuffer));
 }
 
-static const int linear_formats[4 * 2 * 2] = {
-  GST_S8,
-  GST_S8,
-  GST_U8,
-  GST_U8,
-  GST_S16_LE,
-  GST_S16_BE,
-  GST_U16_LE,
-  GST_U16_BE,
-  GST_S24_LE,
-  GST_S24_BE,
-  GST_U24_LE,
-  GST_U24_BE,
-  GST_S32_LE,
-  GST_S32_BE,
-  GST_U32_LE,
-  GST_U32_BE
+typedef struct
+{
+  GstBufferFormat format;
+  guint8 silence[4];
+} FormatDef;
+
+static const FormatDef linear_defs[4 * 2 * 2] = {
+  {GST_S8, {0x00, 0x00, 0x00, 0x00}},
+  {GST_S8, {0x00, 0x00, 0x00, 0x00}},
+  {GST_U8, {0x80, 0x80, 0x80, 0x80}},
+  {GST_U8, {0x80, 0x80, 0x80, 0x80}},
+  {GST_S16_LE, {0x00, 0x00, 0x00, 0x00}},
+  {GST_S16_BE, {0x00, 0x00, 0x00, 0x00}},
+  {GST_U16_LE, {0x00, 0x80, 0x00, 0x80}},
+  {GST_U16_BE, {0x80, 0x00, 0x80, 0x00}},
+  {GST_S24_LE, {0x00, 0x00, 0x00, 0x00}},
+  {GST_S24_BE, {0x00, 0x00, 0x00, 0x00}},
+  {GST_U24_LE, {0x00, 0x00, 0x80, 0x00}},
+  {GST_U24_BE, {0x80, 0x00, 0x00, 0x00}},
+  {GST_S32_LE, {0x00, 0x00, 0x00, 0x00}},
+  {GST_S32_BE, {0x00, 0x00, 0x00, 0x00}},
+  {GST_U32_LE, {0x00, 0x00, 0x00, 0x80}},
+  {GST_U32_BE, {0x80, 0x00, 0x00, 0x00}}
 };
 
-static const int linear24_formats[3 * 2 * 2] = {
-  GST_S24_3LE,
-  GST_S24_3BE,
-  GST_U24_3LE,
-  GST_U24_3BE,
-  GST_S20_3LE,
-  GST_S20_3BE,
-  GST_U20_3LE,
-  GST_U20_3BE,
-  GST_S18_3LE,
-  GST_S18_3BE,
-  GST_U18_3LE,
-  GST_U18_3BE,
+static const FormatDef linear24_defs[3 * 2 * 2] = {
+  {GST_S24_3LE, {0x00, 0x00, 0x00, 0x00}},
+  {GST_S24_3BE, {0x00, 0x00, 0x00, 0x00}},
+  {GST_U24_3LE, {0x00, 0x00, 0x80, 0x00}},
+  {GST_U24_3BE, {0x80, 0x00, 0x00, 0x00}},
+  {GST_S20_3LE, {0x00, 0x00, 0x00, 0x00}},
+  {GST_S20_3BE, {0x00, 0x00, 0x00, 0x00}},
+  {GST_U20_3LE, {0x00, 0x00, 0x08, 0x00}},
+  {GST_U20_3BE, {0x08, 0x00, 0x00, 0x00}},
+  {GST_S18_3LE, {0x00, 0x00, 0x00, 0x00}},
+  {GST_S18_3BE, {0x00, 0x00, 0x00, 0x00}},
+  {GST_U18_3LE, {0x00, 0x00, 0x02, 0x00}},
+  {GST_U18_3BE, {0x02, 0x00, 0x00, 0x00}}
 };
 
-static GstBufferFormat
+static const FormatDef *
 build_linear_format (int depth, int width, int unsignd, int big_endian)
 {
-  const gint *formats;
+  const FormatDef *formats;
 
   if (width == 24) {
     switch (depth) {
       case 24:
-        formats = &linear24_formats[0];
+        formats = &linear24_defs[0];
         break;
       case 20:
-        formats = &linear24_formats[4];
+        formats = &linear24_defs[4];
         break;
       case 18:
-        formats = &linear24_formats[8];
+        formats = &linear24_defs[8];
         break;
       default:
         return GST_UNKNOWN;
@@ -187,16 +194,16 @@ build_linear_format (int depth, int width, int unsignd, int big_endian)
   } else {
     switch (depth) {
       case 8:
-        formats = &linear_formats[0];
+        formats = &linear_defs[0];
         break;
       case 16:
-        formats = &linear_formats[4];
+        formats = &linear_defs[4];
         break;
       case 24:
-        formats = &linear_formats[8];
+        formats = &linear_defs[8];
         break;
       case 32:
-        formats = &linear_formats[12];
+        formats = &linear_defs[12];
         break;
       default:
         return GST_UNKNOWN;
@@ -206,7 +213,8 @@ build_linear_format (int depth, int width, int unsignd, int big_endian)
     formats += 2;
   if (big_endian)
     formats += 1;
-  return (GstBufferFormat) * formats;
+
+  return formats;
 }
 
 /**
@@ -218,6 +226,8 @@ build_linear_format (int depth, int width, int unsignd, int big_endian)
 void
 gst_ring_buffer_debug_spec_caps (GstRingBufferSpec * spec)
 {
+  gint i, bytes;
+
   GST_DEBUG ("spec caps: %p %" GST_PTR_FORMAT, spec->caps, spec->caps);
   GST_DEBUG ("parsed caps: type:         %d", spec->type);
   GST_DEBUG ("parsed caps: format:       %d", spec->format);
@@ -228,6 +238,10 @@ gst_ring_buffer_debug_spec_caps (GstRingBufferSpec * spec)
   GST_DEBUG ("parsed caps: rate:         %d", spec->rate);
   GST_DEBUG ("parsed caps: channels:     %d", spec->channels);
   GST_DEBUG ("parsed caps: sample bytes: %d", spec->bytes_per_sample);
+  bytes = (spec->width >> 3) * spec->channels;
+  for (i = 0; i < bytes; i++) {
+    GST_DEBUG ("silence byte %d: %02x", i, spec->silence_sample[i]);
+  }
 }
 
 /**
@@ -265,14 +279,22 @@ gst_ring_buffer_parse_caps (GstRingBufferSpec * spec, GstCaps * caps)
 {
   const gchar *mimetype;
   GstStructure *structure;
+  gint i;
 
   structure = gst_caps_get_structure (caps, 0);
 
   /* we have to differentiate between int and float formats */
   mimetype = gst_structure_get_name (structure);
 
+  /* get rate and channels */
+  if (!(gst_structure_get_int (structure, "rate", &spec->rate) &&
+          gst_structure_get_int (structure, "channels", &spec->channels)))
+    goto parse_error;
+
   if (!strncmp (mimetype, "audio/x-raw-int", 15)) {
     gint endianness;
+    const FormatDef *def;
+    gint j, bytes;
 
     spec->type = GST_BUFTYPE_LINEAR;
 
@@ -292,9 +314,18 @@ gst_ring_buffer_parse_caps (GstRingBufferSpec * spec, GstCaps * caps)
 
     spec->bigend = endianness == G_LITTLE_ENDIAN ? FALSE : TRUE;
 
-    spec->format =
-        build_linear_format (spec->depth, spec->width, spec->sign ? 0 : 1,
+    def = build_linear_format (spec->depth, spec->width, spec->sign ? 0 : 1,
         spec->bigend ? 1 : 0);
+
+    spec->format = def->format;
+
+    bytes = spec->width >> 3;
+
+    for (i = 0; i < spec->channels; i++) {
+      for (j = 0; j < bytes; j++) {
+        spec->silence_sample[i * bytes + j] = def->silence[j];
+      }
+    }
   } else if (!strncmp (mimetype, "audio/x-raw-float", 17)) {
 
     spec->type = GST_BUFTYPE_FLOAT;
@@ -316,34 +347,40 @@ gst_ring_buffer_parse_caps (GstRingBufferSpec * spec, GstCaps * caps)
       default:
         goto parse_error;
     }
+    /* float silence is all zeros.. */
+    memset (spec->silence_sample, 0, 32);
   } else if (!strncmp (mimetype, "audio/x-alaw", 12)) {
     spec->type = GST_BUFTYPE_A_LAW;
     spec->format = GST_A_LAW;
     spec->width = 8;
     spec->depth = 8;
+    for (i = 0; i < spec->channels; i++)
+      spec->silence_sample[i] = 0xd5;
   } else if (!strncmp (mimetype, "audio/x-mulaw", 13)) {
     spec->type = GST_BUFTYPE_MU_LAW;
     spec->format = GST_MU_LAW;
     spec->width = 8;
     spec->depth = 8;
+    for (i = 0; i < spec->channels; i++)
+      spec->silence_sample[i] = 0xff;
   } else {
     goto parse_error;
   }
 
-  /* get rate and channels */
-  if (!(gst_structure_get_int (structure, "rate", &spec->rate) &&
-          gst_structure_get_int (structure, "channels", &spec->channels)))
-    goto parse_error;
-
   spec->bytes_per_sample = (spec->width >> 3) * spec->channels;
 
   gst_caps_replace (&spec->caps, caps);
 
   g_return_val_if_fail (spec->latency_time != 0, FALSE);
 
-  /* calculate suggested segsize and segtotal */
-  spec->segsize =
-      spec->rate * spec->bytes_per_sample * spec->latency_time / GST_MSECOND;
+  /* calculate suggested segsize and segtotal. segsize should be one unit
+   * of 'latency_time' samples, scaling for the fact that latency_time is
+   * currently stored in microseconds (FIXME: in 0.11) */
+  spec->segsize = gst_util_uint64_scale (spec->rate * spec->bytes_per_sample,
+      spec->latency_time, GST_SECOND / GST_USECOND);
+  /* Round to an integer number of samples */
+  spec->segsize -= spec->segsize % spec->bytes_per_sample;
+
   spec->segtotal = spec->buffer_time / spec->latency_time;
 
   gst_ring_buffer_debug_spec_caps (spec);
@@ -374,7 +411,7 @@ void
 gst_ring_buffer_set_callback (GstRingBuffer * buf, GstRingBufferCallback cb,
     gpointer user_data)
 {
-  g_return_if_fail (buf != NULL);
+  g_return_if_fail (GST_IS_RING_BUFFER (buf));
 
   GST_OBJECT_LOCK (buf);
   buf->callback = cb;
@@ -557,7 +594,7 @@ gst_ring_buffer_acquire (GstRingBuffer * buf, GstRingBufferSpec * spec)
   gint i, j;
   gint segsize, bps;
 
-  g_return_val_if_fail (buf != NULL, FALSE);
+  g_return_val_if_fail (GST_IS_RING_BUFFER (buf), FALSE);
 
   GST_DEBUG_OBJECT (buf, "acquiring device");
 
@@ -587,6 +624,10 @@ gst_ring_buffer_acquire (GstRingBuffer * buf, GstRingBufferSpec * spec)
   /* create an empty segment */
   g_free (buf->empty_seg);
   buf->empty_seg = g_malloc (segsize);
+
+  /* FIXME, we only have 32 silence samples, which might not be enough to
+   * represent silence in all channels */
+  bps = MIN (bps, 32);
   for (i = 0, j = 0; i < segsize; i++) {
     buf->empty_seg[i] = buf->spec.silence_sample[j];
     j = (j + 1) % bps;
@@ -644,7 +685,7 @@ gst_ring_buffer_release (GstRingBuffer * buf)
   gboolean res = FALSE;
   GstRingBufferClass *rclass;
 
-  g_return_val_if_fail (buf != NULL, FALSE);
+  g_return_val_if_fail (GST_IS_RING_BUFFER (buf), FALSE);
 
   GST_DEBUG_OBJECT (buf, "releasing device");
 
@@ -709,7 +750,7 @@ gst_ring_buffer_is_acquired (GstRingBuffer * buf)
 {
   gboolean res;
 
-  g_return_val_if_fail (buf != NULL, FALSE);
+  g_return_val_if_fail (GST_IS_RING_BUFFER (buf), FALSE);
 
   GST_OBJECT_LOCK (buf);
   res = buf->acquired;
@@ -721,6 +762,7 @@ gst_ring_buffer_is_acquired (GstRingBuffer * buf)
 /**
  * gst_ring_buffer_set_flushing:
  * @buf: the #GstRingBuffer to flush
+ * @flushing: the new mode
  *
  * Set the ringbuffer to flushing mode or normal mode.
  *
@@ -729,6 +771,8 @@ gst_ring_buffer_is_acquired (GstRingBuffer * buf)
 void
 gst_ring_buffer_set_flushing (GstRingBuffer * buf, gboolean flushing)
 {
+  g_return_if_fail (GST_IS_RING_BUFFER (buf));
+
   GST_OBJECT_LOCK (buf);
   buf->abidata.ABI.flushing = flushing;
 
@@ -756,7 +800,7 @@ gst_ring_buffer_start (GstRingBuffer * buf)
   GstRingBufferClass *rclass;
   gboolean resume = FALSE;
 
-  g_return_val_if_fail (buf != NULL, FALSE);
+  g_return_val_if_fail (GST_IS_RING_BUFFER (buf), FALSE);
 
   GST_DEBUG_OBJECT (buf, "starting ringbuffer");
 
@@ -865,7 +909,7 @@ gst_ring_buffer_pause (GstRingBuffer * buf)
 {
   gboolean res = FALSE;
 
-  g_return_val_if_fail (buf != NULL, FALSE);
+  g_return_val_if_fail (GST_IS_RING_BUFFER (buf), FALSE);
 
   GST_OBJECT_LOCK (buf);
   if (G_UNLIKELY (buf->abidata.ABI.flushing))
@@ -899,7 +943,7 @@ gst_ring_buffer_stop (GstRingBuffer * buf)
   gboolean res = FALSE;
   GstRingBufferClass *rclass;
 
-  g_return_val_if_fail (buf != NULL, FALSE);
+  g_return_val_if_fail (GST_IS_RING_BUFFER (buf), FALSE);
 
   GST_DEBUG_OBJECT (buf, "stopping");
 
@@ -955,7 +999,7 @@ gst_ring_buffer_delay (GstRingBuffer * buf)
   GstRingBufferClass *rclass;
   guint res;
 
-  g_return_val_if_fail (buf != NULL, 0);
+  g_return_val_if_fail (GST_IS_RING_BUFFER (buf), 0);
 
   res = 0;
 
@@ -989,7 +1033,7 @@ gst_ring_buffer_samples_done (GstRingBuffer * buf)
   guint64 raw, samples;
   guint delay;
 
-  g_return_val_if_fail (buf != NULL, 0);
+  g_return_val_if_fail (GST_IS_RING_BUFFER (buf), 0);
 
   /* get the amount of segments we processed */
   segdone = g_atomic_int_get (&buf->segdone);
@@ -1001,6 +1045,8 @@ gst_ring_buffer_samples_done (GstRingBuffer * buf)
 
   if (G_LIKELY (samples >= delay))
     samples -= delay;
+  else
+    samples = 0;
 
   GST_DEBUG_OBJECT (buf, "processed samples: raw %llu, delay %u, real %llu",
       raw, delay, samples);
@@ -1025,7 +1071,7 @@ gst_ring_buffer_samples_done (GstRingBuffer * buf)
 void
 gst_ring_buffer_set_sample (GstRingBuffer * buf, guint64 sample)
 {
-  g_return_if_fail (buf != NULL);
+  g_return_if_fail (GST_IS_RING_BUFFER (buf));
 
   if (sample == -1)
     sample = 0;
@@ -1057,7 +1103,7 @@ gst_ring_buffer_clear_all (GstRingBuffer * buf)
 {
   gint i;
 
-  g_return_if_fail (buf != NULL);
+  g_return_if_fail (GST_IS_RING_BUFFER (buf));
 
   /* not fatal, we just are not negotiated yet */
   if (G_UNLIKELY (buf->spec.segtotal <= 0))
@@ -1159,7 +1205,7 @@ gst_ring_buffer_commit (GstRingBuffer * buf, guint64 sample, guchar * data,
   guint8 *dest;
   guint to_write;
 
-  g_return_val_if_fail (buf != NULL, -1);
+  g_return_val_if_fail (GST_IS_RING_BUFFER (buf), -1);
   g_return_val_if_fail (buf->data != NULL, -1);
   g_return_val_if_fail (data != NULL, -1);
 
@@ -1265,7 +1311,7 @@ gst_ring_buffer_read (GstRingBuffer * buf, guint64 sample, guchar * data,
   gint segsize, segtotal, bps, sps;
   guint8 *dest;
 
-  g_return_val_if_fail (buf != NULL, -1);
+  g_return_val_if_fail (GST_IS_RING_BUFFER (buf), -1);
   g_return_val_if_fail (buf->data != NULL, -1);
   g_return_val_if_fail (data != NULL, -1);
 
@@ -1370,7 +1416,7 @@ gst_ring_buffer_prepare_read (GstRingBuffer * buf, gint * segment,
   guint8 *data;
   gint segdone;
 
-  g_return_val_if_fail (buf != NULL, FALSE);
+  g_return_val_if_fail (GST_IS_RING_BUFFER (buf), FALSE);
 
   /* buffer must be started */
   if (g_atomic_int_get (&buf->state) != GST_RING_BUFFER_STATE_STARTED)
@@ -1414,7 +1460,7 @@ gst_ring_buffer_prepare_read (GstRingBuffer * buf, gint * segment,
 void
 gst_ring_buffer_advance (GstRingBuffer * buf, guint advance)
 {
-  g_return_if_fail (buf != NULL);
+  g_return_if_fail (GST_IS_RING_BUFFER (buf));
 
   /* update counter */
   g_atomic_int_add (&buf->segdone, advance);
@@ -1445,7 +1491,7 @@ gst_ring_buffer_clear (GstRingBuffer * buf, gint segment)
 {
   guint8 *data;
 
-  g_return_if_fail (buf != NULL);
+  g_return_if_fail (GST_IS_RING_BUFFER (buf));
 
   /* no data means it's already cleared */
   if (G_UNLIKELY (buf->data == NULL))
@@ -1480,6 +1526,8 @@ gst_ring_buffer_clear (GstRingBuffer * buf, gint segment)
 void
 gst_ring_buffer_may_start (GstRingBuffer * buf, gboolean allowed)
 {
+  g_return_if_fail (GST_IS_RING_BUFFER (buf));
+
   GST_LOG_OBJECT (buf, "may start: %d", allowed);
   gst_atomic_int_set (&buf->abidata.ABI.may_start, allowed);
 }
