@@ -230,6 +230,13 @@ gst_riff_create_video_caps (guint32 codec_fcc,
         *codec_name = g_strdup ("VideoSoft H.264");
       break;
 
+    case GST_MAKE_FOURCC ('S', 'E', 'D', 'G'):
+      caps = gst_caps_new_simple ("video/mpeg",
+          "mpegversion", G_TYPE_INT, 4, NULL);
+      if (codec_name)
+        *codec_name = g_strdup ("Samsung MPEG-4");
+      break;
+
     case GST_MAKE_FOURCC ('D', 'I', 'V', '3'):
     case GST_MAKE_FOURCC ('d', 'i', 'v', '3'):
     case GST_MAKE_FOURCC ('D', 'I', 'V', '4'):
@@ -362,6 +369,14 @@ gst_riff_create_video_caps (guint32 codec_fcc,
         *codec_name = g_strdup ("Microsoft Windows Media 9");
       break;
 
+    case GST_MAKE_FOURCC ('W', 'M', 'V', 'A'):
+      caps = gst_caps_new_simple ("video/x-wmv",
+          "wmvversion", G_TYPE_INT, 3, "fourcc", GST_TYPE_FOURCC,
+          codec_fcc, NULL);
+      if (codec_name)
+        *codec_name = g_strdup ("Microsoft Windows Media Advanced Profile");
+      break;
+
     case GST_MAKE_FOURCC ('c', 'v', 'i', 'd'):
       caps = gst_caps_new_simple ("video/x-cinepak", NULL);
       if (codec_name)
@@ -841,6 +856,12 @@ gst_riff_create_audio_caps (guint16 codec_id,
       if (codec_name)
         *codec_name = g_strdup ("AC-3 audio");
       break;
+    case GST_RIFF_WAVE_FORMAT_DTS:
+      channels_max = 6;
+      caps = gst_caps_new_simple ("audio/x-dts", NULL);
+      if (codec_name)
+        *codec_name = g_strdup ("DTS audio");
+      break;
     case GST_RIFF_WAVE_FORMAT_WMAV1:
     case GST_RIFF_WAVE_FORMAT_WMAV2:
     case GST_RIFF_WAVE_FORMAT_WMAV3:
@@ -879,13 +900,15 @@ gst_riff_create_audio_caps (guint16 codec_id,
       guint32 subformat_guid[4];
       const guint8 *data;
 
-      if (GST_BUFFER_SIZE (strf_data) != 22) {
+      /* should be at least 22 bytes */
+      if (strf_data == NULL || GST_BUFFER_SIZE (strf_data) < 22) {
         GST_WARNING ("WAVE_FORMAT_EXTENSIBLE data size is %d (expected: 22)",
-            GST_BUFFER_SIZE (strf_data));
+            (strf_data) ? GST_BUFFER_SIZE (strf_data) : -1);
         return NULL;
       }
 
       data = GST_BUFFER_DATA (strf_data);
+
       valid_bits_per_sample = GST_READ_UINT16_LE (data);
       channel_mask = GST_READ_UINT32_LE (data + 2);
       subformat_guid[0] = GST_READ_UINT32_LE (data + 6);
@@ -932,13 +955,40 @@ gst_riff_create_audio_caps (guint16 codec_id,
           }
         } else if (subformat_guid[0] == 0x00000003) {
           GST_DEBUG ("FIXME: handle IEEE float format");
+        } else if (subformat_guid[0] == 0x00000092) {
+          GST_DEBUG ("FIXME: handle DOLBY AC3 SPDIF format");
         }
-
-        if (caps == NULL) {
-          GST_WARNING ("Unknown WAVE_FORMAT_EXTENSIBLE audio format");
-          return NULL;
-        }
+      } else if (subformat_guid[0] == 0x6ba47966 &&
+          subformat_guid[1] == 0x41783f83 &&
+          subformat_guid[2] == 0xf0006596 && subformat_guid[3] == 0xe59262bf) {
+        caps = gst_caps_new_simple ("application/x-ogg-avi", NULL);
       }
+
+      if (caps == NULL) {
+        GST_WARNING ("Unknown WAVE_FORMAT_EXTENSIBLE audio format");
+        return NULL;
+      }
+      break;
+    }
+
+      /* can anything decode these? pitfdll? */
+    case GST_RIFF_WAVE_FORMAT_VOXWARE:
+    case GST_RIFF_WAVE_FORMAT_VOXWARE_BYTE_ALIGNED:
+    case GST_RIFF_WAVE_FORMAT_VOXWARE_AC8:
+    case GST_RIFF_WAVE_FORMAT_VOXWARE_AC10:
+    case GST_RIFF_WAVE_FORMAT_VOXWARE_AC16:
+    case GST_RIFF_WAVE_FORMAT_VOXWARE_AC20:
+    case GST_RIFF_WAVE_FORMAT_VOXWARE_METAVOICE:
+    case GST_RIFF_WAVE_FORMAT_VOXWARE_METASOUND:
+    case GST_RIFF_WAVE_FORMAT_VOXWARE_RT29HW:
+    case GST_RIFF_WAVE_FORMAT_VOXWARE_VR12:
+    case GST_RIFF_WAVE_FORMAT_VOXWARE_VR18:
+    case GST_RIFF_WAVE_FORMAT_VOXWARE_TQ40:
+    case GST_RIFF_WAVE_FORMAT_VOXWARE_TQ60:{
+      caps = gst_caps_new_simple ("audio/x-voxware",
+          "voxwaretype", G_TYPE_INT, (gint) codec_id, NULL);
+      if (codec_name)
+        *codec_name = g_strdup ("Voxware");
       break;
     }
     default:
@@ -1010,63 +1060,63 @@ gst_riff_create_iavs_caps (guint32 codec_fcc,
 GstCaps *
 gst_riff_create_video_template_caps (void)
 {
-  guint32 tags[] = {
-    GST_MAKE_FOURCC ('I', '4', '2', '0'),
-    GST_MAKE_FOURCC ('Y', 'U', 'Y', '2'),
-    GST_MAKE_FOURCC ('Y', 'V', 'U', '9'),
-    GST_MAKE_FOURCC ('M', 'J', 'P', 'G'),
+  static const guint32 tags[] = {
+    GST_MAKE_FOURCC ('3', 'I', 'V', '1'),
+    GST_MAKE_FOURCC ('A', 'S', 'V', '1'),
+    GST_MAKE_FOURCC ('A', 'S', 'V', '2'),
+    GST_MAKE_FOURCC ('C', 'L', 'J', 'R'),
+    GST_MAKE_FOURCC ('C', 'Y', 'U', 'V'),
+    GST_MAKE_FOURCC ('D', 'I', 'B', ' '),
+    GST_MAKE_FOURCC ('D', 'I', 'V', '3'),
+    GST_MAKE_FOURCC ('D', 'I', 'V', 'X'),
+    GST_MAKE_FOURCC ('D', 'U', 'C', 'K'),
     GST_MAKE_FOURCC ('D', 'V', 'S', 'D'),
+    GST_MAKE_FOURCC ('D', 'X', '5', '0'),
+    GST_MAKE_FOURCC ('F', 'L', 'V', '1'),
+    GST_MAKE_FOURCC ('H', '2', '6', '3'),
+    GST_MAKE_FOURCC ('H', 'F', 'Y', 'U'),
+    GST_MAKE_FOURCC ('I', '2', '6', '3'),
+    GST_MAKE_FOURCC ('I', '4', '2', '0'),
+    GST_MAKE_FOURCC ('I', 'V', '3', '2'),
+    GST_MAKE_FOURCC ('I', 'V', '4', '1'),
+    GST_MAKE_FOURCC ('I', 'V', '5', '0'),
+    GST_MAKE_FOURCC ('M', '4', 'S', '2'),
+    GST_MAKE_FOURCC ('M', 'J', 'P', 'G'),
+    GST_MAKE_FOURCC ('M', 'P', '4', '2'),
+    GST_MAKE_FOURCC ('M', 'P', '4', '3'),
+    GST_MAKE_FOURCC ('M', 'P', 'E', 'G'),
+    GST_MAKE_FOURCC ('M', 'P', 'G', '2'),
+    GST_MAKE_FOURCC ('M', 'P', 'G', '4'),
+    GST_MAKE_FOURCC ('M', 'S', 'Z', 'H'),
+    GST_MAKE_FOURCC ('P', 'N', 'G', ' '),
+    GST_MAKE_FOURCC ('R', 'L', 'E', ' '),
+    GST_MAKE_FOURCC ('R', 'T', '2', '1'),
+    GST_MAKE_FOURCC ('S', 'P', '5', '3'),
+    GST_MAKE_FOURCC ('T', 'M', '2', '0'),
+    GST_MAKE_FOURCC ('T', 'S', 'C', 'C'),
+    GST_MAKE_FOURCC ('U', 'L', 'T', 'I'),
+    GST_MAKE_FOURCC ('V', 'C', 'R', '1'),
+    GST_MAKE_FOURCC ('V', 'C', 'R', '2'),
+    GST_MAKE_FOURCC ('V', 'M', 'n', 'c'),
+    GST_MAKE_FOURCC ('V', 'P', '3', ' '),
     GST_MAKE_FOURCC ('W', 'M', 'V', '1'),
     GST_MAKE_FOURCC ('W', 'M', 'V', '2'),
     GST_MAKE_FOURCC ('W', 'M', 'V', '3'),
-    GST_MAKE_FOURCC ('M', 'P', 'G', '4'),
-    GST_MAKE_FOURCC ('M', 'P', '4', '2'),
-    GST_MAKE_FOURCC ('M', 'P', '4', '3'),
-    GST_MAKE_FOURCC ('H', 'F', 'Y', 'U'),
-    GST_MAKE_FOURCC ('D', 'I', 'V', '3'),
-    GST_MAKE_FOURCC ('M', 'P', 'E', 'G'),
-    GST_MAKE_FOURCC ('H', '2', '6', '3'),
-    GST_MAKE_FOURCC ('I', '2', '6', '3'),
-    GST_MAKE_FOURCC ('h', '2', '6', '4'),
-    GST_MAKE_FOURCC ('D', 'I', 'V', 'X'),
-    GST_MAKE_FOURCC ('D', 'X', '5', '0'),
     GST_MAKE_FOURCC ('X', 'V', 'I', 'D'),
-    GST_MAKE_FOURCC ('3', 'I', 'V', '1'),
-    GST_MAKE_FOURCC ('c', 'v', 'i', 'd'),
-    GST_MAKE_FOURCC ('m', 's', 'v', 'c'),
-    GST_MAKE_FOURCC ('R', 'L', 'E', ' '),
-    GST_MAKE_FOURCC ('D', 'I', 'B', ' '),
     GST_MAKE_FOURCC ('X', 'x', 'a', 'n'),
-    GST_MAKE_FOURCC ('I', 'V', '3', '2'),
-    GST_MAKE_FOURCC ('I', 'V', '5', '0'),
-    GST_MAKE_FOURCC ('M', '4', 'S', '2'),
-    GST_MAKE_FOURCC ('M', 'S', 'Z', 'H'),
+    GST_MAKE_FOURCC ('Y', 'U', 'Y', '2'),
+    GST_MAKE_FOURCC ('Y', 'V', 'U', '9'),
     GST_MAKE_FOURCC ('Z', 'L', 'I', 'B'),
-    GST_MAKE_FOURCC ('A', 'S', 'V', '1'),
-    GST_MAKE_FOURCC ('A', 'S', 'V', '2'),
-    GST_MAKE_FOURCC ('V', 'C', 'R', '1'),
-    GST_MAKE_FOURCC ('V', 'C', 'R', '2'),
-    GST_MAKE_FOURCC ('C', 'L', 'J', 'R'),
-    GST_MAKE_FOURCC ('I', 'V', '4', '1'),
-    GST_MAKE_FOURCC ('R', 'T', '2', '1'),
-    GST_MAKE_FOURCC ('D', 'U', 'C', 'K'),
-    GST_MAKE_FOURCC ('T', 'M', '2', '0'),
-    GST_MAKE_FOURCC ('U', 'L', 'T', 'I'),
-    GST_MAKE_FOURCC ('V', 'P', '3', ' '),
-    GST_MAKE_FOURCC ('T', 'S', 'C', 'C'),
-    GST_MAKE_FOURCC ('S', 'P', '5', '3'),
-    GST_MAKE_FOURCC ('P', 'N', 'G', ' '),
-    GST_MAKE_FOURCC ('C', 'Y', 'U', 'V'),
-    GST_MAKE_FOURCC ('F', 'L', 'V', '1'),
-    GST_MAKE_FOURCC ('V', 'M', 'n', 'c'),
-    /* FILL ME */
-    0
+    GST_MAKE_FOURCC ('c', 'v', 'i', 'd'),
+    GST_MAKE_FOURCC ('h', '2', '6', '4'),
+    GST_MAKE_FOURCC ('m', 's', 'v', 'c')
+        /* FILL ME */
   };
   guint i;
   GstCaps *caps, *one;
 
   caps = gst_caps_new_empty ();
-  for (i = 0; tags[i] != 0; i++) {
+  for (i = 0; i < G_N_ELEMENTS (tags); i++) {
     one = gst_riff_create_video_caps (tags[i], NULL, NULL, NULL, NULL, NULL);
     if (one)
       gst_caps_append (caps, one);
@@ -1078,12 +1128,13 @@ gst_riff_create_video_template_caps (void)
 GstCaps *
 gst_riff_create_audio_template_caps (void)
 {
-  guint16 tags[] = {
+  static const guint16 tags[] = {
     GST_RIFF_WAVE_FORMAT_MPEGL3,
     GST_RIFF_WAVE_FORMAT_MPEGL12,
     GST_RIFF_WAVE_FORMAT_PCM,
     GST_RIFF_WAVE_FORMAT_VORBIS1,
     GST_RIFF_WAVE_FORMAT_A52,
+    GST_RIFF_WAVE_FORMAT_DTS,
     GST_RIFF_WAVE_FORMAT_ALAW,
     GST_RIFF_WAVE_FORMAT_MULAW,
     GST_RIFF_WAVE_FORMAT_ADPCM,
@@ -1091,19 +1142,20 @@ gst_riff_create_audio_template_caps (void)
     GST_RIFF_WAVE_FORMAT_WMAV1,
     GST_RIFF_WAVE_FORMAT_WMAV2,
     GST_RIFF_WAVE_FORMAT_WMAV3,
-    GST_RIFF_WAVE_FORMAT_SONY_ATRAC3,
-    /* FILL ME */
-    0
+    GST_RIFF_WAVE_FORMAT_SONY_ATRAC3
+        /* FILL ME */
   };
   guint i;
   GstCaps *caps, *one;
 
   caps = gst_caps_new_empty ();
-  for (i = 0; tags[i] != 0; i++) {
+  for (i = 0; i < G_N_ELEMENTS (tags); i++) {
     one = gst_riff_create_audio_caps (tags[i], NULL, NULL, NULL, NULL, NULL);
     if (one)
       gst_caps_append (caps, one);
   }
+  one = gst_caps_new_simple ("application/x-ogg-avi", NULL);
+  gst_caps_append (caps, one);
 
   return caps;
 }
@@ -1111,16 +1163,15 @@ gst_riff_create_audio_template_caps (void)
 GstCaps *
 gst_riff_create_iavs_template_caps (void)
 {
-  guint32 tags[] = {
-    GST_MAKE_FOURCC ('D', 'V', 'S', 'D'),
-    /* FILL ME */
-    0
+  static const guint32 tags[] = {
+    GST_MAKE_FOURCC ('D', 'V', 'S', 'D')
+        /* FILL ME */
   };
   guint i;
   GstCaps *caps, *one;
 
   caps = gst_caps_new_empty ();
-  for (i = 0; tags[i] != 0; i++) {
+  for (i = 0; i < G_N_ELEMENTS (tags); i++) {
     one = gst_riff_create_iavs_caps (tags[i], NULL, NULL, NULL, NULL, NULL);
     if (one)
       gst_caps_append (caps, one);
