@@ -28,16 +28,12 @@
 
 G_BEGIN_DECLS
 
-#define GST_TYPE_ADDER \
-  (gst_adder_get_type())
-#define GST_ADDER(obj) \
-  (G_TYPE_CHECK_INSTANCE_CAST((obj),GST_TYPE_ADDER,GstAdder))
-#define GST_ADDER_CLASS(klass) \
-  (G_TYPE_CHECK_CLASS_CAST((klass),GST_TYPE_ADDER,GstAdderClass))
-#define GST_IS_ADDER(obj) \
-  (G_TYPE_CHECK_INSTANCE_TYPE((obj),GST_TYPE_ADDER))
-#define GST_IS_ADDER_CLASS(obj) \
-  (G_TYPE_CHECK_CLASS_TYPE((klass),GST_TYPE_ADDER))
+#define GST_TYPE_ADDER            (gst_adder_get_type())
+#define GST_ADDER(obj)            (G_TYPE_CHECK_INSTANCE_CAST((obj),GST_TYPE_ADDER,GstAdder))
+#define GST_IS_ADDER(obj)         (G_TYPE_CHECK_INSTANCE_TYPE((obj),GST_TYPE_ADDER))
+#define GST_ADDER_CLASS(klass)    (G_TYPE_CHECK_CLASS_CAST((klass) ,GST_TYPE_ADDER,GstAdderClass))
+#define GST_IS_ADDER_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE((klass) ,GST_TYPE_ADDER))
+#define GST_ADDER_GET_CLASS(obj)  (G_TYPE_INSTANCE_GET_CLASS((obj) ,GST_TYPE_ADDER,GstAdderClass))
 
 typedef struct _GstAdder             GstAdder;
 typedef struct _GstAdderClass        GstAdderClass;
@@ -56,7 +52,8 @@ struct _GstAdder {
 
   GstPad         *srcpad;
   GstCollectPads *collect;
-  gint            numpads;
+  /* pad counter, used for creating unique request pads */
+  gint            padcount;
 
   /* the next are valid for both int and float */
   GstAdderFormat  format;
@@ -69,12 +66,21 @@ struct _GstAdder {
   gint            depth;
   gboolean        is_signed;
 
+  /* number of bytes per sample, actually width/8 * channels */
+  gint            bps; 
+
   /* function to add samples */
   GstAdderFunction func;
 
   /* counters to keep track of timestamps */
   gint64          timestamp;
   gint64          offset;
+  
+  /* sink event handling */
+  GstPadEventFunction  collect_event;
+  GstSegment      segment;
+  gboolean        segment_pending;
+  guint64	  segment_position;
 };
 
 struct _GstAdderClass {
