@@ -54,19 +54,12 @@ GST_DEBUG_CATEGORY (audioresample_debug);
 #define GST_CAT_DEFAULT audioresample_debug
 
 /* elementfactory information */
-static GstElementDetails gst_audioresample_details =
+static const GstElementDetails gst_audioresample_details =
 GST_ELEMENT_DETAILS ("Audio scaler",
     "Filter/Converter/Audio",
     "Resample audio",
     "David Schleef <ds@schleef.org>");
 
-/* GstAudioresample signals and args */
-enum
-{
-  /* FILL ME */
-  LAST_SIGNAL
-};
-
 #define DEFAULT_FILTERLEN       16
 
 enum
@@ -83,16 +76,26 @@ GST_STATIC_CAPS ( \
       "endianness = (int) BYTE_ORDER, " \
       "width = (int) 16, " \
       "depth = (int) 16, " \
-      "signed = (boolean) true " \
+      "signed = (boolean) true;" \
+    "audio/x-raw-int, " \
+      "rate = (int) [ 1, MAX ], " \
+      "channels = (int) [ 1, MAX ], " \
+      "endianness = (int) BYTE_ORDER, " \
+      "width = (int) 32, " \
+      "depth = (int) 32, " \
+      "signed = (boolean) true;" \
+    "audio/x-raw-float, " \
+      "rate = (int) [ 1, MAX ], "	\
+      "channels = (int) [ 1, MAX ], " \
+      "endianness = (int) BYTE_ORDER, " \
+      "width = (int) 32; " \
+    "audio/x-raw-float, " \
+      "rate = (int) [ 1, MAX ], "	\
+      "channels = (int) [ 1, MAX ], " \
+      "endianness = (int) BYTE_ORDER, " \
+      "width = (int) 64" \
 )
 
-#if 0
-  /* disabled because it segfaults */
-"audio/x-raw-float, "
-    "rate = (int) [ 1, MAX ], "
-    "channels = (int) [ 1, MAX ], "
-    "endianness = (int) BYTE_ORDER, " "width = (int) 32"
-#endif
 static GstStaticPadTemplate gst_audioresample_sink_template =
 GST_STATIC_PAD_TEMPLATE ("sink",
     GST_PAD_SINK, GST_PAD_ALWAYS, SUPPORTED_CAPS);
@@ -101,8 +104,6 @@ static GstStaticPadTemplate gst_audioresample_src_template =
 GST_STATIC_PAD_TEMPLATE ("src",
     GST_PAD_SRC, GST_PAD_ALWAYS, SUPPORTED_CAPS);
 
-static void gst_audioresample_dispose (GObject * object);
-
 static void gst_audioresample_set_property (GObject * object,
     guint prop_id, const GValue * value, GParamSpec * pspec);
 static void gst_audioresample_get_property (GObject * object,
@@ -123,8 +124,8 @@ static GstFlowReturn audioresample_pushthrough (GstAudioresample *
 static GstFlowReturn audioresample_transform (GstBaseTransform * base,
     GstBuffer * inbuf, GstBuffer * outbuf);
 static gboolean audioresample_event (GstBaseTransform * base, GstEvent * event);
-
-/*static guint gst_audioresample_signals[LAST_SIGNAL] = { 0 }; */
+static gboolean audioresample_start (GstBaseTransform * base);
+static gboolean audioresample_stop (GstBaseTransform * base);
 
 #define DEBUG_INIT(bla) \
   GST_DEBUG_CATEGORY_INIT (audioresample_debug, "audioresample", 0, "audio resampling element");
@@ -154,13 +155,16 @@ gst_audioresample_class_init (GstAudioresampleClass * klass)
 
   gobject_class->set_property = gst_audioresample_set_property;
   gobject_class->get_property = gst_audioresample_get_property;
-  gobject_class->dispose = gst_audioresample_dispose;
 
-  g_object_class_install_property (G_OBJECT_CLASS (klass), PROP_FILTERLEN,
+  g_object_class_install_property (gobject_class, PROP_FILTERLEN,
       g_param_spec_int ("filter_length", "filter_length", "filter_length",
           0, G_MAXINT, DEFAULT_FILTERLEN,
           G_PARAM_READWRITE | G_PARAM_CONSTRUCT));
 
+  GST_BASE_TRANSFORM_CLASS (klass)->start =
+      GST_DEBUG_FUNCPTR (audioresample_start);
+  GST_BASE_TRANSFORM_CLASS (klass)->stop =
+      GST_DEBUG_FUNCPTR (audioresample_stop);
   GST_BASE_TRANSFORM_CLASS (klass)->transform_size =
       GST_DEBUG_FUNCPTR (audioresample_transform_size);
   GST_BASE_TRANSFORM_CLASS (klass)->get_unit_size =
@@ -181,7 +185,6 @@ static void
 gst_audioresample_init (GstAudioresample * audioresample,
     GstAudioresampleClass * klass)
 {
-  ResampleState *r;
   GstBaseTransform *trans;
 
   trans = GST_BASE_TRANSFORM (audioresample);
@@ -190,30 +193,42 @@ gst_audioresample_init (GstAudioresample * audioresample,
    * is trivial in the passtrough case. */
   gst_pad_set_bufferalloc_function (trans->sinkpad, NULL);
 
-  r = resample_new ();
-  audioresample->resample = r;
+  audioresample->filter_length = DEFAULT_FILTERLEN;
+}
+
+/* vmethods */
+static gboolean
+audioresample_start (GstBaseTransform * base)
+{
+  GstAudioresample *audioresample = GST_AUDIORESAMPLE (base);
+
+  audioresample->resample = resample_new ();
   audioresample->ts_offset = -1;
   audioresample->offset = -1;
   audioresample->next_ts = -1;
 
-  resample_set_filter_length (r, DEFAULT_FILTERLEN);
-  resample_set_format (r, RESAMPLE_FORMAT_S16);
+  resample_set_filter_length (audioresample->resample,
+      audioresample->filter_length);
+
+  return TRUE;
 }
 
-static void
-gst_audioresample_dispose (GObject * object)
+static gboolean
+audioresample_stop (GstBaseTransform * base)
 {
-  GstAudioresample *audioresample = GST_AUDIORESAMPLE (object);
+  GstAudioresample *audioresample = GST_AUDIORESAMPLE (base);
 
   if (audioresample->resample) {
     resample_free (audioresample->resample);
     audioresample->resample = NULL;
   }
 
-  G_OBJECT_CLASS (parent_class)->dispose (object);
+  gst_caps_replace (&audioresample->sinkcaps, NULL);
+  gst_caps_replace (&audioresample->srccaps, NULL);
+
+  return TRUE;
 }
 
-/* vmethods */
 gboolean
 audioresample_get_unit_size (GstBaseTransform * base, GstCaps * caps,
     guint * size)
@@ -222,7 +237,7 @@ audioresample_get_unit_size (GstBaseTransform * base, GstCaps * caps,
   GstStructure *structure;
   gboolean ret;
 
-  g_return_val_if_fail (size, FALSE);
+  g_assert (size);
 
   /* this works for both float and int */
   structure = gst_caps_get_structure (caps, 0);
@@ -259,28 +274,49 @@ resample_set_state_from_caps (ResampleState * state, GstCaps * incaps,
   gboolean ret;
   gint myinrate, myoutrate;
   int mychannels;
+  gint width, depth;
+  ResampleFormat format;
 
   GST_DEBUG ("incaps %" GST_PTR_FORMAT ", outcaps %"
       GST_PTR_FORMAT, incaps, outcaps);
 
   structure = gst_caps_get_structure (incaps, 0);
 
-  /* FIXME: once it does float, set the correct format */
-#if 0
-  if (g_str_equal (gst_structure_get_name (structure), "audio/x-raw-float")) {
-    r->format = GST_RESAMPLE_FLOAT;
-  } else {
-    r->format = GST_RESAMPLE_S16;
-  }
-#endif
+  /* get width */
+  ret = gst_structure_get_int (structure, "width", &width);
+  if (!ret)
+    goto no_width;
 
+  /* figure out the format */
+  if (g_str_equal (gst_structure_get_name (structure), "audio/x-raw-float")) {
+    if (width == 32)
+      format = RESAMPLE_FORMAT_F32;
+    else if (width == 64)
+      format = RESAMPLE_FORMAT_F64;
+    else
+      goto wrong_depth;
+  } else {
+    /* for int, depth and width must be the same */
+    ret = gst_structure_get_int (structure, "depth", &depth);
+    if (!ret || width != depth)
+      goto not_equal;
+
+    if (width == 16)
+      format = RESAMPLE_FORMAT_S16;
+    else if (width == 32)
+      format = RESAMPLE_FORMAT_S32;
+    else
+      goto wrong_depth;
+  }
   ret = gst_structure_get_int (structure, "rate", &myinrate);
   ret &= gst_structure_get_int (structure, "channels", &mychannels);
-  g_return_val_if_fail (ret, FALSE);
+  if (!ret)
+    goto no_in_rate_channels;
 
   structure = gst_caps_get_structure (outcaps, 0);
   ret = gst_structure_get_int (structure, "rate", &myoutrate);
-  g_return_val_if_fail (ret, FALSE);
+  if (!ret)
+    goto no_out_rate;
 
   if (channels)
     *channels = mychannels;
@@ -289,11 +325,39 @@ resample_set_state_from_caps (ResampleState * state, GstCaps * incaps,
   if (outrate)
     *outrate = myoutrate;
 
+  resample_set_format (state, format);
   resample_set_n_channels (state, mychannels);
   resample_set_input_rate (state, myinrate);
   resample_set_output_rate (state, myoutrate);
 
   return TRUE;
+
+  /* ERRORS */
+no_width:
+  {
+    GST_DEBUG ("failed to get width from caps");
+    return FALSE;
+  }
+not_equal:
+  {
+    GST_DEBUG ("width %d and depth %d must be the same", width, depth);
+    return FALSE;
+  }
+wrong_depth:
+  {
+    GST_DEBUG ("unknown depth %d found", depth);
+    return FALSE;
+  }
+no_in_rate_channels:
+  {
+    GST_DEBUG ("could not get input rate and channels");
+    return FALSE;
+  }
+no_out_rate:
+  {
+    GST_DEBUG ("could not get output rate");
+    return FALSE;
+  }
 }
 
 gboolean
@@ -377,10 +441,7 @@ audioresample_set_caps (GstBaseTransform * base, GstCaps * incaps,
 
   /* save caps so we can short-circuit in the size_transform if the caps
    * are the same */
-  /* FIXME: clean them up in state change ? */
-  gst_caps_ref (incaps);
   gst_caps_replace (&audioresample->sinkcaps, incaps);
-  gst_caps_ref (outcaps);
   gst_caps_replace (&audioresample->srccaps, outcaps);
 
   return TRUE;
@@ -581,7 +642,6 @@ gst_audioresample_set_property (GObject * object, guint prop_id,
 {
   GstAudioresample *audioresample;
 
-  g_return_if_fail (GST_IS_AUDIORESAMPLE (object));
   audioresample = GST_AUDIORESAMPLE (object);
 
   switch (prop_id) {
@@ -589,8 +649,10 @@ gst_audioresample_set_property (GObject * object, guint prop_id,
       audioresample->filter_length = g_value_get_int (value);
       GST_DEBUG_OBJECT (GST_ELEMENT (audioresample), "new filter length %d",
           audioresample->filter_length);
-      resample_set_filter_length (audioresample->resample,
-          audioresample->filter_length);
+      if (audioresample->resample) {
+        resample_set_filter_length (audioresample->resample,
+            audioresample->filter_length);
+      }
       break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
@@ -604,7 +666,6 @@ gst_audioresample_get_property (GObject * object, guint prop_id,
 {
   GstAudioresample *audioresample;
 
-  g_return_if_fail (GST_IS_AUDIORESAMPLE (object));
   audioresample = GST_AUDIORESAMPLE (object);
 
   switch (prop_id) {
