@@ -43,20 +43,6 @@ GST_DEBUG_CATEGORY_STATIC (gst_play_base_bin_debug);
 #define GROUP_WAIT(pbb) g_cond_wait (pbb->group_cond, pbb->group_lock)
 #define GROUP_SIGNAL(pbb) g_cond_signal (pbb->group_cond)
 
-#ifndef GST_HAVE_GLIB_2_8
-#define _gst_gvalue_set_gstobject(gvalue,obj)  \
-      if (obj != NULL) {                       \
-        gst_object_ref (obj);                  \
-        g_value_set_object (gvalue, obj);      \
-        g_object_unref (obj);                  \
-      } else {                                 \
-        g_value_set_object (gvalue, NULL);     \
-      }
-#else
-#define _gst_gvalue_set_gstobject(gvalue,obj)  \
-      g_value_set_object (gvalue, obj);
-#endif
-
 /* props */
 enum
 {
@@ -67,10 +53,12 @@ enum
   ARG_QUEUE_THRESHOLD,
   ARG_NSTREAMS,
   ARG_STREAMINFO,
+  ARG_STREAMINFO_VALUES,
   ARG_SOURCE,
   ARG_VIDEO,
   ARG_AUDIO,
-  ARG_TEXT
+  ARG_TEXT,
+  ARG_SUBTITLE_ENCODING
 };
 
 static void gst_play_base_bin_class_init (GstPlayBaseBinClass * klass);
@@ -85,6 +73,8 @@ static void gst_play_base_bin_get_property (GObject * object, guint prop_id,
 static GstStateChangeReturn gst_play_base_bin_change_state (GstElement *
     element, GstStateChange transition);
 const GList *gst_play_base_bin_get_streaminfo (GstPlayBaseBin * play_base_bin);
+const GValueArray *gst_play_base_bin_get_streaminfo_value_array (GstPlayBaseBin
+    * play_base_bin);
 
 static gboolean prepare_output (GstPlayBaseBin * play_base_bin);
 static void set_active_source (GstPlayBaseBin * play_base_bin,
@@ -163,6 +153,11 @@ gst_play_base_bin_class_init (GstPlayBaseBinClass * klass)
   g_object_class_install_property (gobject_klass, ARG_STREAMINFO,
       g_param_spec_pointer ("stream-info", "Stream info", "List of streaminfo",
           G_PARAM_READABLE));
+  g_object_class_install_property (gobject_klass, ARG_STREAMINFO_VALUES,
+      g_param_spec_value_array ("stream-info-value-array",
+          "StreamInfo GValueArray", "value array of streaminfo",
+          g_param_spec_object ("streaminfo", "StreamInfo", "Streaminfo object",
+              GST_TYPE_STREAM_INFO, G_PARAM_READABLE), G_PARAM_READABLE));
   g_object_class_install_property (gobject_klass, ARG_SOURCE,
       g_param_spec_object ("source", "Source", "Source element",
           GST_TYPE_ELEMENT, G_PARAM_READABLE));
@@ -179,6 +174,12 @@ gst_play_base_bin_class_init (GstPlayBaseBinClass * klass)
       g_param_spec_int ("current-text", "Current text",
           "Currently playing text stream (-1 = none)",
           -1, G_MAXINT, -1, G_PARAM_READWRITE));
+  g_object_class_install_property (gobject_klass, ARG_SUBTITLE_ENCODING,
+      g_param_spec_string ("subtitle-encoding", "subtitle encoding",
+          "Encoding to assume if input subtitles are not in UTF-8 encoding. "
+          "If not set, the GST_SUBTITLE_ENCODING environment variable will "
+          "be checked for an encoding to use. If that is not set either, "
+          "ISO-8859-15 will be assumed.", NULL, G_PARAM_READWRITE));
 
   GST_DEBUG_CATEGORY_INIT (gst_play_base_bin_debug, "playbasebin", 0,
       "playbasebin");
@@ -200,6 +201,9 @@ gst_play_base_bin_init (GstPlayBaseBin * play_base_bin)
   play_base_bin->source = NULL;
   play_base_bin->decoder = NULL;
   play_base_bin->subtitle = NULL;
+  play_base_bin->subencoding = NULL;
+  play_base_bin->subtitle_elements = NULL;
+  play_base_bin->sub_lock = g_mutex_new ();
 
   play_base_bin->group_lock = g_mutex_new ();
   play_base_bin->group_cond = g_cond_new ();
@@ -221,7 +225,12 @@ gst_play_base_bin_dispose (GObject * object)
   play_base_bin->uri = NULL;
   g_free (play_base_bin->suburi);
   play_base_bin->suburi = NULL;
-
+  g_free (play_base_bin->subencoding);
+  play_base_bin->subencoding = NULL;
+  if (play_base_bin->subtitle_elements) {
+    g_slist_free (play_base_bin->subtitle_elements);
+    play_base_bin->subtitle_elements = NULL;
+  }
   G_OBJECT_CLASS (parent_class)->dispose (object);
 }
 
@@ -233,6 +242,8 @@ gst_play_base_bin_finalize (GObject * object)
   g_mutex_free (play_base_bin->group_lock);
   g_cond_free (play_base_bin->group_cond);
 
+  g_mutex_free (play_base_bin->sub_lock);
+
   G_OBJECT_CLASS (parent_class)->finalize (object);
 }
 
@@ -247,6 +258,9 @@ group_create (GstPlayBaseBin * play_base_bin)
 
   group = g_new0 (GstPlayBaseGroup, 1);
   group->bin = play_base_bin;
+  group->streaminfo_value_array = g_value_array_new (0);
+
+  GST_DEBUG_OBJECT (play_base_bin, "created new group %p", group);
 
   return group;
 }
@@ -338,7 +352,7 @@ group_destroy (GstPlayBaseGroup * group)
       }
     }
 
-    /* if the group is currently being played, we have to remove the element 
+    /* if the group is currently being played, we have to remove the element
      * from the thread */
     gst_element_set_state (element, GST_STATE_NULL);
     gst_element_set_state (group->type[n].selector, GST_STATE_NULL);
@@ -356,11 +370,12 @@ group_destroy (GstPlayBaseGroup * group)
   /* free the streaminfo too */
   g_list_foreach (group->streaminfo, (GFunc) g_object_unref, NULL);
   g_list_free (group->streaminfo);
+  g_value_array_free (group->streaminfo_value_array);
   g_free (group);
 }
 
 /*
- * is called when the current building group is completely finished 
+ * is called when the current building group is completely finished
  * and ready for playback
  *
  * This function grabs lock, so take care when calling.
@@ -376,6 +391,9 @@ group_commit (GstPlayBaseBin * play_base_bin, gboolean fatal, gboolean subtitle)
   group = play_base_bin->building_group;
   had_active_group = (get_active_group (play_base_bin) != NULL);
 
+  GST_DEBUG_OBJECT (play_base_bin, "commit group %p, had active %d",
+      group, had_active_group);
+
   /* if an element signalled a no-more-pads after we stopped due
    * to preroll, the group is NULL. This is not an error */
   if (group == NULL) {
@@ -383,40 +401,49 @@ group_commit (GstPlayBaseBin * play_base_bin, gboolean fatal, gboolean subtitle)
       GROUP_UNLOCK (play_base_bin);
       return;
     } else {
-      GST_DEBUG ("Group loading failed, bailing out");
+      GST_DEBUG_OBJECT (play_base_bin, "Group loading failed, bailing out");
     }
-  } else if (!subtitle) {
-    gint n;
+  } else {
+    if (!subtitle) {
+      gint n;
 
-    GST_DEBUG ("group %p done", group);
+      GST_DEBUG_OBJECT (play_base_bin, "group %p done", group);
 
-    play_base_bin->queued_groups = g_list_append (play_base_bin->queued_groups,
-        group);
+      play_base_bin->queued_groups =
+          g_list_append (play_base_bin->queued_groups, group);
 
-    play_base_bin->building_group = NULL;
+      play_base_bin->building_group = NULL;
 
-    /* remove signals. We don't want anymore signals from the preroll
-     * elements at this stage. */
-    for (n = 0; n < NUM_TYPES; n++) {
-      GstElement *element = group->type[n].preroll;
-      guint sig_id;
+      /* remove signals. We don't want anymore signals from the preroll
+       * elements at this stage. */
+      for (n = 0; n < NUM_TYPES; n++) {
+        GstElement *element = group->type[n].preroll;
+        guint sig_id;
 
-      if (!element)
-        continue;
+        if (!element)
+          continue;
 
-      sig_id =
-          GPOINTER_TO_INT (g_object_get_data (G_OBJECT (element), "signal_id"));
+        sig_id =
+            GPOINTER_TO_INT (g_object_get_data (G_OBJECT (element),
+                "signal_id"));
 
-      if (sig_id) {
-        GST_LOG ("removing preroll signal %s", GST_ELEMENT_NAME (element));
-        g_signal_handler_disconnect (G_OBJECT (element), sig_id);
+        if (sig_id) {
+          GST_LOG_OBJECT (play_base_bin, "removing preroll signal %s",
+              GST_ELEMENT_NAME (element));
+          g_signal_handler_disconnect (G_OBJECT (element), sig_id);
+        }
       }
+    } else {
+      /* this is a special subtitle bin, we don't commit the group but
+       * mark the subtitles as detected before se signal. */
+      GST_DEBUG_OBJECT (play_base_bin, "marking subtitle bin as complete");
+      play_base_bin->subtitle_done = TRUE;
     }
   }
 
-  GST_DEBUG ("signal group done");
+  GST_DEBUG_OBJECT (play_base_bin, "signal group done");
   GROUP_SIGNAL (play_base_bin);
-  GST_DEBUG ("signaled group done");
+  GST_DEBUG_OBJECT (play_base_bin, "signaled group done");
 
   if (!subtitle && !had_active_group) {
     if (!prepare_output (play_base_bin)) {
@@ -425,10 +452,10 @@ group_commit (GstPlayBaseBin * play_base_bin, gboolean fatal, gboolean subtitle)
     }
 
     setup_substreams (play_base_bin);
-    GST_DEBUG ("Emitting signal");
+    GST_DEBUG_OBJECT (play_base_bin, "Emitting signal");
     res = GST_PLAY_BASE_BIN_GET_CLASS (play_base_bin)->
         setup_output_pads (play_base_bin, group);
-    GST_DEBUG ("done");
+    GST_DEBUG_OBJECT (play_base_bin, "done");
 
     GROUP_UNLOCK (play_base_bin);
 
@@ -496,21 +523,50 @@ check_queue (GstPad * pad, GstBuffer * data, gpointer user_data)
 
 /* this signal will be fired when one of the queues with raw
  * data is filled. This means that the group building stage is over 
- * and playback of the new queued group should start */
+ * and playback of the new queued group should start
+ *
+ * If this queue overruns we can potentially create a deadlock when:
+ *
+ *  1) the max-bytes is hit and
+ *  2) the min-time is not hit.
+ *
+ * We recover from this situation in the overrun callback by
+ * setting the max-bytes to unlimited if we see that there is
+ * a current-time-level (which means some sort of timestamping is
+ * done).
+ */
 static void
-queue_overrun (GstElement * element, GstPlayBaseBin * play_base_bin)
+queue_overrun (GstElement * queue, GstPlayBaseBin * play_base_bin)
 {
-  GST_DEBUG ("queue %s overrun", GST_ELEMENT_NAME (element));
+  GST_DEBUG ("queue %s overrun", GST_ELEMENT_NAME (queue));
+
+  /* if we're streaming, check if we had a deadlock with the 
+   * max-bytes <> min-time thresholds */
+  if (play_base_bin->is_stream) {
+    guint64 time, min_time;
+
+    g_object_get (G_OBJECT (queue), "current-level-time", &time,
+        "min-threshold-time", &min_time, NULL);
+
+    GST_DEBUG_OBJECT (play_base_bin, "streaming mode, queue %s current %"
+        GST_TIME_FORMAT ", min %" GST_TIME_FORMAT,
+        GST_ELEMENT_NAME (queue),
+        GST_TIME_ARGS (time), GST_TIME_ARGS (min_time));
+    if (time != 0) {
+      /* queue knows about time, disable bytes checking. */
+      g_object_set (G_OBJECT (queue), "max-size-bytes", 0, NULL);
+    }
+  }
 
   group_commit (play_base_bin, FALSE,
-      GST_OBJECT_PARENT (GST_OBJECT_CAST (element)) ==
+      GST_OBJECT_PARENT (GST_OBJECT_CAST (queue)) ==
       GST_OBJECT (play_base_bin->subtitle));
 
-  g_signal_handlers_disconnect_by_func (element,
+  g_signal_handlers_disconnect_by_func (queue,
       G_CALLBACK (queue_overrun), play_base_bin);
   /* We have disconnected this signal, remove the signal_id from the object
      data */
-  g_object_set_data (G_OBJECT (element), "signal_id", NULL);
+  g_object_set_data (G_OBJECT (queue), "signal_id", NULL);
 }
 
 /* Used for time-based buffering. */
@@ -522,6 +578,8 @@ queue_threshold_reached (GstElement * queue, GstPlayBaseBin * play_base_bin)
   GST_DEBUG ("Running");
 
   /* play */
+  g_signal_handlers_disconnect_by_func (queue,
+      G_CALLBACK (queue_threshold_reached), play_base_bin);
   g_object_set (queue, "min-threshold-time", (guint64) 0, NULL);
 
   data = g_object_get_data (G_OBJECT (queue), "probe");
@@ -542,14 +600,41 @@ queue_threshold_reached (GstElement * queue, GstPlayBaseBin * play_base_bin)
   }
 }
 
+/* this signal is only added when in streaming mode to catch underruns
+ */
 static void
 queue_out_of_data (GstElement * queue, GstPlayBaseBin * play_base_bin)
 {
+  guint64 time, min_time;
+  guint bytes, max_bytes;
+
   GST_DEBUG ("Underrun, re-caching");
 
+  g_object_get (G_OBJECT (queue), "current-level-time", &time,
+      "current-level-bytes", &bytes,
+      "max-size-bytes", &max_bytes, "min-threshold-time", &min_time, NULL);
+
+  GST_DEBUG_OBJECT (play_base_bin, "streaming mode, queue %s current %"
+      GST_TIME_FORMAT ", min %" GST_TIME_FORMAT
+      ", bytes %d",
+      GST_ELEMENT_NAME (queue),
+      GST_TIME_ARGS (time), GST_TIME_ARGS (min_time), bytes);
+
+  /* if the bytes in the queue represent time, we disable bytes
+   * overrun checking to not cause deadlocks.
+   */
+  if (bytes && time != 0 && time < min_time) {
+    /* queue knows about time but is filled with bytes that do
+     * not represent min-threshold time, disable bytes checking so
+     * the queue can grow some more. */
+    g_object_set (G_OBJECT (queue), "max-size-bytes", 0, NULL);
+  }
+
   /* On underrun, we want to temoprarily pause playback, set a "min-size"
    * threshold and wait for the running signal and then play again. Take
    * care of possible deadlocks and so on, */
+  g_signal_connect (G_OBJECT (queue), "running",
+      G_CALLBACK (queue_threshold_reached), play_base_bin);
   g_object_set (queue, "min-threshold-time",
       (guint64) play_base_bin->queue_threshold, NULL);
 
@@ -611,6 +696,19 @@ gen_preroll_element (GstPlayBaseBin * play_base_bin,
   g_free (name);
   g_free (padname);
 
+  /* for buffering of raw data we ideally want to buffer a
+   * very small amount of buffers since the memory used by
+   * this raw data can be enormously huge.
+   *
+   * FIXME: we abuse this buffer to do network buffering since
+   * we can then easily do time-based buffering. The better
+   * solution would be to add a specific network queue right
+   * after the source that measures the datarate and scales this
+   * queue of encoded data instead.
+   *
+   * We use an upper limit of typically a few seconds here but
+   * cap in case no timestamps are set on the raw data (bad!).
+   */
   g_object_set (G_OBJECT (preroll),
       "max-size-buffers", 0, "max-size-bytes",
       ((type == GST_STREAM_TYPE_VIDEO) ? 25 : 1) * 1024 * 1024,
@@ -641,9 +739,22 @@ gen_preroll_element (GstPlayBaseBin * play_base_bin,
     g_object_set_data (G_OBJECT (preroll), "pbb", play_base_bin);
     g_object_set_data (G_OBJECT (preroll), "probe", GINT_TO_POINTER (id));
 
+    /* 
+     * When we connect this queue, it will start running and immediatly
+     * fire an underrun when:
+     *
+     * 1) the max-bytes is hit
+     * 2) the min-time is not hit.
+     *
+     * We recover from this situation in the out_of_data callback by
+     * setting the max-bytes to unlimited if we see that there is
+     * a current-time-level (which means some sort of timestamping is
+     * done).
+     */
     g_signal_connect (G_OBJECT (preroll), "underrun",
         G_CALLBACK (queue_out_of_data), play_base_bin);
   }
+
   /* keep a ref to the signal id so that we can disconnect the signal callback
    * when we are done with the preroll */
   g_object_set_data (G_OBJECT (preroll), "signal_id", GINT_TO_POINTER (sig));
@@ -708,19 +819,23 @@ remove_groups (GstPlayBaseBin * play_base_bin)
   GROUP_UNLOCK (play_base_bin);
 }
 
-/* 
- * Add/remove a single stream to current  building group.
+/*
+ * Add/remove a single stream to current building group.
  *
  * Must be called with group-lock held.
  */
 static void
 add_stream (GstPlayBaseGroup * group, GstStreamInfo * info)
 {
+  GValue v = { 0, };
   GST_DEBUG ("add stream to group %p", group);
 
   /* keep ref to the group */
   g_object_set_data (G_OBJECT (info), "group", group);
 
+  g_value_init (&v, G_TYPE_OBJECT);
+  g_value_set_object (&v, info);
+  g_value_array_append (group->streaminfo_value_array, &v);
   group->streaminfo = g_list_append (group->streaminfo, info);
   if (info->type > 0 && info->type <= NUM_TYPES) {
     group->type[info->type - 1].npads++;
@@ -1049,6 +1164,48 @@ no_type:
   }
 }
 
+static void
+set_encoding_element (GstElement * element, gchar * encoding)
+{
+  GST_DEBUG_OBJECT (element, "setting encoding to %s", GST_STR_NULL (encoding));
+  g_object_set (G_OBJECT (element), "subtitle-encoding", encoding, NULL);
+}
+
+
+static void
+decodebin_element_added_cb (GstBin * decodebin, GstElement * element,
+    gpointer data)
+{
+  GstPlayBaseBin *play_base_bin = GST_PLAY_BASE_BIN (data);
+  gchar *encoding;
+
+  if (!g_object_class_find_property (G_OBJECT_GET_CLASS (element),
+          "subtitle-encoding")) {
+    return;
+  }
+
+  g_mutex_lock (play_base_bin->sub_lock);
+  play_base_bin->subtitle_elements =
+      g_slist_append (play_base_bin->subtitle_elements, element);
+  encoding = g_strdup (play_base_bin->subencoding);
+  g_mutex_unlock (play_base_bin->sub_lock);
+
+  set_encoding_element (element, encoding);
+  g_free (encoding);
+}
+
+static void
+decodebin_element_removed_cb (GstBin * decodebin, GstElement * element,
+    gpointer data)
+{
+  GstPlayBaseBin *play_base_bin = GST_PLAY_BASE_BIN (data);
+
+  g_mutex_lock (play_base_bin->sub_lock);
+  play_base_bin->subtitle_elements =
+      g_slist_remove (play_base_bin->subtitle_elements, element);
+  g_mutex_unlock (play_base_bin->sub_lock);
+}
+
 
 /*
  * Generate source ! subparse bins.
@@ -1057,21 +1214,41 @@ no_type:
 static GstElement *
 setup_subtitle (GstPlayBaseBin * play_base_bin, gchar * sub_uri)
 {
-  GstElement *source, *subparse, *subbin;
+  GstElement *source, *subdecodebin, *subbin;
 
   source = gst_element_make_from_uri (GST_URI_SRC, sub_uri, NULL);
   if (!source)
     return NULL;
-  subparse = gst_element_factory_make ("decodebin", "subtitle-decoder");
-
+  subdecodebin = gst_element_factory_make ("decodebin", "subtitle-decoder");
+  g_signal_connect (subdecodebin, "element-added",
+      G_CALLBACK (decodebin_element_added_cb), play_base_bin);
+  g_signal_connect (subdecodebin, "element-removed",
+      G_CALLBACK (decodebin_element_removed_cb), play_base_bin);
   subbin = gst_bin_new ("subtitle-bin");
-  gst_bin_add_many (GST_BIN (subbin), source, subparse, NULL);
-  gst_element_link (source, subparse);
+  gst_bin_add_many (GST_BIN (subbin), source, subdecodebin, NULL);
+
+  gst_element_link (source, subdecodebin);
 
   /* return the subtitle GstElement object */
   return subbin;
 }
 
+/* lame - FIXME, maybe we can use seek_types/mask here? */
+static gboolean
+is_stream (GstPlayBaseBin * play_base_bin, gchar * uri)
+{
+  static const gchar *uris[] = { "http://", "mms://", "mmsh://",
+    "mmsu://", "mmst://", "rtp://", "rtsp://", NULL
+  };
+  gint i;
+
+  for (i = 0; uris[i]; i++) {
+    if (g_str_has_prefix (uri, uris[i]))
+      return TRUE;
+  }
+  return FALSE;
+}
+
 /*
  * Generate a source element that does caching for network streams.
  */
@@ -1084,6 +1261,13 @@ gen_source_element (GstPlayBaseBin * play_base_bin, GstElement ** subbin)
   /* stip subtitle from uri */
   if (!play_base_bin->uri)
     return NULL;
+
+  if (g_str_has_prefix (play_base_bin->uri, "rtsp://")) {
+    GST_ELEMENT_ERROR (play_base_bin, RESOURCE, FAILED,
+        (_("RTSP streams cannot be played yet.")), (NULL));
+    return NULL;
+  }
+
   if (play_base_bin->suburi) {
     /* subtitle specified */
     *subbin = setup_subtitle (play_base_bin, play_base_bin->suburi);
@@ -1097,11 +1281,15 @@ gen_source_element (GstPlayBaseBin * play_base_bin, GstElement ** subbin)
   if (!source)
     return NULL;
 
-  /* lame - FIXME, maybe we can use seek_types/mask here? */
-  play_base_bin->is_stream = !strncmp (play_base_bin->uri, "http://", 7) ||
-      !strncmp (play_base_bin->uri, "mms://", 6) ||
-      !strncmp (play_base_bin->uri, "rtp://", 6) ||
-      !strncmp (play_base_bin->uri, "rtsp://", 7);
+  play_base_bin->is_stream = is_stream (play_base_bin, play_base_bin->uri);
+
+  /* make HTTP sources send extra headers so we get icecast
+   * metadata in case the stream is an icecast stream */
+  if (!strncmp (play_base_bin->uri, "http://", 7) &&
+      g_object_class_find_property (G_OBJECT_GET_CLASS (source),
+          "iradio-mode")) {
+    g_object_set (source, "iradio-mode", TRUE, NULL);
+  }
 
   return source;
 }
@@ -1173,6 +1361,8 @@ setup_source (GstPlayBaseBin * play_base_bin, gchar ** new_location)
   if (!play_base_bin->need_rebuild)
     return TRUE;
 
+  GST_DEBUG_OBJECT (play_base_bin, "setup source");
+
   /* delete old src */
   if (play_base_bin->source) {
     GST_DEBUG_OBJECT (play_base_bin, "removing old src element");
@@ -1221,12 +1411,20 @@ setup_source (GstPlayBaseBin * play_base_bin, gchar ** new_location)
        * has no more dynamic streams, the cond is unlocked. We can remove
        * the signal handlers then
        */
+      GST_DEBUG_OBJECT (play_base_bin, "starting subtitle bin");
+
+      /* for subtitles in a separate bin we will not commit the
+       * current building group since we need to add the other
+       * audio/video streams to the group. We check if we managed
+       * to commit the subtitle group using an extra flag. */
+      play_base_bin->subtitle_done = FALSE;
 
       gst_element_set_state (subbin, GST_STATE_PAUSED);
 
       GROUP_LOCK (play_base_bin);
-      GST_DEBUG ("waiting for first group...");
-      GROUP_WAIT (play_base_bin);
+      GST_DEBUG ("waiting for subtitle to complete...");
+      while (!play_base_bin->subtitle_done)
+        GROUP_WAIT (play_base_bin);
       GST_DEBUG ("group done !");
       GROUP_UNLOCK (play_base_bin);
 
@@ -1331,6 +1529,11 @@ setup_source (GstPlayBaseBin * play_base_bin, gchar ** new_location)
           gst_element_factory_make ("decodebin", "decoder")))
     goto no_decodebin;
 
+  g_signal_connect (play_base_bin->decoder, "element-added",
+      G_CALLBACK (decodebin_element_added_cb), play_base_bin);
+  g_signal_connect (play_base_bin->decoder, "element-removed",
+      G_CALLBACK (decodebin_element_removed_cb), play_base_bin);
+
   gst_bin_add (GST_BIN (play_base_bin), play_base_bin->decoder);
 
   if (!gst_element_link (play_base_bin->source, play_base_bin->decoder))
@@ -1696,6 +1899,34 @@ gst_play_base_bin_set_property (GObject * object, guint prop_id,
           GST_STREAM_TYPE_TEXT, g_value_get_int (value));
       GROUP_UNLOCK (play_base_bin);
       break;
+    case ARG_SUBTITLE_ENCODING:
+    {
+      const gchar *encoding;
+      GSList *list;
+
+      encoding = g_value_get_string (value);
+      if (encoding && play_base_bin->subencoding &&
+          !strcmp (play_base_bin->subencoding, encoding)) {
+        return;
+      }
+      if (encoding == NULL && play_base_bin->subencoding == NULL)
+        return;
+
+      g_mutex_lock (play_base_bin->sub_lock);
+      g_free (play_base_bin->subencoding);
+      play_base_bin->subencoding = g_strdup (encoding);
+      list = g_slist_copy (play_base_bin->subtitle_elements);
+      g_slist_foreach (list, (GFunc) gst_object_ref, NULL);
+      g_mutex_unlock (play_base_bin->sub_lock);
+
+      /* we can't hold a lock when calling g_object_set() on a child, since
+       * the notify event will trigger GstObject to send a deep-notify event
+       * which will try to take the lock ... */
+      g_slist_foreach (list, (GFunc) set_encoding_element, (gpointer) encoding);
+      g_slist_foreach (list, (GFunc) gst_object_unref, NULL);
+      g_slist_free (list);
+      break;
+    }
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
       break;
@@ -1744,8 +1975,13 @@ gst_play_base_bin_get_property (GObject * object, guint prop_id, GValue * value,
       g_value_set_pointer (value,
           (gpointer) gst_play_base_bin_get_streaminfo (play_base_bin));
       break;
+    case ARG_STREAMINFO_VALUES:{
+      g_value_set_boxed (value,
+          gst_play_base_bin_get_streaminfo_value_array (play_base_bin));
+      break;
+    }
     case ARG_SOURCE:
-      _gst_gvalue_set_gstobject (value, play_base_bin->source);
+      g_value_set_object (value, play_base_bin->source);
       break;
     case ARG_VIDEO:
       GROUP_LOCK (play_base_bin);
@@ -1765,6 +2001,11 @@ gst_play_base_bin_get_property (GObject * object, guint prop_id, GValue * value,
               GST_STREAM_TYPE_TEXT));
       GROUP_UNLOCK (play_base_bin);
       break;
+    case ARG_SUBTITLE_ENCODING:
+      GST_OBJECT_LOCK (play_base_bin);
+      g_value_set_string (value, play_base_bin->subencoding);
+      GST_OBJECT_UNLOCK (play_base_bin);
+      break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
       break;
@@ -1830,3 +2071,16 @@ gst_play_base_bin_get_streaminfo (GstPlayBaseBin * play_base_bin)
   }
   return info;
 }
+
+const GValueArray *
+gst_play_base_bin_get_streaminfo_value_array (GstPlayBaseBin * play_base_bin)
+{
+  GstPlayBaseGroup *group = get_active_group (play_base_bin);
+  GValueArray *array = NULL;
+
+  if (group) {
+    array = group->streaminfo_value_array;
+  }
+
+  return array;
+}
