@@ -17,6 +17,220 @@
  * Boston, MA 02111-1307, USA.
  */
 
+/**
+ * SECTION:element-playbin
+ *
+ * <refsect2>
+ * <para>
+ * Playbin provides a stand-alone everything-in-one abstraction for an
+ * audio and/or video player.
+ * </para>
+ * <para>
+ * It can handle both audio and video files and features 
+ * <itemizedlist>
+ * <listitem>
+ * automatic file type recognition and based on that automatic
+ * selection and usage of the right audio/video/subtitle demuxers/decoders
+ * </listitem>
+ * <listitem>
+ * visualisations for audio files
+ * </listitem>
+ * <listitem>
+ * subtitle support for video files
+ * </listitem>
+ * <listitem>
+ * stream selection between different audio/subtitles streams
+ * </listitem>
+ * <listitem>
+ * meta info (tag) extraction
+ * </listitem>
+ * <listitem>
+ * easy access to the last video frame
+ * </listitem>
+ * <listitem>
+ * buffering when playing streams over a network
+ * </listitem>
+ * <listitem>
+ * volume control
+ * </listitem>
+ * </itemizedlist>
+ * </para>
+ * <title>Usage</title>
+ * <para>
+ * A playbin element can be created just like any other element using
+ * gst_element_factory_make(). The file/URI to play should be set via the "uri"
+ * property. This must be an absolute URI, relative file paths are not allowed.
+ * Example URIs are file:///home/joe/movie.avi or http://www.joedoe.com/foo.ogg
+ * </para>
+ * <para>
+ * Playbin is a #GstPipeline. It will notify the application of everything
+ * that's happening (errors, end of stream, tags found, state changes, etc.)
+ * by posting messages on its #GstBus. The application needs to watch the
+ * bus.
+ * </para>
+ * <para>
+ * Playback can be initiated by setting the element to PLAYING state using
+ * gst_element_set_state(). Note that the state change will take place in
+ * the background in a separate thread, when the function returns playback
+ * is probably not happening yet and any errors might not have occured yet.
+ * Applications using playbin should ideally be written to deal with things
+ * completely asynchroneous.
+ * </para>
+ * <para>
+ * When playback has finished (an EOS message has been received on the bus)
+ * or an error has occured (an ERROR message has been received on the bus) or
+ * the user wants to play a different track, playbin should be set back to
+ * READY or NULL state, then the "uri" property should be set to the new
+ * location and then playbin be set to PLAYING state again.
+ * </para>
+ * <para>
+ * Seeking can be done using gst_element_seek_simple() or gst_element_seek()
+ * on the playbin element. Again, the seek will not be executed
+ * instantaneously, but will be done in a background thread. When the seek
+ * call returns the seek will most likely still be in process. An application
+ * may wait for the seek to finish (or fail) using gst_element_get_state() with
+ * -1 as the timeout, but this will block the user interface and is not
+ * recommended at all.
+ * </para>
+ * <para>
+ * Applications may query the current position and duration of the stream
+ * via gst_element_query_position() and gst_element_query_duration() and
+ * setting the format passed to GST_FORMAT_TIME. If the query was successful,
+ * the duration or position will have been returned in units of nanoseconds.
+ * </para>
+ * <title>Advanced Usage: specifying the audio and video sink</title>
+ * <para>
+ * By default, if no audio sink or video sink has been specified via the
+ * "audio-sink" or "video-sink" property, playbin will use the autoaudiosink
+ * and autovideosink elements to find the first-best available output method.
+ * This should work in most cases, but is not always desirable. Often either
+ * the user or application might want to specify more explicitly what to use
+ * for audio and video output.
+ * </para>
+ * <para>
+ * If the application wants more control over how audio or video should be
+ * output, it may create the audio/video sink elements itself (for example
+ * using gst_element_factory_make()) and provide them to playbin using the
+ * "audio-sink" or "video-sink" property.
+ * </para>
+ * <para>
+ * GNOME-based applications, for example, will usually want to create
+ * gconfaudiosink and gconfvideosink elements and make playbin use those,
+ * so that output happens to whatever the user has configured in the GNOME
+ * Multimedia System Selector confinguration dialog.
+ * </para>
+ * <para>
+ * The sink elements do not necessarily need to be ready-made sinks. It is
+ * possible to create container elements that look like a sink to playbin,
+ * but in reality contain a number of custom elements linked together. This
+ * can be achieved by creating a #GstBin and putting elements in there and
+ * linking them, and then creating a sink #GstGhostPad for the bin and pointing
+ * it to the sink pad of the first element within the bin. This can be used
+ * for a number of purposes, for example to force output to a particular
+ * format or to modify or observe the data before it is output.
+ * </para>
+ * <para>
+ * It is also possible to 'suppress' audio and/or video output by using
+ * 'fakesink' elements (or capture it from there using the fakesink element's
+ * "handoff" signal, which, nota bene, is fired from the streaming thread!).
+ * </para>
+ * <title>Retrieving Tags and Other Meta Data</title>
+ * <para>
+ * Most of the common meta data (artist, title, etc.) can be retrieved by
+ * watching for TAG messages on the pipeline's bus (see above).
+ * </para>
+ * <para>
+ * Other more specific meta information like width/height/framerate of video
+ * streams or samplerate/number of channels of audio streams can be obtained
+ * using the "stream-info" property, which will return a GList of stream info
+ * objects, one for each stream. These are opaque objects that can only be
+ * accessed via the standard GObject property interface, ie. g_object_get().
+ * Each stream info object has the following properties:
+ * <itemizedlist>
+ * <listitem>"object" (GstObject) (the decoder source pad usually)</listitem>
+ * <listitem>"type" (enum) (if this is an audio/video/subtitle stream)</listitem>
+ * <listitem>"decoder" (string) (name of decoder used to decode this stream)</listitem>
+ * <listitem>"mute" (boolean) (to mute or unmute this stream)</listitem>
+ * <listitem>"caps" (GstCaps) (caps of the decoded stream)</listitem>
+ * <listitem>"language-code" (string) (ISO-639 language code for this stream, mostly used for audio/subtitle streams)</listitem>
+ * <listitem>"codec" (string) (format this stream was encoded in)</listitem>
+ * </itemizedlist>
+ * Stream information from the stream-info properties is best queried once
+ * playbin has changed into PAUSED or PLAYING state (which can be detected
+ * via a state-changed message on the bus where old_state=READY and
+ * new_state=PAUSED), since before that the list might not be complete yet or
+ * not contain all available information (like language-codes).
+ * </para>
+ * <title>Buffering status</title>
+ * <para>
+ * Playbin handles buffering automatically. When playbin is buffering, it
+ * will post BUFFERING messages on the bus with a percentage value that
+ * shows the progress of the buffering. Applications may want to show this
+ * information to the user in some form or another. Here is how to extract
+ * the percentage information from the message:
+ * </para>
+ * <para>
+ * <programlisting>
+ * switch (GST_MESSAGE_TYPE (msg)) {
+ *   case GST_MESSAGE_BUFFERING: {
+ *     gint percent = 0;
+ *     gst_structure_get_int (msg-&gt;structure, "buffer-percent", &amp;percent);
+ *     g_print ("Buffering (%%u percent done)", percent);
+ *     break;
+ *   }
+ *   ...
+ * }
+ * </programlisting>
+ * </para>
+ * <title>Embedding the video window in your application</title>
+ * <para>
+ * By default, playbin (or rather the video sinks used) will create their own
+ * window. Applications will usually want to force output to a window of their
+ * own, however. This can be done using the GstXOverlay interface, which most
+ * video sinks implement. See the documentation there for more details.
+ * </para>
+ * <title>Specifying which CD/DVD device to use</title>
+ * <para>
+ * The device to use for CDs/DVDs needs to be set on the source element
+ * playbin creates before it is opened. The only way to do this at the moment
+ * is to connect to playbin's "notify::source" signal, which will be emitted
+ * by playbin when it has created the source element for a particular URI.
+ * In the signal callback you can check if the source element has a "device"
+ * property and set it appropriately. In future ways might be added to specify
+ * the device as part of the URI, but at the time of writing this is not
+ * possible yet.
+ * </para>
+ * <title>Examples</title>
+ * <para>
+ * Here is a simple pipeline to play back a video or audio file:
+ * <programlisting>
+ * gst-launch -v playbin uri=file:///path/to/somefile.avi
+ * </programlisting>
+ * This will play back the given AVI video file, given that the video and
+ * audio decoders required to decode the content are installed. Since no
+ * special audio sink or video sink is supplied (not possible via gst-launch),
+ * playbin will try to find a suitable audio and video sink automatically
+ * using the autoaudiosink and autovideosink elements.
+ * </para>
+ * <para>
+ * Here is a another pipeline to play track 4 of an audio CD:
+ * <programlisting>
+ * gst-launch -v playbin uri=cdda://4
+ * </programlisting>
+ * This will play back track 4 on an audio CD in your disc drive (assuming
+ * the drive is detected automatically by the plugin).
+ * </para>
+ * <para>
+ * Here is a another pipeline to play title 1 of a DVD:
+ * <programlisting>
+ * gst-launch -v playbin uri=dvd://1
+ * </programlisting>
+ * This will play back title 1 of a DVD in your disc drive (assuming
+ * the drive is detected automatically by the plugin).
+ * </para>
+ * </refsect2>
+ */
+
 #ifdef HAVE_CONFIG_H
 #include "config.h"
 #endif
@@ -37,21 +251,8 @@ GST_DEBUG_CATEGORY_STATIC (gst_play_bin_debug);
 #define GST_IS_PLAY_BIN(obj)            (G_TYPE_CHECK_INSTANCE_TYPE((obj),GST_TYPE_PLAY_BIN))
 #define GST_IS_PLAY_BIN_CLASS(klass)    (G_TYPE_CHECK_CLASS_TYPE((klass),GST_TYPE_PLAY_BIN))
 
-#define VOLUME_MAX_DOUBLE 4.0
-
-#ifndef GST_HAVE_GLIB_2_8
-#define _gst_gvalue_set_gstobject(gvalue,obj)  \
-      if (obj != NULL) {                       \
-        gst_object_ref (obj);                  \
-        g_value_set_object (gvalue, obj);      \
-        g_object_unref (obj);                  \
-      } else {                                 \
-        g_value_set_object (gvalue, NULL);     \
-      }
-#else
-#define _gst_gvalue_set_gstobject(gvalue,obj)  \
-      g_value_set_object (gvalue, obj);
-#endif
+#define VOLUME_MAX_DOUBLE 10.0
+#define CONNECTION_SPEED_DEFAULT 0
 
 typedef struct _GstPlayBin GstPlayBin;
 typedef struct _GstPlayBinClass GstPlayBinClass;
@@ -81,6 +282,9 @@ struct _GstPlayBin
 
   /* font description */
   gchar *font_desc;
+
+  /* connection speed in bits/sec (0 = unknown) */
+  guint connection_speed;
 };
 
 struct _GstPlayBinClass
@@ -97,7 +301,8 @@ enum
   ARG_VIS_PLUGIN,
   ARG_VOLUME,
   ARG_FRAME,
-  ARG_FONT_DESC
+  ARG_FONT_DESC,
+  ARG_CONNECTION_SPEED
 };
 
 /* signals */
@@ -123,12 +328,13 @@ static gboolean gst_play_bin_send_event (GstElement * element,
     GstEvent * event);
 static GstStateChangeReturn gst_play_bin_change_state (GstElement * element,
     GstStateChange transition);
+static void gst_play_bin_handle_message (GstBin * bin, GstMessage * message);
 
 static GstElementClass *parent_class;
 
 //static guint gst_play_bin_signals[LAST_SIGNAL] = { 0 };
 
-static GstElementDetails gst_play_bin_details =
+static const GstElementDetails gst_play_bin_details =
 GST_ELEMENT_DETAILS ("Player Bin",
     "Generic/Bin/Player",
     "Autoplug and play media from an uri",
@@ -202,6 +408,17 @@ gst_play_bin_class_init (GstPlayBinClass * klass)
           "Subtitle font description",
           "Pango font description of font "
           "to be used for subtitle rendering", NULL, G_PARAM_WRITABLE));
+  /**
+   * GstPlayBin::connection-speed
+   *
+   * Network connection speed in kbps (0 = unknown)
+   *
+   * Since: 0.10.10
+   **/
+  g_object_class_install_property (gobject_klass, ARG_CONNECTION_SPEED,
+      g_param_spec_uint ("connection-speed", "Connection Speed",
+          "Network connection speed in kbps (0 = unknown)",
+          0, G_MAXUINT, CONNECTION_SPEED_DEFAULT, G_PARAM_READWRITE));
 
   gobject_klass->dispose = GST_DEBUG_FUNCPTR (gst_play_bin_dispose);
 
@@ -211,6 +428,9 @@ gst_play_bin_class_init (GstPlayBinClass * klass)
       GST_DEBUG_FUNCPTR (gst_play_bin_change_state);
   gstelement_klass->send_event = GST_DEBUG_FUNCPTR (gst_play_bin_send_event);
 
+  gstbin_klass->handle_message =
+      GST_DEBUG_FUNCPTR (gst_play_bin_handle_message);
+
   playbasebin_klass->setup_output_pads = setup_sinks;
 }
 
@@ -390,8 +610,6 @@ gst_play_bin_set_property (GObject * object, guint prop_id,
 {
   GstPlayBin *play_bin;
 
-  g_return_if_fail (GST_IS_PLAY_BIN (object));
-
   play_bin = GST_PLAY_BIN (object);
 
   switch (prop_id) {
@@ -500,6 +718,9 @@ gst_play_bin_set_property (GObject * object, guint prop_id,
             "font-desc", g_value_get_string (value), NULL);
       }
       break;
+    case ARG_CONNECTION_SPEED:
+      play_bin->connection_speed = g_value_get_uint (value) * 1000;
+      break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
       break;
@@ -512,19 +733,17 @@ gst_play_bin_get_property (GObject * object, guint prop_id, GValue * value,
 {
   GstPlayBin *play_bin;
 
-  g_return_if_fail (GST_IS_PLAY_BIN (object));
-
   play_bin = GST_PLAY_BIN (object);
 
   switch (prop_id) {
     case ARG_VIDEO_SINK:
-      _gst_gvalue_set_gstobject (value, play_bin->video_sink);
+      g_value_set_object (value, play_bin->video_sink);
       break;
     case ARG_AUDIO_SINK:
-      _gst_gvalue_set_gstobject (value, play_bin->audio_sink);
+      g_value_set_object (value, play_bin->audio_sink);
       break;
     case ARG_VIS_PLUGIN:
-      _gst_gvalue_set_gstobject (value, play_bin->visualisation);
+      g_value_set_object (value, play_bin->visualisation);
       break;
     case ARG_VOLUME:
       g_value_set_double (value, play_bin->volume);
@@ -532,6 +751,9 @@ gst_play_bin_get_property (GObject * object, guint prop_id, GValue * value,
     case ARG_FRAME:
       gst_value_set_mini_object (value, GST_MINI_OBJECT (play_bin->frame));
       break;
+    case ARG_CONNECTION_SPEED:
+      g_value_set_uint (value, play_bin->connection_speed / 1000);
+      break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
       break;
@@ -751,10 +973,9 @@ gen_audio_element (GstPlayBin * play_bin)
 {
   GstElement *element;
   GstElement *conv;
+  GstElement *scale;
   GstElement *sink;
   GstElement *volume;
-
-  /* GstElement *scale; */
   GstPad *pad;
 
   element = g_hash_table_lookup (play_bin->cache, "abin");
@@ -765,11 +986,11 @@ gen_audio_element (GstPlayBin * play_bin)
   conv = gst_element_factory_make ("audioconvert", "aconv");
   if (conv == NULL)
     goto no_audioconvert;
-/*
+
   scale = gst_element_factory_make ("audioresample", "aresample");
   if (scale == NULL)
     goto no_audioresample;
-*/
+
   volume = gst_element_factory_make ("volume", "volume");
   g_object_set (G_OBJECT (volume), "volume", play_bin->volume, NULL);
   play_bin->volume_element = volume;
@@ -793,12 +1014,12 @@ gen_audio_element (GstPlayBin * play_bin)
   g_hash_table_insert (play_bin->cache, "audio_sink", sink);
 
   gst_bin_add (GST_BIN (element), conv);
-  /* gst_bin_add (GST_BIN (element), scale); */
+  gst_bin_add (GST_BIN (element), scale);
   gst_bin_add (GST_BIN (element), volume);
   gst_bin_add (GST_BIN (element), sink);
 
-  gst_element_link_pads (conv, "src",   /*scale, "sink");
-                                           gst_element_link_pads (scale, "src", */ volume, "sink");
+  gst_element_link_pads (conv, "src", scale, "sink");
+  gst_element_link_pads (scale, "src", volume, "sink");
   gst_element_link_pads (volume, "src", sink, "sink");
 
   pad = gst_element_get_pad (conv, "sink");
@@ -822,17 +1043,15 @@ no_audioconvert:
     gst_object_unref (element);
     return NULL;
   }
-/*
+
 no_audioresample:
   {
     GST_ELEMENT_ERROR (play_bin, CORE, MISSING_PLUGIN,
         (_("Missing element '%s' - check your GStreamer installation."),
-         "audioresample"),
-        ("possibly a liboil version mismatch?"));
+            "audioresample"), ("possibly a liboil version mismatch?"));
     gst_object_unref (element);
     return NULL;
   }
-*/
 }
 
 /* make the element (bin) that contains the elements needed to perform
@@ -1109,7 +1328,7 @@ link_failed:
     GST_DEBUG_OBJECT (play_bin,
         "link failed when adding sink, caps %s, reason %d", capsstr, linkres);
     g_free (capsstr);
-    g_free (caps);
+    gst_caps_unref (caps);
 
     gst_element_set_state (sink, GST_STATE_NULL);
     gst_bin_remove (GST_BIN (play_bin), sink);
@@ -1127,7 +1346,7 @@ subtitle_failed:
         "subtitle link failed when adding sink, caps %s, reason %d", capsstr,
         linkres);
     g_free (capsstr);
-    g_free (caps);
+    gst_caps_unref (caps);
 
     return TRUE;
   }
@@ -1366,6 +1585,104 @@ gst_play_bin_send_event (GstElement * element, GstEvent * event)
   return res;
 }
 
+static void
+value_list_append_structure_list (GValue * list_val, GstStructure ** first,
+    GList * structure_list)
+{
+  GList *l;
+
+  for (l = structure_list; l != NULL; l = l->next) {
+    GValue val = { 0, };
+
+    if (*first == NULL)
+      *first = gst_structure_copy ((GstStructure *) l->data);
+
+    g_value_init (&val, GST_TYPE_STRUCTURE);
+    g_value_take_boxed (&val, gst_structure_copy ((GstStructure *) l->data));
+    gst_value_list_append_value (list_val, &val);
+    g_value_unset (&val);
+  }
+}
+
+/* if it's a redirect message with multiple redirect locations we might
+ * want to pick a different 'best' location depending on the required
+ * bitrates and the connection speed */
+static GstMessage *
+gst_play_bin_handle_redirect_message (GstPlayBin * playbin, GstMessage * msg)
+{
+  const GValue *locations_list, *location_val;
+  GstMessage *new_msg;
+  GstStructure *new_structure = NULL;
+  GList *l_good = NULL, *l_neutral = NULL, *l_bad = NULL;
+  GValue new_list = { 0, };
+  guint size, i;
+
+  GST_DEBUG_OBJECT (playbin, "redirect message: %" GST_PTR_FORMAT, msg);
+  GST_DEBUG_OBJECT (playbin, "connection speed: %u", playbin->connection_speed);
+
+  if (playbin->connection_speed == 0 || msg->structure == NULL)
+    return msg;
+
+  locations_list = gst_structure_get_value (msg->structure, "locations");
+  if (locations_list == NULL)
+    return msg;
+
+  size = gst_value_list_get_size (locations_list);
+  if (size < 2)
+    return msg;
+
+  /* maintain existing order as much as possible, just sort references
+   * with too high a bitrate to the end (the assumption being that if
+   * bitrates are given they are given for all interesting streams and
+   * that the you-need-at-least-version-xyz redirect has the same bitrate
+   * as the lowest referenced redirect alternative) */
+  for (i = 0; i < size; ++i) {
+    const GstStructure *s;
+    gint bitrate = 0;
+
+    location_val = gst_value_list_get_value (locations_list, i);
+    s = (const GstStructure *) g_value_get_boxed (location_val);
+    if (!gst_structure_get_int (s, "minimum-bitrate", &bitrate) || bitrate <= 0) {
+      GST_DEBUG_OBJECT (playbin, "no bitrate: %" GST_PTR_FORMAT, s);
+      l_neutral = g_list_append (l_neutral, (gpointer) s);
+    } else if (bitrate > playbin->connection_speed) {
+      GST_DEBUG_OBJECT (playbin, "bitrate too high: %" GST_PTR_FORMAT, s);
+      l_bad = g_list_append (l_bad, (gpointer) s);
+    } else if (bitrate <= playbin->connection_speed) {
+      GST_DEBUG_OBJECT (playbin, "bitrate OK: %" GST_PTR_FORMAT, s);
+      l_good = g_list_append (l_good, (gpointer) s);
+    }
+  }
+
+  g_value_init (&new_list, GST_TYPE_LIST);
+  value_list_append_structure_list (&new_list, &new_structure, l_good);
+  value_list_append_structure_list (&new_list, &new_structure, l_neutral);
+  value_list_append_structure_list (&new_list, &new_structure, l_bad);
+  gst_structure_set_value (new_structure, "locations", &new_list);
+  g_value_unset (&new_list);
+
+  g_list_free (l_good);
+  g_list_free (l_neutral);
+  g_list_free (l_bad);
+
+  new_msg = gst_message_new_element (msg->src, new_structure);
+  gst_message_unref (msg);
+
+  GST_DEBUG_OBJECT (playbin, "new redirect message: %" GST_PTR_FORMAT, new_msg);
+  return new_msg;
+}
+
+static void
+gst_play_bin_handle_message (GstBin * bin, GstMessage * msg)
+{
+  if (GST_MESSAGE_TYPE (msg) == GST_MESSAGE_ELEMENT && msg->structure != NULL
+      && gst_structure_has_name (msg->structure, "redirect")) {
+    msg = gst_play_bin_handle_redirect_message (GST_PLAY_BIN (bin), msg);
+  }
+
+  GST_BIN_CLASS (parent_class)->handle_message (bin, msg);
+}
+
 static GstStateChangeReturn
 gst_play_bin_change_state (GstElement * element, GstStateChange transition)
 {
