@@ -20,7 +20,6 @@
 
 /**
  * SECTION:element-videoscale
- * @short_description: rescale video
  * @see_also: videorate, ffmpegcolorspace
  *
  * <refsect2>
@@ -31,24 +30,26 @@
  * get more robust behaviour without any cost if no scaling is needed.
  * </para>
  * <para>
- * This element supports a wide range of color spaces including various YUV and 
- * RGB formats and is therefore generally able to operate anywhere in a pipeline.
+ * This element supports a wide range of color spaces including various YUV and
+ * RGB formats and is therefore generally able to operate anywhere in a
+ * pipeline.
  * </para>
  * <title>Example pipelines</title>
  * <para>
  * <programlisting>
  * gst-launch -v filesrc location=videotestsrc.ogg ! oggdemux ! theoradec ! ffmpegcolorspace ! videoscale ! ximagesink
  * </programlisting>
- * Decode an Ogg/Theora and display the video using ximagesink. Since ximagesink cannot
- * perform scaling, the video scaling will be performed by videoscale when you resize the
- * video window.
+ * Decode an Ogg/Theora and display the video using ximagesink. Since
+ * ximagesink cannot perform scaling, the video scaling will be performed by
+ * videoscale when you resize the video window.
  * To create the test Ogg/Theora file refer to the documentation of theoraenc.
  * </para>
  * <para>
  * <programlisting>
  * gst-launch -v filesrc location=videotestsrc.ogg ! oggdemux ! theoradec ! videoscale ! video/x-raw-yuv, width=50 ! xvimagesink
  * </programlisting>
- * Decode an Ogg/Theora and display the video using xvimagesink with a width of 50.
+ * Decode an Ogg/Theora and display the video using xvimagesink with a width of
+ * 50.
  * </para>
  * </refsect2>
  *
@@ -72,7 +73,7 @@
 GST_DEBUG_CATEGORY (video_scale_debug);
 
 /* elementfactory information */
-static GstElementDetails video_scale_details =
+static const GstElementDetails video_scale_details =
 GST_ELEMENT_DETAILS ("Video scaler",
     "Filter/Effect/Video",
     "Resizes video",
@@ -92,6 +93,10 @@ static GstStaticCaps gst_video_scale_format_caps[] = {
   GST_STATIC_CAPS (GST_VIDEO_CAPS_xRGB),
   GST_STATIC_CAPS (GST_VIDEO_CAPS_BGRx),
   GST_STATIC_CAPS (GST_VIDEO_CAPS_xBGR),
+  GST_STATIC_CAPS (GST_VIDEO_CAPS_RGBA),
+  GST_STATIC_CAPS (GST_VIDEO_CAPS_ARGB),
+  GST_STATIC_CAPS (GST_VIDEO_CAPS_BGRA),
+  GST_STATIC_CAPS (GST_VIDEO_CAPS_ABGR),
   GST_STATIC_CAPS (GST_VIDEO_CAPS_RGB),
   GST_STATIC_CAPS (GST_VIDEO_CAPS_BGR),
   GST_STATIC_CAPS (GST_VIDEO_CAPS_YUV ("AYUV")),
@@ -111,6 +116,10 @@ enum
   GST_VIDEO_SCALE_xRGB,
   GST_VIDEO_SCALE_BGRx,
   GST_VIDEO_SCALE_xBGR,
+  GST_VIDEO_SCALE_RGBA,
+  GST_VIDEO_SCALE_ARGB,
+  GST_VIDEO_SCALE_BGRA,
+  GST_VIDEO_SCALE_ABGR,
   GST_VIDEO_SCALE_RGB,
   GST_VIDEO_SCALE_BGR,
   GST_VIDEO_SCALE_AYUV,
@@ -129,7 +138,7 @@ static GType
 gst_video_scale_method_get_type (void)
 {
   static GType video_scale_method_type = 0;
-  static GEnumValue video_scale_methods[] = {
+  static const GEnumValue video_scale_methods[] = {
     {GST_VIDEO_SCALE_NEAREST, "Nearest Neighbour", "nearest-neighbour"},
     {GST_VIDEO_SCALE_BILINEAR, "Bilinear", "bilinear"},
     {0, NULL, NULL},
@@ -178,6 +187,7 @@ gst_video_scale_sink_template_factory (void)
 static void gst_video_scale_base_init (gpointer g_class);
 static void gst_video_scale_class_init (GstVideoScaleClass * klass);
 static void gst_video_scale_init (GstVideoScale * videoscale);
+static void gst_video_scale_finalize (GstVideoScale * videoscale);
 static gboolean gst_video_scale_src_event (GstBaseTransform * trans,
     GstEvent * event);
 
@@ -248,10 +258,11 @@ gst_video_scale_class_init (GstVideoScaleClass * klass)
   gobject_class = (GObjectClass *) klass;
   trans_class = (GstBaseTransformClass *) klass;
 
+  gobject_class->finalize = (GObjectFinalizeFunc) gst_video_scale_finalize;
   gobject_class->set_property = gst_video_scale_set_property;
   gobject_class->get_property = gst_video_scale_get_property;
 
-  g_object_class_install_property (G_OBJECT_CLASS (klass), PROP_METHOD,
+  g_object_class_install_property (gobject_class, PROP_METHOD,
       g_param_spec_enum ("method", "method", "method",
           GST_TYPE_VIDEO_SCALE_METHOD, DEFAULT_PROP_METHOD, G_PARAM_READWRITE));
 
@@ -277,6 +288,14 @@ gst_video_scale_init (GstVideoScale * videoscale)
   videoscale->method = DEFAULT_PROP_METHOD;
 }
 
+static void
+gst_video_scale_finalize (GstVideoScale * videoscale)
+{
+  if (videoscale->tmp_buf)
+    g_free (videoscale->tmp_buf);
+
+  G_OBJECT_CLASS (parent_class)->finalize (G_OBJECT (videoscale));
+}
 
 static void
 gst_video_scale_set_property (GObject * object, guint prop_id,
@@ -316,22 +335,42 @@ gst_video_scale_transform_caps (GstBaseTransform * trans,
 {
   GstVideoScale *videoscale;
   GstCaps *ret;
-  int i;
+  GstStructure *structure;
+  const GValue *par;
 
   videoscale = GST_VIDEO_SCALE (trans);
 
   ret = gst_caps_copy (caps);
 
-  for (i = 0; i < gst_caps_get_size (ret); i++) {
-    GstStructure *structure = gst_caps_get_structure (ret, i);
+  /* this function is always called with a simple caps */
+  g_return_val_if_fail (GST_CAPS_IS_SIMPLE (ret), NULL);
 
-    gst_structure_set (structure,
-        "width", GST_TYPE_INT_RANGE, 16, 4096,
-        "height", GST_TYPE_INT_RANGE, 16, 4096, NULL);
-    gst_structure_remove_field (structure, "pixel-aspect-ratio");
+  structure = gst_caps_get_structure (ret, 0);
+
+  gst_structure_set (structure,
+      "width", GST_TYPE_INT_RANGE, 16, 4096,
+      "height", GST_TYPE_INT_RANGE, 16, 4096, NULL);
+
+  /* if pixel aspect ratio, make a range of it */
+  if ((par = gst_structure_get_value (structure, "pixel-aspect-ratio"))) {
+    GstCaps *copy;
+    GstStructure *cstruct;
+
+    /* copy input PAR first, this is the prefered PAR */
+    gst_structure_set_value (structure, "pixel-aspect-ratio", par);
+
+    /* then make a copy with a fraction range as a second choice */
+    copy = gst_caps_copy (ret);
+    cstruct = gst_caps_get_structure (copy, 0);
+    gst_structure_set (cstruct,
+        "pixel-aspect-ratio", GST_TYPE_FRACTION_RANGE, 0, 1, G_MAXINT, 1, NULL);
+
+    /* and append */
+    gst_caps_append (ret, copy);
   }
 
   GST_DEBUG_OBJECT (trans, "returning caps: %" GST_PTR_FORMAT, ret);
+
   return ret;
 }
 
@@ -369,6 +408,10 @@ gst_video_scale_prepare_size (GstVideoScale * videoscale, gint format,
     case GST_VIDEO_SCALE_xRGB:
     case GST_VIDEO_SCALE_BGRx:
     case GST_VIDEO_SCALE_xBGR:
+    case GST_VIDEO_SCALE_RGBA:
+    case GST_VIDEO_SCALE_ARGB:
+    case GST_VIDEO_SCALE_BGRA:
+    case GST_VIDEO_SCALE_ABGR:
     case GST_VIDEO_SCALE_AYUV:
       img->stride = img->width * 4;
       *size = img->stride * img->height;
@@ -489,7 +532,7 @@ gst_video_scale_get_unit_size (GstBaseTransform * trans, GstCaps * caps,
   gint format, width, height;
   VSImage img;
 
-  g_return_val_if_fail (size, FALSE);
+  g_assert (size);
 
   videoscale = GST_VIDEO_SCALE (trans);
 
@@ -521,10 +564,28 @@ gst_video_scale_fixate_caps (GstBaseTransform * base, GstPadDirection direction,
   from_par = gst_structure_get_value (ins, "pixel-aspect-ratio");
   to_par = gst_structure_get_value (outs, "pixel-aspect-ratio");
 
+  /* we have both PAR but they might not be fixated */
   if (from_par && to_par) {
-    GValue to_ratio = { 0, };   /* w/h of output video */
     gint from_w, from_h, from_par_n, from_par_d, to_par_n, to_par_d;
-    gint count = 0, w = 0, h = 0, num, den;
+    gint count = 0, w = 0, h = 0;
+    guint num, den;
+
+    /* from_par should be fixed */
+    g_return_if_fail (gst_value_is_fixed (from_par));
+
+    from_par_n = gst_value_get_fraction_numerator (from_par);
+    from_par_d = gst_value_get_fraction_denominator (from_par);
+
+    /* fixate the out PAR */
+    if (!gst_value_is_fixed (to_par)) {
+      GST_DEBUG_OBJECT (base, "fixating to_par to %dx%d", from_par_n,
+          from_par_d);
+      gst_structure_fixate_field_nearest_fraction (outs, "pixel-aspect-ratio",
+          from_par_n, from_par_d);
+    }
+
+    to_par_n = gst_value_get_fraction_numerator (to_par);
+    to_par_d = gst_value_get_fraction_denominator (to_par);
 
     /* if both width and height are already fixed, we can't do anything
      * about it anymore */
@@ -540,16 +601,14 @@ gst_video_scale_fixate_caps (GstBaseTransform * base, GstPadDirection direction,
 
     gst_structure_get_int (ins, "width", &from_w);
     gst_structure_get_int (ins, "height", &from_h);
-    from_par_n = gst_value_get_fraction_numerator (from_par);
-    from_par_d = gst_value_get_fraction_denominator (from_par);
-    to_par_n = gst_value_get_fraction_numerator (to_par);
-    to_par_d = gst_value_get_fraction_denominator (to_par);
 
-    g_value_init (&to_ratio, GST_TYPE_FRACTION);
-    gst_value_set_fraction (&to_ratio, from_w * from_par_n * to_par_d,
-        from_h * from_par_d * to_par_n);
-    num = gst_value_get_fraction_numerator (&to_ratio);
-    den = gst_value_get_fraction_denominator (&to_ratio);
+    if (!gst_video_calculate_display_ratio (&num, &den, from_w, from_h,
+            from_par_n, from_par_d, to_par_n, to_par_d)) {
+      GST_ELEMENT_ERROR (base, CORE, NEGOTIATION, (NULL),
+          ("Error calculating the output scaled size - integer overflow"));
+      return;
+    }
+
     GST_DEBUG_OBJECT (base,
         "scaling input with %dx%d and PAR %d/%d to output PAR %d/%d",
         from_w, from_h, from_par_n, from_par_d, to_par_n, to_par_d);
@@ -560,20 +619,30 @@ gst_video_scale_fixate_caps (GstBaseTransform * base, GstPadDirection direction,
      * prefer those that have one of w/h the same as the incoming video
      * using wd / hd = num / den */
 
-    /* start with same height, because of interlaced video */
-    /* check hd / den is an integer scale factor, and scale wd with the PAR */
-    if (from_h % den == 0) {
-      GST_DEBUG_OBJECT (base, "keeping video height");
-      h = from_h;
-      w = h * num / den;
-    } else if (from_w % num == 0) {
-      GST_DEBUG_OBJECT (base, "keeping video width");
-      w = from_w;
-      h = w * den / num;
+    /* if one of the output width or height is fixed, we work from there */
+    if (h) {
+      GST_DEBUG_OBJECT (base, "height is fixed,scaling width");
+      w = (guint) gst_util_uint64_scale_int (h, num, den);
+    } else if (w) {
+      GST_DEBUG_OBJECT (base, "width is fixed, scaling height");
+      h = (guint) gst_util_uint64_scale_int (w, den, num);
     } else {
-      GST_DEBUG_OBJECT (base, "approximating but keeping video height");
-      h = from_h;
-      w = h * num / den;
+      /* none of width or height is fixed, figure out both of them based only on
+       * the input width and height */
+      /* check hd / den is an integer scale factor, and scale wd with the PAR */
+      if (from_h % den == 0) {
+        GST_DEBUG_OBJECT (base, "keeping video height");
+        h = from_h;
+        w = (guint) gst_util_uint64_scale_int (h, num, den);
+      } else if (from_w % num == 0) {
+        GST_DEBUG_OBJECT (base, "keeping video width");
+        w = from_w;
+        h = (guint) gst_util_uint64_scale_int (w, den, num);
+      } else {
+        GST_DEBUG_OBJECT (base, "approximating but keeping video height");
+        h = from_h;
+        w = (guint) gst_util_uint64_scale_int (h, num, den);
+      }
     }
     GST_DEBUG_OBJECT (base, "scaling to %dx%d", w, h);
 
@@ -653,6 +722,10 @@ gst_video_scale_transform (GstBaseTransform * trans, GstBuffer * in,
         case GST_VIDEO_SCALE_xRGB:
         case GST_VIDEO_SCALE_BGRx:
         case GST_VIDEO_SCALE_xBGR:
+        case GST_VIDEO_SCALE_RGBA:
+        case GST_VIDEO_SCALE_ARGB:
+        case GST_VIDEO_SCALE_BGRA:
+        case GST_VIDEO_SCALE_ABGR:
         case GST_VIDEO_SCALE_AYUV:
           vs_image_scale_nearest_RGBA (dest, src, videoscale->tmp_buf);
           break;
@@ -692,6 +765,10 @@ gst_video_scale_transform (GstBaseTransform * trans, GstBuffer * in,
         case GST_VIDEO_SCALE_xRGB:
         case GST_VIDEO_SCALE_BGRx:
         case GST_VIDEO_SCALE_xBGR:
+        case GST_VIDEO_SCALE_RGBA:
+        case GST_VIDEO_SCALE_ARGB:
+        case GST_VIDEO_SCALE_BGRA:
+        case GST_VIDEO_SCALE_ABGR:
         case GST_VIDEO_SCALE_AYUV:
           vs_image_scale_linear_RGBA (dest, src, videoscale->tmp_buf);
           break;
