@@ -49,11 +49,11 @@
 
 #define USE_PEER_BUFFERALLOC
 
-GST_DEBUG_CATEGORY (video_test_src_debug);
+GST_DEBUG_CATEGORY_STATIC (video_test_src_debug);
 #define GST_CAT_DEFAULT video_test_src_debug
 
 
-static GstElementDetails video_test_src_details =
+static const GstElementDetails video_test_src_details =
 GST_ELEMENT_DETAILS ("Video test source",
     "Source/Video",
     "Creates a test video stream",
@@ -101,11 +101,15 @@ static GType
 gst_video_test_src_pattern_get_type (void)
 {
   static GType video_test_src_pattern_type = 0;
-  static GEnumValue pattern_types[] = {
+  static const GEnumValue pattern_types[] = {
     {GST_VIDEO_TEST_SRC_SMPTE, "SMPTE 100% color bars", "smpte"},
     {GST_VIDEO_TEST_SRC_SNOW, "Random (television snow)", "snow"},
     {GST_VIDEO_TEST_SRC_BLACK, "100% Black", "black"},
-    {0, NULL, NULL},
+    {GST_VIDEO_TEST_SRC_WHITE, "100% White", "white"},
+    {GST_VIDEO_TEST_SRC_RED, "Red", "red"},
+    {GST_VIDEO_TEST_SRC_GREEN, "Green", "green"},
+    {GST_VIDEO_TEST_SRC_BLUE, "Blue", "blue"},
+    {0, NULL, NULL}
   };
 
   if (!video_test_src_pattern_type) {
@@ -141,11 +145,11 @@ gst_video_test_src_class_init (GstVideoTestSrcClass * klass)
   gobject_class->set_property = gst_video_test_src_set_property;
   gobject_class->get_property = gst_video_test_src_get_property;
 
-  g_object_class_install_property (G_OBJECT_CLASS (klass), PROP_PATTERN,
+  g_object_class_install_property (gobject_class, PROP_PATTERN,
       g_param_spec_enum ("pattern", "Pattern",
           "Type of test pattern to generate", GST_TYPE_VIDEO_TEST_SRC_PATTERN,
           1, G_PARAM_READWRITE));
-  g_object_class_install_property (G_OBJECT_CLASS (klass),
+  g_object_class_install_property (gobject_class,
       PROP_TIMESTAMP_OFFSET, g_param_spec_int64 ("timestamp-offset",
           "Timestamp offset",
           "An offset added to timestamps set on buffers (in ns)", G_MININT64,
@@ -211,6 +215,18 @@ gst_video_test_src_set_pattern (GstVideoTestSrc * videotestsrc,
     case GST_VIDEO_TEST_SRC_BLACK:
       videotestsrc->make_image = gst_video_test_src_black;
       break;
+    case GST_VIDEO_TEST_SRC_WHITE:
+      videotestsrc->make_image = gst_video_test_src_white;
+      break;
+    case GST_VIDEO_TEST_SRC_RED:
+      videotestsrc->make_image = gst_video_test_src_red;
+      break;
+    case GST_VIDEO_TEST_SRC_GREEN:
+      videotestsrc->make_image = gst_video_test_src_green;
+      break;
+    case GST_VIDEO_TEST_SRC_BLUE:
+      videotestsrc->make_image = gst_video_test_src_blue;
+      break;
     default:
       g_assert_not_reached ();
   }
@@ -504,6 +520,10 @@ gst_video_test_src_create (GstPushSrc * psrc, GstBuffer ** buffer)
   if (src->fourcc == NULL)
     goto not_negotiated;
 
+  /* 0 framerate and we are at the second frame, eos */
+  if (src->rate_numerator == 0 && src->n_frames == 1)
+    goto eos;
+
   newsize = gst_video_test_src_get_size (src, src->width, src->height);
 
   g_return_val_if_fail (newsize > 0, GST_FLOW_ERROR);
@@ -552,6 +572,11 @@ not_negotiated:
         ("format wasn't negotiated before get function"));
     return GST_FLOW_NOT_NEGOTIATED;
   }
+eos:
+  {
+    GST_DEBUG_OBJECT (src, "eos: 0 framerate and frame %d", src->n_frames);
+    return GST_FLOW_UNEXPECTED;
+  }
 no_buffer:
   {
     GST_DEBUG_OBJECT (src, "could not allocate buffer, reason %s",
