@@ -30,14 +30,14 @@
 #include <sys/ioctl.h>
 
 
-static GstElementDetails gst_v4lsrc_details =
+static const GstElementDetails gst_v4lsrc_details =
 GST_ELEMENT_DETAILS ("Video (video4linux/raw) Source",
     "Source/Video",
     "Reads raw frames from a video4linux device",
     "Ronald Bultje <rbultje@ronald.bitfreak.net>");
 
 
-GST_DEBUG_CATEGORY (v4lsrc_debug);
+GST_DEBUG_CATEGORY_STATIC (v4lsrc_debug);
 #define GST_CAT_DEFAULT v4lsrc_debug
 
 
