@@ -143,7 +143,7 @@ static void gst_ximagesink_xwindow_update_geometry (GstXImageSink * ximagesink,
 static void gst_ximagesink_expose (GstXOverlay * overlay);
 
 /* ElementFactory information */
-static GstElementDetails gst_ximagesink_details =
+static const GstElementDetails gst_ximagesink_details =
 GST_ELEMENT_DETAILS ("Video sink",
     "Sink/Video",
     "A standard X based videosink",
@@ -196,6 +196,7 @@ gst_ximage_buffer_finalize (GstXImageBuffer * ximage)
 {
   GstXImageSink *ximagesink = NULL;
   gboolean recycled = FALSE;
+  gboolean running;
 
   g_return_if_fail (ximage != NULL);
 
@@ -205,9 +206,18 @@ gst_ximage_buffer_finalize (GstXImageBuffer * ximage)
     goto beach;
   }
 
-  /* If our geometry changed we can't reuse that image. */
-  if ((ximage->width != GST_VIDEO_SINK_WIDTH (ximagesink)) ||
+  GST_OBJECT_LOCK (ximagesink);
+  running = ximagesink->running;
+  GST_OBJECT_UNLOCK (ximagesink);
+
+  if (running == FALSE) {
+    /* If the sink is shutting down, need to clear the image */
+    GST_DEBUG_OBJECT (ximagesink,
+        "destroy image %p because the sink is shutting down", ximage);
+    gst_ximagesink_ximage_destroy (ximagesink, ximage);
+  } else if ((ximage->width != GST_VIDEO_SINK_WIDTH (ximagesink)) ||
       (ximage->height != GST_VIDEO_SINK_HEIGHT (ximagesink))) {
+    /* If our geometry changed we can't reuse that image. */
     GST_DEBUG_OBJECT (ximagesink,
         "destroy image %p as its size changed %dx%d vs current %dx%d",
         ximage, ximage->width, ximage->height,
@@ -217,7 +227,7 @@ gst_ximage_buffer_finalize (GstXImageBuffer * ximage)
     /* In that case we can reuse the image and add it to our image pool. */
     GST_LOG_OBJECT (ximagesink, "recycling image %p in pool", ximage);
     /* need to increment the refcount again to recycle */
-    gst_buffer_ref (GST_BUFFER (ximage));
+    gst_buffer_ref (GST_BUFFER_CAST (ximage));
     g_mutex_lock (ximagesink->pool_lock);
     ximagesink->buffer_pool = g_slist_prepend (ximagesink->buffer_pool, ximage);
     g_mutex_unlock (ximagesink->pool_lock);
@@ -234,7 +244,7 @@ gst_ximage_buffer_free (GstXImageBuffer * ximage)
   /* make sure it is not recycled */
   ximage->width = -1;
   ximage->height = -1;
-  gst_buffer_unref (GST_BUFFER (ximage));
+  gst_buffer_unref (GST_BUFFER_CAST (ximage));
 }
 
 static void
@@ -418,8 +428,12 @@ gst_ximagesink_ximage_new (GstXImageSink * ximagesink, GstCaps * caps)
         ximagesink->xcontext->depth,
         ZPixmap, NULL, &ximage->SHMInfo, ximage->width, ximage->height);
     if (!ximage->ximage) {
-      GST_ELEMENT_ERROR (ximagesink, RESOURCE, WRITE, (NULL),
-          ("could not XShmCreateImage a %dx%d image"));
+      g_mutex_unlock (ximagesink->x_lock);
+      GST_ELEMENT_ERROR (ximagesink, RESOURCE, WRITE,
+          ("Failed to create output image buffer of %dx%d pixels",
+              ximage->width, ximage->height),
+          ("could not XShmCreateImage a %dx%d image",
+              ximage->width, ximage->height));
       goto beach;
     }
 
@@ -431,14 +445,20 @@ gst_ximagesink_ximage_new (GstXImageSink * ximagesink, GstCaps * caps)
     ximage->SHMInfo.shmid = shmget (IPC_PRIVATE, ximage->size,
         IPC_CREAT | 0777);
     if (ximage->SHMInfo.shmid == -1) {
-      GST_ELEMENT_ERROR (ximagesink, RESOURCE, WRITE, (NULL),
+      g_mutex_unlock (ximagesink->x_lock);
+      GST_ELEMENT_ERROR (ximagesink, RESOURCE, WRITE,
+          ("Failed to create output image buffer of %dx%d pixels",
+              ximage->width, ximage->height),
           ("could not get shared memory of %d bytes", ximage->size));
       goto beach;
     }
 
     ximage->SHMInfo.shmaddr = shmat (ximage->SHMInfo.shmid, 0, 0);
     if (ximage->SHMInfo.shmaddr == ((void *) -1)) {
-      GST_ELEMENT_ERROR (ximagesink, RESOURCE, WRITE, (NULL),
+      g_mutex_unlock (ximagesink->x_lock);
+      GST_ELEMENT_ERROR (ximagesink, RESOURCE, WRITE,
+          ("Failed to create output image buffer of %dx%d pixels",
+              ximage->width, ximage->height),
           ("Failed to shmat: %s", g_strerror (errno)));
       /* Clean up the shared memory segment */
       shmctl (ximage->SHMInfo.shmid, IPC_RMID, 0);
@@ -454,8 +474,10 @@ gst_ximagesink_ximage_new (GstXImageSink * ximagesink, GstCaps * caps)
     ximage->SHMInfo.readOnly = FALSE;
 
     if (XShmAttach (ximagesink->xcontext->disp, &ximage->SHMInfo) == 0) {
-      GST_ELEMENT_ERROR (ximagesink, RESOURCE, WRITE, (NULL),
-          ("Failed to XShmAttach"));
+      g_mutex_unlock (ximagesink->x_lock);
+      GST_ELEMENT_ERROR (ximagesink, RESOURCE, WRITE,
+          ("Failed to create output image buffer of %dx%d pixels",
+              ximage->width, ximage->height), ("Failed to XShmAttach"));
       goto beach;
     }
 
@@ -469,8 +491,12 @@ gst_ximagesink_ximage_new (GstXImageSink * ximagesink, GstCaps * caps)
         ZPixmap, 0, NULL,
         ximage->width, ximage->height, ximagesink->xcontext->bpp, 0);
     if (!ximage->ximage) {
-      GST_ELEMENT_ERROR (ximagesink, RESOURCE, WRITE, (NULL),
-          ("could not XCreateImage a %dx%d image"));
+      g_mutex_unlock (ximagesink->x_lock);
+      GST_ELEMENT_ERROR (ximagesink, RESOURCE, WRITE,
+          ("Failed to create output image buffer of %dx%d pixels",
+              ximage->width, ximage->height),
+          ("could not XCreateImage a %dx%d image",
+              ximage->width, ximage->height));
       goto beach;
     }
 
@@ -488,9 +514,8 @@ gst_ximagesink_ximage_new (GstXImageSink * ximagesink, GstCaps * caps)
   /* Keep a ref to our sink */
   ximage->ximagesink = gst_object_ref (ximagesink);
 
-beach:
   g_mutex_unlock (ximagesink->x_lock);
-
+beach:
   if (!succeeded) {
     gst_ximage_buffer_free (ximage);
     ximage = NULL;
@@ -512,8 +537,17 @@ gst_ximagesink_ximage_destroy (GstXImageSink * ximagesink,
     ximagesink->cur_image = NULL;
   }
 
+  /* Hold the object lock to ensure the XContext doesn't disappear */
+  GST_OBJECT_LOCK (ximagesink);
+
   /* We might have some buffers destroyed after changing state to NULL */
   if (!ximagesink->xcontext) {
+    GST_DEBUG_OBJECT (ximagesink, "Destroying XImage after XContext");
+#ifdef HAVE_XSHM
+    if (ximage->SHMInfo.shmaddr != ((void *) -1)) {
+      shmdt (ximage->SHMInfo.shmaddr);
+    }
+#endif
     goto beach;
   }
 
@@ -542,6 +576,8 @@ gst_ximagesink_ximage_destroy (GstXImageSink * ximagesink,
   g_mutex_unlock (ximagesink->x_lock);
 
 beach:
+  GST_OBJECT_UNLOCK (ximagesink);
+
   if (ximage->ximagesink) {
     /* Release the ref to our sink */
     ximage->ximagesink = NULL;
@@ -607,7 +643,7 @@ gst_ximagesink_ximage_put (GstXImageSink * ximagesink, GstXImageBuffer * ximage)
     }
     GST_LOG_OBJECT (ximagesink, "reffing %p as our current image", ximage);
     ximagesink->cur_image =
-        GST_XIMAGE_BUFFER (gst_buffer_ref (GST_BUFFER (ximage)));
+        GST_XIMAGE_BUFFER (gst_buffer_ref (GST_BUFFER_CAST (ximage)));
   }
 
   /* Expose sends a NULL image, we take the latest frame */
@@ -1022,8 +1058,8 @@ gst_ximagesink_xcontext_get (GstXImageSink * ximagesink)
   if (!xcontext->disp) {
     g_mutex_unlock (ximagesink->x_lock);
     g_free (xcontext);
-    GST_ELEMENT_ERROR (ximagesink, RESOURCE, WRITE, (NULL),
-        ("Could not open display"));
+    GST_ELEMENT_ERROR (ximagesink, RESOURCE, WRITE,
+        ("Could not initialise X output"), ("Could not open display"));
     return NULL;
   }
 
@@ -1133,8 +1169,23 @@ gst_ximagesink_xcontext_get (GstXImageSink * ximagesink)
 static void
 gst_ximagesink_xcontext_clear (GstXImageSink * ximagesink)
 {
+  GstXContext *xcontext;
+
   g_return_if_fail (GST_IS_XIMAGESINK (ximagesink));
-  g_return_if_fail (ximagesink->xcontext != NULL);
+
+  GST_OBJECT_LOCK (ximagesink);
+  if (ximagesink->xcontext == NULL) {
+    GST_OBJECT_UNLOCK (ximagesink);
+    return;
+  }
+
+  /* Take the xcontext reference and NULL it while we
+   * clean it up, so that any buffer-alloced buffers 
+   * arriving after this will be freed correctly */
+  xcontext = ximagesink->xcontext;
+  ximagesink->xcontext = NULL;
+
+  GST_OBJECT_UNLOCK (ximagesink);
 
   /* Wait for our event thread */
   if (ximagesink->event_thread) {
@@ -1142,19 +1193,18 @@ gst_ximagesink_xcontext_clear (GstXImageSink * ximagesink)
     ximagesink->event_thread = NULL;
   }
 
-  gst_caps_unref (ximagesink->xcontext->caps);
-  g_free (ximagesink->xcontext->par);
+  gst_caps_unref (xcontext->caps);
+  g_free (xcontext->par);
   g_free (ximagesink->par);
   ximagesink->par = NULL;
 
   g_mutex_lock (ximagesink->x_lock);
 
-  XCloseDisplay (ximagesink->xcontext->disp);
+  XCloseDisplay (xcontext->disp);
 
   g_mutex_unlock (ximagesink->x_lock);
 
   g_free (ximagesink->xcontext);
-  ximagesink->xcontext = NULL;
 }
 
 static void
@@ -1287,7 +1337,7 @@ gst_ximagesink_setcaps (GstBaseSink * bsink, GstCaps * caps)
           (GST_VIDEO_SINK_HEIGHT (ximagesink) != ximagesink->ximage->height))) {
     GST_DEBUG_OBJECT (ximagesink, "our image is not usable anymore, unref %p",
         ximagesink->ximage);
-    gst_buffer_unref (GST_BUFFER (ximagesink->ximage));
+    gst_buffer_unref (GST_BUFFER_CAST (ximagesink->ximage));
     ximagesink->ximage = NULL;
   }
 
@@ -1306,15 +1356,23 @@ gst_ximagesink_change_state (GstElement * element, GstStateChange transition)
 {
   GstXImageSink *ximagesink;
   GstStateChangeReturn ret = GST_STATE_CHANGE_SUCCESS;
+  GstXContext *xcontext = NULL;
 
   ximagesink = GST_XIMAGESINK (element);
 
   switch (transition) {
     case GST_STATE_CHANGE_NULL_TO_READY:
-      ximagesink->running = TRUE;
+
       /* Initializing the XContext */
       if (!ximagesink->xcontext)
-        ximagesink->xcontext = gst_ximagesink_xcontext_get (ximagesink);
+        xcontext = gst_ximagesink_xcontext_get (ximagesink);
+
+      GST_OBJECT_LOCK (ximagesink);
+      ximagesink->running = TRUE;
+      if (xcontext)
+        ximagesink->xcontext = xcontext;
+      GST_OBJECT_UNLOCK (ximagesink);
+
       if (!ximagesink->xcontext) {
         ret = GST_STATE_CHANGE_FAILURE;
         goto beach;
@@ -1350,7 +1408,10 @@ gst_ximagesink_change_state (GstElement * element, GstStateChange transition)
       GST_VIDEO_SINK_HEIGHT (ximagesink) = 0;
       break;
     case GST_STATE_CHANGE_READY_TO_NULL:
+      GST_OBJECT_LOCK (ximagesink);
       ximagesink->running = FALSE;
+      GST_OBJECT_UNLOCK (ximagesink);
+
       if (ximagesink->ximage) {
         gst_buffer_unref (ximagesink->ximage);
         ximagesink->ximage = NULL;
@@ -1359,18 +1420,15 @@ gst_ximagesink_change_state (GstElement * element, GstStateChange transition)
         gst_buffer_unref (ximagesink->cur_image);
         ximagesink->cur_image = NULL;
       }
-      if (ximagesink->buffer_pool)
-        gst_ximagesink_bufferpool_clear (ximagesink);
+
+      gst_ximagesink_bufferpool_clear (ximagesink);
 
       if (ximagesink->xwindow) {
         gst_ximagesink_xwindow_destroy (ximagesink, ximagesink->xwindow);
         ximagesink->xwindow = NULL;
       }
 
-      if (ximagesink->xcontext) {
-        gst_ximagesink_xcontext_clear (ximagesink);
-        ximagesink->xcontext = NULL;
-      }
+      gst_ximagesink_xcontext_clear (ximagesink);
       break;
     default:
       break;
@@ -1423,7 +1481,19 @@ gst_ximagesink_show_frame (GstBaseSink * bsink, GstBuffer * buf)
       ximagesink->ximage = gst_ximagesink_ximage_new (ximagesink,
           GST_BUFFER_CAPS (buf));
       if (!ximagesink->ximage)
+        /* The create method should have posted an informative error */
         goto no_ximage;
+
+      if (ximagesink->ximage->size < GST_BUFFER_SIZE (buf)) {
+        GST_ELEMENT_ERROR (ximagesink, RESOURCE, WRITE,
+            ("Failed to create output image buffer of %dx%d pixels",
+                ximagesink->ximage->width, ximagesink->ximage->height),
+            ("XServer allocated buffer size did not match input buffer"));
+
+        gst_ximagesink_ximage_destroy (ximagesink, ximagesink->ximage);
+        ximagesink->ximage = NULL;
+        goto no_ximage;
+      }
     }
     memcpy (GST_BUFFER_DATA (ximagesink->ximage), GST_BUFFER_DATA (buf),
         MIN (GST_BUFFER_SIZE (buf), ximagesink->ximage->size));
@@ -1437,14 +1507,20 @@ no_ximage:
   {
     /* No image available. That's very bad ! */
     GST_DEBUG ("could not create image");
-    GST_ELEMENT_ERROR (ximagesink, CORE, NEGOTIATION, (NULL),
-        ("Failed creating an XImage in ximagesink chain function."));
     return GST_FLOW_ERROR;
   }
 }
 
-/* Buffer management */
-
+/* Buffer management
+ *
+ * The buffer_alloc function must either return a buffer with given size and
+ * caps or create a buffer with different caps attached to the buffer. This
+ * last option is called reverse negotiation, ie, where the sink suggests a
+ * different format from the upstream peer. 
+ *
+ * We try to do reverse negotiation when our geometry changes and we like a
+ * resized buffer.
+ */
 static GstFlowReturn
 gst_ximagesink_buffer_alloc (GstBaseSink * bsink, guint64 offset, guint size,
     GstCaps * caps, GstBuffer ** buf)
@@ -1452,9 +1528,9 @@ gst_ximagesink_buffer_alloc (GstBaseSink * bsink, guint64 offset, guint size,
   GstXImageSink *ximagesink;
   GstXImageBuffer *ximage = NULL;
   GstStructure *structure = NULL;
-  GstCaps *desired_caps = NULL;
   GstFlowReturn ret = GST_FLOW_OK;
-  gboolean rev_nego = FALSE;
+  GstCaps *alloc_caps;
+  gboolean alloc_unref = FALSE;
   gint width, height;
 
   ximagesink = GST_XIMAGESINK (bsink);
@@ -1463,9 +1539,14 @@ gst_ximagesink_buffer_alloc (GstBaseSink * bsink, guint64 offset, guint size,
       "a buffer of %d bytes was requested with caps %" GST_PTR_FORMAT
       " and offset %" G_GUINT64_FORMAT, size, caps, offset);
 
-  desired_caps = gst_caps_copy (caps);
+  /* assume we're going to alloc what was requested, keep track of
+   * wheter we need to unref or not. When we suggest a new format 
+   * upstream we will create a new caps that we need to unref. */
+  alloc_caps = caps;
+  alloc_unref = FALSE;
 
-  structure = gst_caps_get_structure (desired_caps, 0);
+  /* get struct to see what is requested */
+  structure = gst_caps_get_structure (caps, 0);
 
   if (gst_structure_get_int (structure, "width", &width) &&
       gst_structure_get_int (structure, "height", &height)) {
@@ -1476,7 +1557,6 @@ gst_ximagesink_buffer_alloc (GstBaseSink * bsink, guint64 offset, guint size,
 
     /* We take the flow_lock because the window might go away */
     g_mutex_lock (ximagesink->flow_lock);
-
     if (!ximagesink->xwindow) {
       g_mutex_unlock (ximagesink->flow_lock);
       goto alloc;
@@ -1504,35 +1584,44 @@ gst_ximagesink_buffer_alloc (GstBaseSink * bsink, guint64 offset, guint size,
     /* We would like another geometry */
     if (width != result.w || height != result.h) {
       int nom, den;
-      GstPad *peer = gst_pad_get_peer (GST_VIDEO_SINK_PAD (ximagesink));
+      GstCaps *desired_caps;
+      GstStructure *desired_struct;
 
-      if (!GST_IS_PAD (peer)) {
-        /* Is this situation possible ? */
-        goto alloc;
-      }
+      /* make a copy of the incomming caps to create the new
+       * suggestion. We can't use make_writable because we might
+       * then destroy the original caps which we still need when the
+       * peer does not accept the suggestion. */
+      desired_caps = gst_caps_copy (caps);
+      desired_struct = gst_caps_get_structure (desired_caps, 0);
 
       GST_DEBUG ("we would love to receive a %dx%d video", result.w, result.h);
-      gst_structure_set (structure, "width", G_TYPE_INT, result.w, NULL);
-      gst_structure_set (structure, "height", G_TYPE_INT, result.h, NULL);
+      gst_structure_set (desired_struct, "width", G_TYPE_INT, result.w, NULL);
+      gst_structure_set (desired_struct, "height", G_TYPE_INT, result.h, NULL);
 
       /* PAR property overrides the X calculated one */
       if (ximagesink->par) {
         nom = gst_value_get_fraction_numerator (ximagesink->par);
         den = gst_value_get_fraction_denominator (ximagesink->par);
-        gst_structure_set (structure, "pixel-aspect-ratio",
+        gst_structure_set (desired_struct, "pixel-aspect-ratio",
             GST_TYPE_FRACTION, nom, den, NULL);
       } else if (ximagesink->xcontext->par) {
         nom = gst_value_get_fraction_numerator (ximagesink->xcontext->par);
         den = gst_value_get_fraction_denominator (ximagesink->xcontext->par);
-        gst_structure_set (structure, "pixel-aspect-ratio",
+        gst_structure_set (desired_struct, "pixel-aspect-ratio",
             GST_TYPE_FRACTION, nom, den, NULL);
       }
 
-      if (gst_pad_accept_caps (peer, desired_caps)) {
+      /* see if peer accepts our new suggestion, if there is no peer, this 
+       * function returns true. */
+      if (gst_pad_peer_accept_caps (GST_VIDEO_SINK_PAD (ximagesink),
+              desired_caps)) {
         gint bpp;
 
         bpp = size / height / width;
-        rev_nego = TRUE;
+        /* we will not alloc a buffer of the new suggested caps. Make sure
+         * we also unref this new caps after we set it on the buffer. */
+        alloc_caps = desired_caps;
+        alloc_unref = TRUE;
         width = result.w;
         height = result.h;
         size = bpp * width * height;
@@ -1541,11 +1630,10 @@ gst_ximagesink_buffer_alloc (GstBaseSink * bsink, guint64 offset, guint size,
       } else {
         GST_DEBUG ("peer pad does not accept our desired caps %" GST_PTR_FORMAT,
             desired_caps);
-        rev_nego = FALSE;
+        /* we alloc a buffer with the original incomming caps */
         width = GST_VIDEO_SINK_WIDTH (ximagesink);
         height = GST_VIDEO_SINK_HEIGHT (ximagesink);
       }
-      gst_object_unref (peer);
     }
   }
 
@@ -1574,24 +1662,18 @@ alloc:
 
   /* We haven't found anything, creating a new one */
   if (!ximage) {
-    if (rev_nego) {
-      ximage = gst_ximagesink_ximage_new (ximagesink, desired_caps);
-    } else {
-      ximage = gst_ximagesink_ximage_new (ximagesink, caps);
-    }
+    ximage = gst_ximagesink_ximage_new (ximagesink, alloc_caps);
   }
   /* Now we should have a ximage, set appropriate caps on it */
   if (ximage) {
-    if (rev_nego) {
-      gst_buffer_set_caps (GST_BUFFER (ximage), desired_caps);
-    } else {
-      gst_buffer_set_caps (GST_BUFFER (ximage), caps);
-    }
+    gst_buffer_set_caps (GST_BUFFER_CAST (ximage), alloc_caps);
   }
 
-  gst_caps_unref (desired_caps);
+  /* could be our new reffed suggestion or the original unreffed caps */
+  if (alloc_unref)
+    gst_caps_unref (alloc_caps);
 
-  *buf = GST_BUFFER (ximage);
+  *buf = GST_BUFFER_CAST (ximage);
 
   return ret;
 }
@@ -1822,7 +1904,7 @@ gst_ximagesink_get_property (GObject * object, guint prop_id,
 
   switch (prop_id) {
     case PROP_DISPLAY:
-      g_value_set_string (value, g_strdup (ximagesink->display_name));
+      g_value_set_string (value, ximagesink->display_name);
       break;
     case PROP_SYNCHRONOUS:
       g_value_set_boolean (value, ximagesink->synchronous);
@@ -2006,6 +2088,10 @@ gst_ximagesink_get_type (void)
         &navigation_info);
     g_type_add_interface_static (ximagesink_type, GST_TYPE_X_OVERLAY,
         &overlay_info);
+
+    /* register type in a more safe place instead of at runtime since the
+     * type registration is not threadsafe. */
+    gst_ximage_buffer_get_type ();
   }
 
   return ximagesink_type;
