@@ -125,6 +125,8 @@
 #include <gst/interfaces/navigation.h>
 #include <gst/interfaces/xoverlay.h>
 #include <gst/interfaces/colorbalance.h>
+/* Helper functions */
+#include <gst/video/video.h>
 
 /* Object header */
 #include "xvimagesink.h"
@@ -157,7 +159,7 @@ static void gst_xvimagesink_expose (GstXOverlay * overlay);
 //static void gst_xvimagesink_send_pending_navigation (GstXvImageSink * xvimagesink);
 
 /* ElementFactory information */
-static GstElementDetails gst_xvimagesink_details =
+static const GstElementDetails gst_xvimagesink_details =
 GST_ELEMENT_DETAILS ("Video sink",
     "Sink/Video",
     "A Xv based videosink",
@@ -213,6 +215,8 @@ gst_xvimage_buffer_destroy (GstXvImageBuffer * xvimage)
 {
   GstXvImageSink *xvimagesink;
 
+  GST_DEBUG_OBJECT (xvimage, "Destroying buffer");
+
   xvimagesink = xvimage->xvimagesink;
   if (xvimagesink == NULL)
     goto no_sink;
@@ -224,7 +228,16 @@ gst_xvimage_buffer_destroy (GstXvImageBuffer * xvimage)
     xvimagesink->cur_image = NULL;
 
   /* We might have some buffers destroyed after changing state to NULL */
-  if (xvimagesink->xcontext) {
+  GST_OBJECT_LOCK (xvimagesink);
+  if (xvimagesink->xcontext == NULL) {
+    GST_DEBUG_OBJECT (xvimagesink, "Destroying XvImage after Xcontext");
+#ifdef HAVE_XSHM
+    /* Need to free the shared memory segment even if the x context
+     * was already cleaned up */
+    if (xvimage->SHMInfo.shmaddr != ((void *) -1)) {
+      shmdt (xvimage->SHMInfo.shmaddr);
+    }
+#endif
     goto beach;
   }
 
@@ -233,8 +246,11 @@ gst_xvimage_buffer_destroy (GstXvImageBuffer * xvimage)
 #ifdef HAVE_XSHM
   if (xvimagesink->xcontext->use_xshm) {
     if (xvimage->SHMInfo.shmaddr != ((void *) -1)) {
+      GST_DEBUG_OBJECT (xvimagesink, "XServer ShmDetaching from 0x%x id 0x%x\n",
+          xvimage->SHMInfo.shmid, xvimage->SHMInfo.shmseg);
       XShmDetach (xvimagesink->xcontext->disp, &xvimage->SHMInfo);
       XSync (xvimagesink->xcontext->disp, FALSE);
+
       shmdt (xvimage->SHMInfo.shmaddr);
     }
     if (xvimage->xvimage)
@@ -255,6 +271,7 @@ gst_xvimage_buffer_destroy (GstXvImageBuffer * xvimage)
   g_mutex_unlock (xvimagesink->x_lock);
 
 beach:
+  GST_OBJECT_UNLOCK (xvimagesink);
   xvimage->xvimagesink = NULL;
   gst_object_unref (xvimagesink);
 
@@ -271,13 +288,21 @@ static void
 gst_xvimage_buffer_finalize (GstXvImageBuffer * xvimage)
 {
   GstXvImageSink *xvimagesink;
+  gboolean running;
 
   xvimagesink = xvimage->xvimagesink;
-  if (xvimagesink == NULL)
+  if (G_UNLIKELY (xvimagesink == NULL))
     goto no_sink;
 
+  GST_OBJECT_LOCK (xvimagesink);
+  running = xvimagesink->running;
+  GST_OBJECT_UNLOCK (xvimagesink);
+
   /* If our geometry changed we can't reuse that image. */
-  if ((xvimage->width != xvimagesink->video_width) ||
+  if (running == FALSE) {
+    GST_LOG_OBJECT (xvimage, "destroy image as sink is shutting down");
+    gst_xvimage_buffer_destroy (xvimage);
+  } else if ((xvimage->width != xvimagesink->video_width) ||
       (xvimage->height != xvimagesink->video_height)) {
     GST_LOG_OBJECT (xvimage,
         "destroy image as its size changed %dx%d vs current %dx%d",
@@ -438,6 +463,9 @@ gst_xvimagesink_check_xshm_calls (GstXContext * xcontext)
   /* Sync to ensure we see any errors we caused */
   XSync (xcontext->disp, FALSE);
 
+  GST_DEBUG ("XServer ShmAttached to 0x%x, id 0x%x\n", SHMInfo.shmid,
+      SHMInfo.shmseg);
+
   if (!error_caught) {
     did_attach = TRUE;
     /* store whether we succeeded in result */
@@ -452,6 +480,8 @@ beach:
   XSetErrorHandler (handler);
 
   if (did_attach) {
+    GST_DEBUG ("XServer ShmDetaching from 0x%x id 0x%x\n",
+        SHMInfo.shmid, SHMInfo.shmseg);
     XShmDetach (xcontext->disp, &SHMInfo);
     XSync (xcontext->disp, FALSE);
   }
@@ -474,6 +504,7 @@ gst_xvimagesink_xvimage_new (GstXvImageSink * xvimagesink, GstCaps * caps)
   g_return_val_if_fail (GST_IS_XVIMAGESINK (xvimagesink), NULL);
 
   xvimage = (GstXvImageBuffer *) gst_mini_object_new (GST_TYPE_XVIMAGE_BUFFER);
+  GST_DEBUG_OBJECT (xvimage, "Creating new XvImageBuffer");
 
   structure = gst_caps_get_structure (caps, 0);
 
@@ -486,9 +517,12 @@ gst_xvimagesink_xvimage_new (GstXvImageSink * xvimagesink, GstCaps * caps)
       xvimage->height);
 
   xvimage->im_format = gst_xvimagesink_get_format_from_caps (xvimagesink, caps);
-  if (!xvimage->im_format) {
+  if (xvimage->im_format == -1) {
     GST_WARNING_OBJECT (xvimagesink, "failed to get format from caps %"
         GST_PTR_FORMAT, caps);
+    GST_ELEMENT_ERROR (xvimagesink, RESOURCE, WRITE,
+        ("Failed to create output image buffer of %dx%d pixels",
+            xvimage->width, xvimage->height), ("Invalid input caps"));
     goto beach_unlocked;
   }
   xvimage->xvimagesink = gst_object_ref (xvimagesink);
@@ -502,9 +536,13 @@ gst_xvimagesink_xvimage_new (GstXvImageSink * xvimagesink, GstCaps * caps)
         xvimage->im_format, NULL,
         xvimage->width, xvimage->height, &xvimage->SHMInfo);
     if (!xvimage->xvimage) {
-      GST_ELEMENT_ERROR (xvimagesink, RESOURCE, WRITE, (NULL),
-          ("could not XvShmCreateImage a %dx%d image"));
-      goto beach;
+      g_mutex_unlock (xvimagesink->x_lock);
+      GST_ELEMENT_ERROR (xvimagesink, RESOURCE, WRITE,
+          ("Failed to create output image buffer of %dx%d pixels",
+              xvimage->width, xvimage->height),
+          ("could not XvShmCreateImage a %dx%d image",
+              xvimage->width, xvimage->height));
+      goto beach_unlocked;
     }
 
     /* we have to use the returned data_size for our shm size */
@@ -514,18 +552,24 @@ gst_xvimagesink_xvimage_new (GstXvImageSink * xvimagesink, GstCaps * caps)
     xvimage->SHMInfo.shmid = shmget (IPC_PRIVATE, xvimage->size,
         IPC_CREAT | 0777);
     if (xvimage->SHMInfo.shmid == -1) {
-      GST_ELEMENT_ERROR (xvimagesink, RESOURCE, WRITE, (NULL),
+      g_mutex_unlock (xvimagesink->x_lock);
+      GST_ELEMENT_ERROR (xvimagesink, RESOURCE, WRITE,
+          ("Failed to create output image buffer of %dx%d pixels",
+              xvimage->width, xvimage->height),
           ("could not get shared memory of %d bytes", xvimage->size));
-      goto beach;
+      goto beach_unlocked;
     }
 
     xvimage->SHMInfo.shmaddr = shmat (xvimage->SHMInfo.shmid, 0, 0);
     if (xvimage->SHMInfo.shmaddr == ((void *) -1)) {
-      GST_ELEMENT_ERROR (xvimagesink, RESOURCE, WRITE, (NULL),
+      g_mutex_unlock (xvimagesink->x_lock);
+      GST_ELEMENT_ERROR (xvimagesink, RESOURCE, WRITE,
+          ("Failed to create output image buffer of %dx%d pixels",
+              xvimage->width, xvimage->height),
           ("Failed to shmat: %s", g_strerror (errno)));
       /* Clean up the shared memory segment */
       shmctl (xvimage->SHMInfo.shmid, IPC_RMID, 0);
-      goto beach;
+      goto beach_unlocked;
     }
 
     /* Delete the shared memory segment as soon as we manage to attach.
@@ -537,12 +581,16 @@ gst_xvimagesink_xvimage_new (GstXvImageSink * xvimagesink, GstCaps * caps)
     xvimage->SHMInfo.readOnly = FALSE;
 
     if (XShmAttach (xvimagesink->xcontext->disp, &xvimage->SHMInfo) == 0) {
-      GST_ELEMENT_ERROR (xvimagesink, RESOURCE, WRITE, (NULL),
-          ("Failed to XShmAttach"));
-      goto beach;
+      g_mutex_unlock (xvimagesink->x_lock);
+      GST_ELEMENT_ERROR (xvimagesink, RESOURCE, WRITE,
+          ("Failed to create output image buffer of %dx%d pixels",
+              xvimage->width, xvimage->height), ("Failed to XShmAttach"));
+      goto beach_unlocked;
     }
 
     XSync (xvimagesink->xcontext->disp, FALSE);
+    GST_DEBUG_OBJECT (xvimagesink, "XServer ShmAttached to 0x%x, id 0x%x\n",
+        xvimage->SHMInfo.shmid, xvimage->SHMInfo.shmseg);
   } else
 #endif /* HAVE_XSHM */
   {
@@ -550,9 +598,13 @@ gst_xvimagesink_xvimage_new (GstXvImageSink * xvimagesink, GstCaps * caps)
         xvimagesink->xcontext->xv_port_id,
         xvimage->im_format, NULL, xvimage->width, xvimage->height);
     if (!xvimage->xvimage) {
-      GST_ELEMENT_ERROR (xvimagesink, RESOURCE, WRITE, (NULL),
-          ("could not XvCreateImage a %dx%d image"));
-      goto beach;
+      g_mutex_unlock (xvimagesink->x_lock);
+      GST_ELEMENT_ERROR (xvimagesink, RESOURCE, WRITE,
+          ("Failed to create outputimage buffer of %dx%d pixels",
+              xvimage->width, xvimage->height),
+          ("could not XvCreateImage a %dx%d image",
+              xvimage->width, xvimage->height));
+      goto beach_unlocked;
     }
 
     /* we have to use the returned data_size for our image size */
@@ -566,7 +618,6 @@ gst_xvimagesink_xvimage_new (GstXvImageSink * xvimagesink, GstCaps * caps)
   GST_BUFFER_DATA (xvimage) = (guchar *) xvimage->xvimage->data;
   GST_BUFFER_SIZE (xvimage) = xvimage->size;
 
-beach:
   g_mutex_unlock (xvimagesink->x_lock);
 
 beach_unlocked:
@@ -675,9 +726,11 @@ gst_xvimagesink_xvimage_put (GstXvImageSink * xvimagesink,
 #ifdef HAVE_XSHM
   if (xvimagesink->xcontext->use_xshm) {
     GST_LOG_OBJECT (xvimagesink,
-        "XvShmPutImage with image %dx%d and window %dx%d",
+        "XvShmPutImage with image %dx%d and window %dx%d, from xvimage %"
+        GST_PTR_FORMAT,
         xvimage->width, xvimage->height,
-        xvimagesink->xwindow->width, xvimagesink->xwindow->height);
+        xvimagesink->xwindow->width, xvimagesink->xwindow->height, xvimage);
+
     XvShmPutImage (xvimagesink->xcontext->disp,
         xvimagesink->xcontext->xv_port_id,
         xvimagesink->xwindow->win,
@@ -1053,13 +1106,18 @@ gst_xvimagesink_get_xv_support (GstXvImageSink * xvimagesink,
   XvAdaptorInfo *adaptors;
   gint nb_formats;
   XvImageFormatValues *formats = NULL;
+  guint nb_encodings;
+  XvEncodingInfo *encodings = NULL;
+  gulong max_w = G_MAXINT, max_h = G_MAXINT;
   GstCaps *caps = NULL;
+  GstCaps *rgb_caps = NULL;
 
   g_return_val_if_fail (xcontext != NULL, NULL);
 
   /* First let's check that XVideo extension is available */
   if (!XQueryExtension (xcontext->disp, "XVideo", &i, &i, &i)) {
-    GST_ELEMENT_ERROR (xvimagesink, RESOURCE, SETTINGS, (NULL),
+    GST_ELEMENT_ERROR (xvimagesink, RESOURCE, SETTINGS,
+        ("Could not initialise Xv output"),
         ("XVideo extension is not available"));
     return NULL;
   }
@@ -1067,7 +1125,8 @@ gst_xvimagesink_get_xv_support (GstXvImageSink * xvimagesink,
   /* Then we get adaptors list */
   if (Success != XvQueryAdaptors (xcontext->disp, xcontext->root,
           &nb_adaptors, &adaptors)) {
-    GST_ELEMENT_ERROR (xvimagesink, RESOURCE, SETTINGS, (NULL),
+    GST_ELEMENT_ERROR (xvimagesink, RESOURCE, SETTINGS,
+        ("Could not initialise Xv output"),
         ("Failed getting XV adaptors list"));
     return NULL;
   }
@@ -1097,8 +1156,8 @@ gst_xvimagesink_get_xv_support (GstXvImageSink * xvimagesink,
   XvFreeAdaptorInfo (adaptors);
 
   if (!xcontext->xv_port_id) {
-    GST_ELEMENT_ERROR (xvimagesink, RESOURCE, BUSY, (NULL),
-        ("No port available"));
+    GST_ELEMENT_ERROR (xvimagesink, RESOURCE, BUSY,
+        ("Could not initialise Xv output"), ("No port available"));
     return NULL;
   }
 
@@ -1141,12 +1200,32 @@ gst_xvimagesink_get_xv_support (GstXvImageSink * xvimagesink,
     XFree (attr);
   }
 
+  /* Get the list of encodings supported by the adapter and look for the
+   * XV_IMAGE encoding so we can determine the maximum width and height
+   * supported */
+  XvQueryEncodings (xcontext->disp, xcontext->xv_port_id, &nb_encodings,
+      &encodings);
+
+  for (i = 0; i < nb_encodings; i++) {
+    GST_LOG_OBJECT (xvimagesink,
+        "Encoding %d, name %s, max wxh %lux%lu rate %d/%d",
+        i, encodings[i].name, encodings[i].width, encodings[i].height,
+        encodings[i].rate.numerator, encodings[i].rate.denominator);
+    if (strcmp (encodings[i].name, "XV_IMAGE") == 0) {
+      max_w = encodings[i].width;
+      max_h = encodings[i].height;
+    }
+  }
+
+  XvFreeEncodingInfo (encodings);
+
   /* We get all image formats supported by our port */
   formats = XvListImageFormats (xcontext->disp,
       xcontext->xv_port_id, &nb_formats);
   caps = gst_caps_new_empty ();
   for (i = 0; i < nb_formats; i++) {
     GstCaps *format_caps = NULL;
+    gboolean is_rgb_format = FALSE;
 
     /* We set the image format of the xcontext to an existing one. Sink
        connect method will override that but we need to have at least a
@@ -1157,23 +1236,44 @@ gst_xvimagesink_get_xv_support (GstXvImageSink * xvimagesink,
     switch (formats[i].type) {
       case XvRGB:
       {
+        XvImageFormatValues *fmt = &(formats[i]);
+        gint endianness = G_BIG_ENDIAN;
+
+        if (fmt->byte_order == LSBFirst) {
+          /* our caps system handles 24/32bpp RGB as big-endian. */
+          if (fmt->bits_per_pixel == 24 || fmt->bits_per_pixel == 32) {
+            fmt->red_mask = GUINT32_TO_BE (fmt->red_mask);
+            fmt->green_mask = GUINT32_TO_BE (fmt->green_mask);
+            fmt->blue_mask = GUINT32_TO_BE (fmt->blue_mask);
+
+            if (fmt->bits_per_pixel == 24) {
+              fmt->red_mask >>= 8;
+              fmt->green_mask >>= 8;
+              fmt->blue_mask >>= 8;
+            }
+          } else
+            endianness = G_LITTLE_ENDIAN;
+        }
+
         format_caps = gst_caps_new_simple ("video/x-raw-rgb",
-            "endianness", G_TYPE_INT, xcontext->endianness,
-            "depth", G_TYPE_INT, xcontext->depth,
-            "bpp", G_TYPE_INT, xcontext->bpp,
-            "blue_mask", G_TYPE_INT, formats[i].red_mask,
-            "green_mask", G_TYPE_INT, formats[i].green_mask,
-            "red_mask", G_TYPE_INT, formats[i].blue_mask,
-            "width", GST_TYPE_INT_RANGE, 1, G_MAXINT,
-            "height", GST_TYPE_INT_RANGE, 1, G_MAXINT,
+            "endianness", G_TYPE_INT, endianness,
+            "depth", G_TYPE_INT, fmt->depth,
+            "bpp", G_TYPE_INT, fmt->bits_per_pixel,
+            "blue_mask", G_TYPE_INT, fmt->red_mask,
+            "green_mask", G_TYPE_INT, fmt->green_mask,
+            "red_mask", G_TYPE_INT, fmt->blue_mask,
+            "width", GST_TYPE_INT_RANGE, 1, max_w,
+            "height", GST_TYPE_INT_RANGE, 1, max_h,
             "framerate", GST_TYPE_FRACTION_RANGE, 0, 1, G_MAXINT, 1, NULL);
+
+        is_rgb_format = TRUE;
         break;
       }
       case XvYUV:
         format_caps = gst_caps_new_simple ("video/x-raw-yuv",
             "format", GST_TYPE_FOURCC, formats[i].id,
-            "width", GST_TYPE_INT_RANGE, 1, G_MAXINT,
-            "height", GST_TYPE_INT_RANGE, 1, G_MAXINT,
+            "width", GST_TYPE_INT_RANGE, 1, max_w,
+            "height", GST_TYPE_INT_RANGE, 1, max_h,
             "framerate", GST_TYPE_FRACTION_RANGE, 0, 1, G_MAXINT, 1, NULL);
         break;
       default:
@@ -1190,11 +1290,22 @@ gst_xvimagesink_get_xv_support (GstXvImageSink * xvimagesink,
         format->caps = gst_caps_copy (format_caps);
         xcontext->formats_list = g_list_append (xcontext->formats_list, format);
       }
-    }
 
-    gst_caps_append (caps, format_caps);
+      if (is_rgb_format) {
+        if (rgb_caps == NULL)
+          rgb_caps = format_caps;
+        else
+          gst_caps_append (rgb_caps, format_caps);
+      } else
+        gst_caps_append (caps, format_caps);
+    }
   }
 
+  /* Collected all caps into either the caps or rgb_caps structures.
+   * Append rgb_caps on the end of YUV, so that YUV is always preferred */
+  if (rgb_caps)
+    gst_caps_append (caps, rgb_caps);
+
   if (formats)
     XFree (formats);
 
@@ -1310,8 +1421,8 @@ gst_xvimagesink_xcontext_get (GstXvImageSink * xvimagesink)
   if (!xcontext->disp) {
     g_mutex_unlock (xvimagesink->x_lock);
     g_free (xcontext);
-    GST_ELEMENT_ERROR (xvimagesink, RESOURCE, WRITE, (NULL),
-        ("Could not open display"));
+    GST_ELEMENT_ERROR (xvimagesink, RESOURCE, WRITE,
+        ("Could not initialise Xv output"), ("Could not open display"));
     return NULL;
   }
 
@@ -1340,8 +1451,8 @@ gst_xvimagesink_xcontext_get (GstXvImageSink * xvimagesink)
     XCloseDisplay (xcontext->disp);
     g_mutex_unlock (xvimagesink->x_lock);
     g_free (xcontext);
-    GST_ELEMENT_ERROR (xvimagesink, RESOURCE, SETTINGS, (NULL),
-        ("Could not get pixel formats"));
+    GST_ELEMENT_ERROR (xvimagesink, RESOURCE, SETTINGS,
+        ("Could not initialise Xv output"), ("Could not get pixel formats"));
     return NULL;
   }
 
@@ -1459,9 +1570,21 @@ static void
 gst_xvimagesink_xcontext_clear (GstXvImageSink * xvimagesink)
 {
   GList *formats_list, *channels_list;
+  GstXContext *xcontext;
 
   g_return_if_fail (GST_IS_XVIMAGESINK (xvimagesink));
-  g_return_if_fail (xvimagesink->xcontext != NULL);
+
+  GST_OBJECT_LOCK (xvimagesink);
+  if (xvimagesink->xcontext == NULL) {
+    GST_OBJECT_UNLOCK (xvimagesink);
+    return;
+  }
+
+  /* Take the XContext from the sink and clean it up */
+  xcontext = xvimagesink->xcontext;
+  xvimagesink->xcontext = NULL;
+
+  GST_OBJECT_UNLOCK (xvimagesink);
 
   /* Wait for our event thread */
   if (xvimagesink->event_thread) {
@@ -1469,7 +1592,7 @@ gst_xvimagesink_xcontext_clear (GstXvImageSink * xvimagesink)
     xvimagesink->event_thread = NULL;
   }
 
-  formats_list = xvimagesink->xcontext->formats_list;
+  formats_list = xcontext->formats_list;
 
   while (formats_list) {
     GstXvImageFormat *format = formats_list->data;
@@ -1479,10 +1602,10 @@ gst_xvimagesink_xcontext_clear (GstXvImageSink * xvimagesink)
     formats_list = g_list_next (formats_list);
   }
 
-  if (xvimagesink->xcontext->formats_list)
-    g_list_free (xvimagesink->xcontext->formats_list);
+  if (xcontext->formats_list)
+    g_list_free (xcontext->formats_list);
 
-  channels_list = xvimagesink->xcontext->channels_list;
+  channels_list = xcontext->channels_list;
 
   while (channels_list) {
     GstColorBalanceChannel *channel = channels_list->data;
@@ -1491,23 +1614,26 @@ gst_xvimagesink_xcontext_clear (GstXvImageSink * xvimagesink)
     channels_list = g_list_next (channels_list);
   }
 
-  if (xvimagesink->xcontext->channels_list)
-    g_list_free (xvimagesink->xcontext->channels_list);
+  if (xcontext->channels_list)
+    g_list_free (xcontext->channels_list);
 
-  gst_caps_unref (xvimagesink->xcontext->caps);
-  g_free (xvimagesink->xcontext->par);
+  gst_caps_unref (xcontext->caps);
+  if (xcontext->last_caps)
+    gst_caps_replace (&xcontext->last_caps, NULL);
+
+  g_free (xcontext->par);
 
   g_mutex_lock (xvimagesink->x_lock);
 
-  XvUngrabPort (xvimagesink->xcontext->disp,
-      xvimagesink->xcontext->xv_port_id, 0);
+  GST_DEBUG_OBJECT (xvimagesink, "Closing display and freeing X Context");
 
-  XCloseDisplay (xvimagesink->xcontext->disp);
+  XvUngrabPort (xcontext->disp, xcontext->xv_port_id, 0);
+
+  XCloseDisplay (xcontext->disp);
 
   g_mutex_unlock (xvimagesink->x_lock);
 
-  g_free (xvimagesink->xcontext);
-  xvimagesink->xcontext = NULL;
+  g_free (xcontext);
 }
 
 static void
@@ -1556,7 +1682,7 @@ gst_xvimagesink_get_format_from_caps (GstXvImageSink * xvimagesink,
     list = g_list_next (list);
   }
 
-  return 0;
+  return -1;
 }
 
 static GstCaps *
@@ -1585,10 +1711,9 @@ gst_xvimagesink_setcaps (GstBaseSink * bsink, GstCaps * caps)
   gint video_width, video_height;
   gint video_par_n, video_par_d;        /* video's PAR */
   gint display_par_n, display_par_d;    /* display's PAR */
-  GValue display_ratio = { 0, };        /* display w/h ratio */
   const GValue *caps_par;
   const GValue *fps;
-  gint num, den;
+  guint num, den;
 
   xvimagesink = GST_XVIMAGESINK (bsink);
 
@@ -1611,8 +1736,11 @@ gst_xvimagesink_setcaps (GstBaseSink * bsink, GstCaps * caps)
   fps = gst_structure_get_value (structure, "framerate");
   ret &= (fps != NULL);
 
-  if (!ret)
+  if (!ret) {
+    GST_DEBUG_OBJECT (xvimagesink, "Failed to retrieve either width, "
+        "height or framerate from intersected caps");
     return FALSE;
+  }
 
   xvimagesink->fps_n = gst_value_get_fraction_numerator (fps);
   xvimagesink->fps_d = gst_value_get_fraction_denominator (fps);
@@ -1620,15 +1748,15 @@ gst_xvimagesink_setcaps (GstBaseSink * bsink, GstCaps * caps)
   xvimagesink->video_width = video_width;
   xvimagesink->video_height = video_height;
   im_format = gst_xvimagesink_get_format_from_caps (xvimagesink, caps);
-  if (im_format == 0) {
+  if (im_format == -1) {
+    GST_DEBUG_OBJECT (xvimagesink,
+        "Could not locate image format from caps %" GST_PTR_FORMAT, caps);
     return FALSE;
   }
 
   /* get aspect ratio from caps if it's present, and
    * convert video width and height to a display width and height
-   * using wd / hd = wv / hv * PARv / PARd
-   * the ratio wd / hd will be stored in display_ratio */
-  g_value_init (&display_ratio, GST_TYPE_FRACTION);
+   * using wd / hd = wv / hv * PARv / PARd */
 
   /* get video's PAR */
   caps_par = gst_structure_get_value (structure, "pixel-aspect-ratio");
@@ -1648,12 +1776,14 @@ gst_xvimagesink_setcaps (GstBaseSink * bsink, GstCaps * caps)
     display_par_d = 1;
   }
 
-  gst_value_set_fraction (&display_ratio,
-      video_width * video_par_n * display_par_d,
-      video_height * video_par_d * display_par_n);
+  if (!gst_video_calculate_display_ratio (&num, &den, video_width,
+          video_height, video_par_n, video_par_d, display_par_n,
+          display_par_d)) {
+    GST_ELEMENT_ERROR (xvimagesink, CORE, NEGOTIATION, (NULL),
+        ("Error calculating the output display ratio of the video."));
+    return FALSE;
+  }
 
-  num = gst_value_get_fraction_numerator (&display_ratio);
-  den = gst_value_get_fraction_denominator (&display_ratio);
   GST_DEBUG_OBJECT (xvimagesink,
       "video width/height: %dx%d, calculated display ratio: %d/%d",
       video_width, video_height, num, den);
@@ -1666,15 +1796,18 @@ gst_xvimagesink_setcaps (GstBaseSink * bsink, GstCaps * caps)
   /* check hd / den is an integer scale factor, and scale wd with the PAR */
   if (video_height % den == 0) {
     GST_DEBUG_OBJECT (xvimagesink, "keeping video height");
-    GST_VIDEO_SINK_WIDTH (xvimagesink) = video_height * num / den;
+    GST_VIDEO_SINK_WIDTH (xvimagesink) = (guint)
+        gst_util_uint64_scale_int (video_height, num, den);
     GST_VIDEO_SINK_HEIGHT (xvimagesink) = video_height;
   } else if (video_width % num == 0) {
     GST_DEBUG_OBJECT (xvimagesink, "keeping video width");
     GST_VIDEO_SINK_WIDTH (xvimagesink) = video_width;
-    GST_VIDEO_SINK_HEIGHT (xvimagesink) = video_width * den / num;
+    GST_VIDEO_SINK_HEIGHT (xvimagesink) = (guint)
+        gst_util_uint64_scale_int (video_width, den, num);
   } else {
     GST_DEBUG_OBJECT (xvimagesink, "approximating while keeping video height");
-    GST_VIDEO_SINK_WIDTH (xvimagesink) = video_height * num / den;
+    GST_VIDEO_SINK_WIDTH (xvimagesink) = (guint)
+        gst_util_uint64_scale_int (video_height, num, den);
     GST_VIDEO_SINK_HEIGHT (xvimagesink) = video_height;
   }
   GST_DEBUG_OBJECT (xvimagesink, "scaling to %dx%d",
@@ -1686,8 +1819,13 @@ gst_xvimagesink_setcaps (GstBaseSink * bsink, GstCaps * caps)
   }
 
   /* Creating our window and our image with the display size in pixels */
-  g_assert (GST_VIDEO_SINK_WIDTH (xvimagesink) > 0);
-  g_assert (GST_VIDEO_SINK_HEIGHT (xvimagesink) > 0);
+  if (GST_VIDEO_SINK_WIDTH (xvimagesink) <= 0 ||
+      GST_VIDEO_SINK_HEIGHT (xvimagesink) <= 0) {
+    GST_ELEMENT_ERROR (xvimagesink, CORE, NEGOTIATION, (NULL),
+        ("Error calculating the output display ratio of the video."));
+    return FALSE;
+  }
+
   if (!xvimagesink->xwindow)
     xvimagesink->xwindow = gst_xvimagesink_xwindow_new (xvimagesink,
         GST_VIDEO_SINK_WIDTH (xvimagesink),
@@ -1718,16 +1856,25 @@ gst_xvimagesink_change_state (GstElement * element, GstStateChange transition)
 {
   GstStateChangeReturn ret = GST_STATE_CHANGE_SUCCESS;
   GstXvImageSink *xvimagesink;
+  GstXContext *xcontext = NULL;
 
   xvimagesink = GST_XVIMAGESINK (element);
 
   switch (transition) {
     case GST_STATE_CHANGE_NULL_TO_READY:
-      xvimagesink->running = TRUE;
       /* Initializing the XContext */
-      if (!xvimagesink->xcontext &&
-          !(xvimagesink->xcontext = gst_xvimagesink_xcontext_get (xvimagesink)))
-        return GST_STATE_CHANGE_FAILURE;
+      if (xvimagesink->xcontext == NULL) {
+        xcontext = gst_xvimagesink_xcontext_get (xvimagesink);
+        if (xcontext == NULL)
+          return GST_STATE_CHANGE_FAILURE;
+      }
+
+      GST_OBJECT_LOCK (xvimagesink);
+      xvimagesink->running = TRUE;
+      if (xcontext)
+        xvimagesink->xcontext = xcontext;
+      GST_OBJECT_UNLOCK (xvimagesink);
+
       /* update object's par with calculated one if not set yet */
       if (!xvimagesink->par) {
         xvimagesink->par = g_new0 (GValue, 1);
@@ -1764,7 +1911,9 @@ gst_xvimagesink_change_state (GstElement * element, GstStateChange transition)
       GST_VIDEO_SINK_HEIGHT (xvimagesink) = 0;
       break;
     case GST_STATE_CHANGE_READY_TO_NULL:
+      GST_OBJECT_LOCK (xvimagesink);
       xvimagesink->running = FALSE;
+      GST_OBJECT_UNLOCK (xvimagesink);
       if (xvimagesink->cur_image) {
         gst_buffer_unref (xvimagesink->cur_image);
         xvimagesink->cur_image = NULL;
@@ -1773,18 +1922,15 @@ gst_xvimagesink_change_state (GstElement * element, GstStateChange transition)
         gst_buffer_unref (xvimagesink->xvimage);
         xvimagesink->xvimage = NULL;
       }
-      if (xvimagesink->image_pool)
-        gst_xvimagesink_imagepool_clear (xvimagesink);
+
+      gst_xvimagesink_imagepool_clear (xvimagesink);
 
       if (xvimagesink->xwindow) {
         gst_xvimagesink_xwindow_destroy (xvimagesink, xvimagesink->xwindow);
         xvimagesink->xwindow = NULL;
       }
 
-      if (xvimagesink->xcontext) {
-        gst_xvimagesink_xcontext_clear (xvimagesink);
-        xvimagesink->xcontext = NULL;
-      }
+      gst_xvimagesink_xcontext_clear (xvimagesink);
       break;
     default:
       break;
@@ -1835,9 +1981,20 @@ gst_xvimagesink_show_frame (GstBaseSink * bsink, GstBuffer * buf)
       xvimagesink->xvimage = gst_xvimagesink_xvimage_new (xvimagesink,
           GST_BUFFER_CAPS (buf));
 
-      if ((!xvimagesink->xvimage) ||
-          (xvimagesink->xvimage->size < GST_BUFFER_SIZE (buf)))
+      if (!xvimagesink->xvimage)
+        /* The create method should have posted an informative error */
         goto no_image;
+
+      if (xvimagesink->xvimage->size < GST_BUFFER_SIZE (buf)) {
+        GST_ELEMENT_ERROR (xvimagesink, RESOURCE, WRITE,
+            ("Failed to create output image buffer of %dx%d pixels",
+                xvimagesink->xvimage->width, xvimagesink->xvimage->height),
+            ("XServer allocated buffer size did not match input buffer"));
+
+        gst_xvimage_buffer_destroy (xvimagesink->xvimage);
+        xvimagesink->xvimage = NULL;
+        goto no_image;
+      }
     }
 
     memcpy (xvimagesink->xvimage->xvimage->data,
@@ -1854,8 +2011,6 @@ no_image:
   {
     /* No image available. That's very bad ! */
     GST_WARNING_OBJECT (xvimagesink, "could not create image");
-    GST_ELEMENT_ERROR (xvimagesink, CORE, NEGOTIATION, (NULL),
-        ("Failed creating an XvImage in xvimagesink chain function."));
     return GST_FLOW_ERROR;
   }
 }
@@ -1875,12 +2030,26 @@ gst_xvimagesink_buffer_alloc (GstBaseSink * bsink, guint64 offset, guint size,
 
   xvimagesink = GST_XVIMAGESINK (bsink);
 
+  if (G_LIKELY (xvimagesink->xcontext->last_caps &&
+          gst_caps_is_equal (caps, xvimagesink->xcontext->last_caps))) {
+    GST_DEBUG_OBJECT (xvimagesink,
+        "buffer alloc for same last_caps, reusing caps");
+    intersection = gst_caps_ref (caps);
+    image_format = xvimagesink->xcontext->last_format;
+
+    goto reuse_last_caps;
+  }
+
   GST_DEBUG_OBJECT (xvimagesink, "buffer alloc requested with caps %"
       GST_PTR_FORMAT "intersecting with our caps %" GST_PTR_FORMAT, caps,
       xvimagesink->xcontext->caps);
 
   /* Check the caps against our xcontext */
   intersection = gst_caps_intersect (xvimagesink->xcontext->caps, caps);
+
+  /* Ensure the returned caps are fixed */
+  gst_caps_truncate (intersection);
+
   GST_DEBUG_OBJECT (xvimagesink, "intersection in buffer alloc returned %"
       GST_PTR_FORMAT, intersection);
 
@@ -1930,16 +2099,28 @@ gst_xvimagesink_buffer_alloc (GstBaseSink * bsink, guint64 offset, guint size,
 
     GST_DEBUG_OBJECT (xvimagesink, "allocating a buffer with caps %"
         GST_PTR_FORMAT, intersection);
+  } else if (gst_caps_is_equal (intersection, caps)) {
+    /* Things work better if we return a buffer with the same caps ptr
+     * as was asked for when we can */
+    gst_caps_replace (&intersection, caps);
   }
 
   /* Get image format from caps */
   image_format = gst_xvimagesink_get_format_from_caps (xvimagesink,
       intersection);
 
+  /* Store our caps and format as the last_caps to avoid expensive 
+   * caps intersection next time */
+  gst_caps_replace (&xvimagesink->xcontext->last_caps, intersection);
+  xvimagesink->xcontext->last_format = image_format;
+
+reuse_last_caps:
+
   /* Get geometry from caps */
   structure = gst_caps_get_structure (intersection, 0);
   if (!gst_structure_get_int (structure, "width", &width) ||
-      !gst_structure_get_int (structure, "height", &height) || !image_format) {
+      !gst_structure_get_int (structure, "height", &height) ||
+      image_format == -1) {
     GST_WARNING_OBJECT (xvimagesink, "invalid caps for buffer allocation %"
         GST_PTR_FORMAT, intersection);
     ret = GST_FLOW_UNEXPECTED;
@@ -1978,7 +2159,7 @@ gst_xvimagesink_buffer_alloc (GstBaseSink * bsink, guint64 offset, guint size,
     /* We found no suitable image in the pool. Creating... */
     GST_DEBUG_OBJECT (xvimagesink, "no usable image in pool, creating xvimage");
     xvimage = gst_xvimagesink_xvimage_new (xvimagesink, intersection);
-    if (xvimage->size < size) {
+    if (xvimage && xvimage->size < size) {
       /* This image is unusable. Destroying... */
       gst_xvimage_buffer_free (xvimage);
       xvimage = NULL;
@@ -2353,7 +2534,7 @@ gst_xvimagesink_get_property (GObject * object, guint prop_id,
       g_value_set_int (value, xvimagesink->saturation);
       break;
     case ARG_DISPLAY:
-      g_value_set_string (value, g_strdup (xvimagesink->display_name));
+      g_value_set_string (value, xvimagesink->display_name);
       break;
     case ARG_SYNCHRONOUS:
       g_value_set_boolean (value, xvimagesink->synchronous);
@@ -2566,6 +2747,10 @@ gst_xvimagesink_get_type (void)
         &overlay_info);
     g_type_add_interface_static (xvimagesink_type, GST_TYPE_COLOR_BALANCE,
         &colorbalance_info);
+
+    /* make sure type is registered and not at runtime as this is not
+     * threadsafe */
+    gst_xvimage_buffer_get_type ();
   }
 
   return xvimagesink_type;
